#ifndef CFMUVALIDATE_H
#define CFMUVALIDATE_H

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "sysdepsgui.h"

#include <gtkmm.h>
#include <sigc++/sigc++.h>
#ifdef HAVE_WEBKITGTK
#include <webkit/webkit.h>
#endif

#include "tfr.hh"

class CFMUValidate : public Gtk::Main {
protected:
	class MainWindow : public Gtk::Window {
	public:
		explicit MainWindow(BaseObjectType* cobject, const Glib::RefPtr<Builder>& builder);
		virtual ~MainWindow();
		void set_app(CFMUValidate *app);
		void set_dbdirs(const std::string& dir_main = "", Engine::auxdb_mode_t auxdbmode = Engine::auxdb_prefs, const std::string& dir_aux = "");

	protected:
		AirportsDb m_airportdb;
		NavaidsDb m_navaiddb;
		WaypointsDb m_waypointdb;
		AirwaysDb m_airwaydb;
		AirspacesDb m_airspacedb;
		Engine *m_engine;
		TrafficFlowRestrictions m_tfr;
		std::string m_dir_main;
		std::string m_dir_aux;
		Glib::RefPtr<Gtk::TextTag> m_txttagmsginfo;
		Glib::RefPtr<Gtk::TextTag> m_txttagmsgwarn;
		Glib::RefPtr<Gtk::TextTag> m_txttagrule;
		Glib::RefPtr<Gtk::TextTag> m_txttagdesc;
		Glib::RefPtr<Gtk::TextTag> m_txttagoprgoal;
		Glib::RefPtr<Gtk::TextTag> m_txttagcond;
		Glib::RefPtr<Gtk::TextTag> m_txttagrestrict;
		Glib::RefPtr<Gtk::TextTag> m_txttagfpl;
		Glib::RefPtr<Builder> m_builder;
		Gtk::AboutDialog *m_aboutwindow;
		Gtk::TextView *m_textviewfplan;
		Gtk::TextView *m_textviewresults;
		Gtk::ScrolledWindow *m_scrolledwindowbrowser;
		Gtk::Button *m_buttonabout;
		Gtk::Button *m_buttonclear;
		Gtk::Button *m_buttonvalidate;
		Gtk::Button *m_buttonvalidatelocal;
		Gtk::Button *m_buttonquit;
		Gtk::Widget *m_portal_webview;
		Gtk::Widget *m_validate_webview;
		gulong m_portal_webview_close_sighandler;
		gulong m_portal_webview_create_sighandler;
		gulong m_portal_webview_ready_sighandler;
		gulong m_portal_webview_loadstatus_sighandler;
		gulong m_portal_webview_navpolicy_sighandler;
		gulong m_portal_webview_newwndpolicy_sighandler;
		gulong m_portal_webview_rsrcload_sighandler;
		gulong m_portal_webview_consolemsg_sighandler;
		gulong m_validate_webview_close_sighandler;
		gulong m_validate_webview_create_sighandler;
		gulong m_validate_webview_ready_sighandler;
		gulong m_validate_webview_loadstatus_sighandler;
		gulong m_validate_webview_navpolicy_sighandler;
		gulong m_validate_webview_newwndpolicy_sighandler;
		gulong m_validate_webview_rsrcload_sighandler;
		gulong m_validate_webview_consolemsg_sighandler;
		CFMUValidate *m_app;
		typedef enum {
			wvstate_loadmainpage,
			wvstate_loadquerypage,
			wvstate_ready,
			wvstate_query,
			wvstate_waitresult,
			wvstate_failed
		} wvstate_t;
		wvstate_t m_wvstate;
		Engine::auxdb_mode_t m_auxdbmode;
		bool m_tfrloaded;

		void validate_clicked(void);
		void validate_local_clicked(void);
		void clear_clicked(void);
		void quit_clicked(void);
		void about_clicked(void);
		void about_done(int resp);
		void fpl_changed(void);
		void add_portal_sighandlers(void);
		void remove_portal_sighandlers(void);
		void add_validate_sighandlers(void);
		void remove_validate_sighandlers(void);
		void prepare_webview(void);
		void set_wvstate(wvstate_t newstate);
		void parse_results(void);
		void zap_errors(void);
		static gboolean close_webview_1(WebKitWebView* webview, gpointer userdata);
		gboolean close_webview(WebKitWebView* webview);
		static WebKitWebView *create_webview_1(WebKitWebView *webview, WebKitWebFrame *frame, gpointer userdata);
		WebKitWebView *create_webview(WebKitWebView *webview, WebKitWebFrame *frame);
		static gboolean webview_ready_1(WebKitWebView* webview, gpointer userdata);
		gboolean webview_ready(WebKitWebView* webview);
		static void loadstatus_webview_1(WebKitWebView *webview, GParamSpec *pspec, gpointer userdata);
		void loadstatus_webview(WebKitWebView *webview, GParamSpec *pspec);
		static gboolean policy_decision_required_1(WebKitWebView *webview, WebKitWebFrame *webframe, WebKitNetworkRequest *request,
							   WebKitWebNavigationAction *action, WebKitWebPolicyDecision *decision, gpointer userdata);
		gboolean policy_decision_required(WebKitWebView *webview, WebKitWebFrame *webframe, WebKitNetworkRequest *request,
						  WebKitWebNavigationAction *action, WebKitWebPolicyDecision *decision);
		static void resourceload_finished_1(WebKitWebView *webview, WebKitWebFrame *webframe, WebKitWebResource *webresource, gpointer userdata);
		void resourceload_finished(WebKitWebView *webview, WebKitWebFrame *webframe, WebKitWebResource *webresource);
		static gboolean result_change_1(WebKitDOMElement *target, WebKitDOMEvent *e, gpointer userdata);
		gboolean result_change(WebKitDOMElement *target, WebKitDOMEvent *e);
		static gboolean console_message_1(WebKitWebView *webview, gchar *message, gint line, gchar *source_id, gpointer userdata);
		gboolean console_message(WebKitWebView *webview, gchar *message, gint line, gchar *source_id);

		void add_resulttext(const Glib::ustring& txt, const Glib::RefPtr<Gtk::TextTag>& tag);
		void add_resulttext(const TrafficFlowRestrictions::Message& m);
		void add_resulttext(const TrafficFlowRestrictions::Result& res);
		void add_resulttext(const TrafficFlowRestrictions::RuleResult& rule);
		void print_tracerules(void);
		void print_disabledrules(void);
	};

public:
	CFMUValidate(int& argc, char**& argv, Glib::OptionContext& option_context);
	~CFMUValidate();
	void run(void);
	void set_dbdirs(const std::string& dir_main = "", Engine::auxdb_mode_t auxdbmode = Engine::auxdb_prefs, const std::string& dir_aux = "");

protected:
	Glib::RefPtr<Builder> m_builder;
	MainWindow *m_mainwindow;
};

#endif /* CFMUVALIDATE_H */
