//
// C++ Implementation: FlightDeckWindow
//
// Description: Flight Deck Windows, Text Entry Dialog
//
//
// Author: Thomas Sailer <t.sailer@alumni.ethz.ch>, (C) 2011, 2012, 2013
//
// Copyright: See COPYING file that comes with this distribution
//
//

#include "sysdeps.h"
#include <iomanip>

#include "flightdeckwindow.h"
#include "icaofpl.h"

const unsigned int FlightDeckWindow::FPLAutoRouteDialog::nrkeybuttons;
const struct FlightDeckWindow::FPLAutoRouteDialog::keybutton FlightDeckWindow::FPLAutoRouteDialog::keybuttons[nrkeybuttons] = {
	{ "fplautoroutebutton0", GDK_KEY_0, GDK_KEY_parenright, 19 },
	{ "fplautoroutebutton1", GDK_KEY_1, GDK_KEY_exclam, 10 },
	{ "fplautoroutebutton2", GDK_KEY_2, GDK_KEY_at, 11 },
	{ "fplautoroutebutton3", GDK_KEY_3, GDK_KEY_numbersign, 12 },
	{ "fplautoroutebutton4", GDK_KEY_4, GDK_KEY_dollar, 13 },
	{ "fplautoroutebutton5", GDK_KEY_5, GDK_KEY_percent, 14 },
	{ "fplautoroutebutton6", GDK_KEY_6, GDK_KEY_asciicircum, 15 },
	{ "fplautoroutebutton7", GDK_KEY_7, GDK_KEY_ampersand, 16 },
	{ "fplautoroutebutton8", GDK_KEY_8, GDK_KEY_asterisk, 17 },
	{ "fplautoroutebutton9", GDK_KEY_9, GDK_KEY_parenleft, 18 },
	{ "fplautoroutebuttona", GDK_KEY_a, GDK_KEY_A, 38 },
	{ "fplautoroutebuttonb", GDK_KEY_b, GDK_KEY_B, 56 },
	{ "fplautoroutebuttonc", GDK_KEY_c, GDK_KEY_C, 54 },
	{ "fplautoroutebuttond", GDK_KEY_d, GDK_KEY_D, 40 },
	{ "fplautoroutebuttone", GDK_KEY_e, GDK_KEY_E, 26 },
	{ "fplautoroutebuttonf", GDK_KEY_f, GDK_KEY_F, 41 },
	{ "fplautoroutebuttong", GDK_KEY_g, GDK_KEY_G, 42 },
	{ "fplautoroutebuttonh", GDK_KEY_h, GDK_KEY_H, 43 },
	{ "fplautoroutebuttoni", GDK_KEY_i, GDK_KEY_I, 31 },
	{ "fplautoroutebuttonj", GDK_KEY_j, GDK_KEY_J, 44 },
	{ "fplautoroutebuttonk", GDK_KEY_k, GDK_KEY_K, 45 },
	{ "fplautoroutebuttonl", GDK_KEY_l, GDK_KEY_L, 46 },
	{ "fplautoroutebuttonm", GDK_KEY_m, GDK_KEY_M, 58 },
	{ "fplautoroutebuttonn", GDK_KEY_n, GDK_KEY_N, 57 },
	{ "fplautoroutebuttono", GDK_KEY_o, GDK_KEY_O, 32 },
	{ "fplautoroutebuttonp", GDK_KEY_p, GDK_KEY_P, 33 },
	{ "fplautoroutebuttonq", GDK_KEY_q, GDK_KEY_Q, 24 },
	{ "fplautoroutebuttonr", GDK_KEY_r, GDK_KEY_R, 27 },
	{ "fplautoroutebuttons", GDK_KEY_s, GDK_KEY_S, 39 },
	{ "fplautoroutebuttont", GDK_KEY_t, GDK_KEY_T, 28 },
	{ "fplautoroutebuttonu", GDK_KEY_u, GDK_KEY_U, 30 },
	{ "fplautoroutebuttonv", GDK_KEY_v, GDK_KEY_V, 55 },
	{ "fplautoroutebuttonw", GDK_KEY_w, GDK_KEY_W, 25 },
	{ "fplautoroutebuttonx", GDK_KEY_x, GDK_KEY_X, 53 },
	{ "fplautoroutebuttony", GDK_KEY_y, GDK_KEY_Y, 29 },
	{ "fplautoroutebuttonz", GDK_KEY_z, GDK_KEY_Z, 52 },
	{ "fplautoroutebuttonapostrophe", GDK_KEY_quoteright, GDK_KEY_quotedbl, 48 },
	{ "fplautoroutebuttonbracketclose", GDK_KEY_bracketright, GDK_KEY_braceright, 35 },
	{ "fplautoroutebuttonbracketopen", GDK_KEY_bracketleft, GDK_KEY_braceleft, 34 },
	{ "fplautoroutebuttoncomma", GDK_KEY_comma, GDK_KEY_less, 59 },
	{ "fplautoroutebuttondash", GDK_KEY_minus, GDK_KEY_underscore, 20 },
	{ "fplautoroutebuttondot", GDK_KEY_period, GDK_KEY_greater, 60 },
	{ "fplautoroutebuttonequal", GDK_KEY_equal, GDK_KEY_plus, 21 },
	{ "fplautoroutebuttonquestionmark", GDK_KEY_question, GDK_KEY_slash, 61 },
	{ "fplautoroutebuttonsemicolon", GDK_KEY_semicolon, GDK_KEY_colon, 47 },
	{ "fplautoroutebuttonspace", GDK_KEY_space, GDK_KEY_space, 65 },
	{ "fplautoroutebuttonbackslash", GDK_KEY_backslash, GDK_KEY_bar, 51 },
	{ "fplautoroutebuttonbackspace", GDK_KEY_BackSpace, GDK_KEY_BackSpace, 22 }
};

FlightDeckWindow::FPLAutoRouteDialog::FPLAutoRouteDialog(BaseObjectType *castitem, const Glib::RefPtr<Builder>& refxml)
	: Gtk::Window(castitem), m_sensors(0), m_fplautoroutenotebook(0),
	  m_fplautorouteentrydepicao(0), m_fplautorouteentrydepname(0), m_fplautoroutecheckbuttondepifr(0),
	  m_fplautoroutelabelsid(0), m_fplautorouteentrysidicao(0), m_fplautorouteentrysidname(0),
	  m_fplautorouteentrydesticao(0), m_fplautorouteentrydestname(0), m_fplautoroutecheckbuttondestifr(0),
	  m_fplautoroutelabelstar(0), m_fplautorouteentrystaricao(0), m_fplautorouteentrystarname(0),
	  m_fplautoroutespinbuttondctlimit(0), m_fplautoroutespinbuttondctpenalty(0),
	  m_fplautoroutespinbuttonvfraspclimit(0), m_fplautoroutecheckbuttonforceenrouteifr(0), m_fplautoroutecheckbuttonawylevels(0),
	  m_fplautoroutecheckbuttontfr(0), m_fplautorouteentrytracerules(0), m_fplautorouteentrydisablerules(0),
	  m_fplautorouteentryexclude(0), m_fplautoroutespinbuttonsidlimit(0), m_fplautoroutespinbuttonstarlimit(0),
	  m_fplautoroutespinbuttonbaselevel(0), m_fplautoroutespinbuttontoplevel(0), m_fplautoroutecomboboxopttarget(0),
	  m_fplautoroutebuttonpg1close(0), m_fplautoroutebuttonpg1copy(0), m_fplautoroutebuttonpg1next(0),
	  m_fplautoroutenotebookpg2(0), m_fplautoroutedrawingarea(0), m_fplautorouteentrypg2dep(0), m_fplautorouteentrypg2dest(0),
	  m_fplautorouteentrypg2gcdist(0), m_fplautorouteentrypg2routedist(0),m_fplautorouteentrypg2mintime(0),
	  m_fplautorouteentrypg2routetime(0), m_fplautorouteentrypg2minfuel(0),
	  m_fplautorouteentrypg2routefuel(0), m_fplautorouteentrypg2iteration(0),
	  m_fplautoroutetextviewpg2route(0), m_fplautoroutetextviewpg2results(0), m_fplautoroutetextviewpg2log(0),
	  m_fplautoroutetextviewpg2graph(0), m_fplautoroutetextviewpg2comm(0), m_fplautoroutetextviewpg2debug(0),
	  m_fplautoroutebuttonpg2zoomin(0), m_fplautoroutebuttonpg2zoomout(0), m_fplautoroutebuttonpg2back(0),
	  m_fplautoroutebuttonpg2close(0), m_fplautoroutebuttonpg2copy(0), m_fplautoroutetogglebuttonpg2run(0),
	  m_fplautoroutetogglebuttonshiftleft(0), m_fplautoroutetogglebuttonshiftright(0), m_fplautoroutetogglebuttoncapslock(0),
	  m_shiftstate(shiftstate_none), m_sidend(Point::invalid), m_starstart(Point::invalid),
	  m_depifr(true), m_destifr(true), m_firstrun(true), m_fplvalclear(true),
	  m_flags(MapRenderer::drawflags_route | MapRenderer::drawflags_route_labels), m_mapscale(0.1)
{
#ifndef GLIBMM_DEFAULT_SIGNAL_HANDLERS_ENABLED
        signal_delete_event().connect(sigc::mem_fun(*this, &FPLAutoRouteDialog::on_delete_event));
#endif //GLIBMM_DEFAULT_SIGNAL_HANDLERS_ENABLED
	refxml->get_widget("fplautoroutenotebook", m_fplautoroutenotebook);
	refxml->get_widget("fplautorouteentrydepicao", m_fplautorouteentrydepicao);
	refxml->get_widget("fplautorouteentrydepname", m_fplautorouteentrydepname);
	refxml->get_widget("fplautoroutecheckbuttondepifr", m_fplautoroutecheckbuttondepifr);
	refxml->get_widget("fplautoroutelabelsid", m_fplautoroutelabelsid);
	refxml->get_widget("fplautorouteentrysidicao", m_fplautorouteentrysidicao);
	refxml->get_widget("fplautorouteentrysidname", m_fplautorouteentrysidname);
	refxml->get_widget("fplautorouteentrydesticao", m_fplautorouteentrydesticao);
	refxml->get_widget("fplautorouteentrydestname", m_fplautorouteentrydestname);
	refxml->get_widget("fplautoroutecheckbuttondestifr", m_fplautoroutecheckbuttondestifr);
	refxml->get_widget("fplautoroutelabelstar", m_fplautoroutelabelstar);
	refxml->get_widget("fplautorouteentrystaricao", m_fplautorouteentrystaricao);
	refxml->get_widget("fplautorouteentrystarname", m_fplautorouteentrystarname);
	refxml->get_widget("fplautoroutespinbuttondctlimit", m_fplautoroutespinbuttondctlimit);
	refxml->get_widget("fplautoroutespinbuttondctpenalty", m_fplautoroutespinbuttondctpenalty);
	refxml->get_widget("fplautoroutespinbuttonvfraspclimit", m_fplautoroutespinbuttonvfraspclimit);
	refxml->get_widget("fplautoroutecheckbuttonforceenrouteifr", m_fplautoroutecheckbuttonforceenrouteifr);
	refxml->get_widget("fplautoroutecheckbuttonawylevels", m_fplautoroutecheckbuttonawylevels);
	refxml->get_widget("fplautoroutecheckbuttontfr", m_fplautoroutecheckbuttontfr);
	refxml->get_widget("fplautorouteentrytracerules", m_fplautorouteentrytracerules);
	refxml->get_widget("fplautorouteentrydisablerules", m_fplautorouteentrydisablerules);
	refxml->get_widget("fplautorouteentryexclude", m_fplautorouteentryexclude);
	refxml->get_widget("fplautoroutespinbuttonsidlimit", m_fplautoroutespinbuttonsidlimit);
	refxml->get_widget("fplautoroutespinbuttonstarlimit", m_fplautoroutespinbuttonstarlimit);
	refxml->get_widget("fplautoroutespinbuttonbaselevel", m_fplautoroutespinbuttonbaselevel);
	refxml->get_widget("fplautoroutespinbuttontoplevel", m_fplautoroutespinbuttontoplevel);
	refxml->get_widget("fplautoroutecomboboxopttarget", m_fplautoroutecomboboxopttarget);
	refxml->get_widget("fplautoroutebuttonpg1close", m_fplautoroutebuttonpg1close);
	refxml->get_widget("fplautoroutebuttonpg1copy", m_fplautoroutebuttonpg1copy);
	refxml->get_widget("fplautoroutebuttonpg1next", m_fplautoroutebuttonpg1next);
	refxml->get_widget("fplautoroutenotebookpg2", m_fplautoroutenotebookpg2);
	refxml->get_widget_derived("fplautoroutedrawingarea", m_fplautoroutedrawingarea);
	refxml->get_widget("fplautorouteentrypg2dep", m_fplautorouteentrypg2dep);
	refxml->get_widget("fplautorouteentrypg2dest", m_fplautorouteentrypg2dest);
	refxml->get_widget("fplautorouteentrypg2gcdist", m_fplautorouteentrypg2gcdist);
	refxml->get_widget("fplautorouteentrypg2routedist", m_fplautorouteentrypg2routedist);
	refxml->get_widget("fplautorouteentrypg2mintime", m_fplautorouteentrypg2mintime);
	refxml->get_widget("fplautorouteentrypg2routetime", m_fplautorouteentrypg2routetime);
	refxml->get_widget("fplautorouteentrypg2minfuel", m_fplautorouteentrypg2minfuel);
	refxml->get_widget("fplautorouteentrypg2routefuel", m_fplautorouteentrypg2routefuel);
	refxml->get_widget("fplautorouteentrypg2iteration",m_fplautorouteentrypg2iteration);
	refxml->get_widget("fplautoroutetextviewpg2route", m_fplautoroutetextviewpg2route);
	refxml->get_widget("fplautoroutetextviewpg2results", m_fplautoroutetextviewpg2results);
	refxml->get_widget("fplautoroutetextviewpg2log", m_fplautoroutetextviewpg2log);
	refxml->get_widget("fplautoroutetextviewpg2graph", m_fplautoroutetextviewpg2graph);
	refxml->get_widget("fplautoroutetextviewpg2comm", m_fplautoroutetextviewpg2comm);
	refxml->get_widget("fplautoroutetextviewpg2debug", m_fplautoroutetextviewpg2debug);
	refxml->get_widget("fplautoroutebuttonpg2zoomin", m_fplautoroutebuttonpg2zoomin);
	refxml->get_widget("fplautoroutebuttonpg2zoomout", m_fplautoroutebuttonpg2zoomout);
	refxml->get_widget("fplautoroutebuttonpg2back", m_fplautoroutebuttonpg2back);
	refxml->get_widget("fplautoroutebuttonpg2close", m_fplautoroutebuttonpg2close);
	refxml->get_widget("fplautoroutebuttonpg2copy", m_fplautoroutebuttonpg2copy);
	refxml->get_widget("fplautoroutetogglebuttonpg2run", m_fplautoroutetogglebuttonpg2run);
	refxml->get_widget("fplautoroutetogglebuttonshiftleft", m_fplautoroutetogglebuttonshiftleft);
	refxml->get_widget("fplautoroutetogglebuttonshiftright", m_fplautoroutetogglebuttonshiftright);
	refxml->get_widget("fplautoroutetogglebuttoncapslock", m_fplautoroutetogglebuttoncapslock);
	if (m_fplautoroutetogglebuttonshiftleft)
		m_connshiftleft = m_fplautoroutetogglebuttonshiftleft->signal_clicked().connect(sigc::mem_fun(*this, &FPLAutoRouteDialog::shift_clicked));
	if (m_fplautoroutetogglebuttonshiftright)
		m_connshiftright = m_fplautoroutetogglebuttonshiftright->signal_clicked().connect(sigc::mem_fun(*this, &FPLAutoRouteDialog::shift_clicked));
	if (m_fplautoroutetogglebuttoncapslock)
		m_conncapslock = m_fplautoroutetogglebuttoncapslock->signal_clicked().connect(sigc::mem_fun(*this, &FPLAutoRouteDialog::capslock_clicked));
	for (unsigned int i = 0; i < nrkeybuttons; ++i) {
		m_keybutton[i] = 0;
		refxml->get_widget(keybuttons[i].widget_name, m_keybutton[i]);
		if (!m_keybutton[i])
			continue;
		m_keybutton[i]->signal_clicked().connect(sigc::bind(sigc::mem_fun(*this, &FPLAutoRouteDialog::key_clicked), i));
		m_keybutton[i]->set_can_focus(false);
	}
	if (m_fplautoroutenotebook) {
		m_fplautoroutenotebook->set_current_page(0);
		m_fplautoroutenotebook->set_show_border(false);
		m_fplautoroutenotebook->set_show_tabs(false);
	}
	if (m_fplautoroutebuttonpg1close)
		m_fplautoroutebuttonpg1close->signal_clicked().connect(sigc::mem_fun(*this, &FPLAutoRouteDialog::pg1_close_clicked));
	if (m_fplautoroutebuttonpg1copy)
		m_fplautoroutebuttonpg1copy->signal_clicked().connect(sigc::mem_fun(*this, &FPLAutoRouteDialog::pg1_copy_clicked));
	if (m_fplautoroutebuttonpg1next)
		m_fplautoroutebuttonpg1next->signal_clicked().connect(sigc::mem_fun(*this, &FPLAutoRouteDialog::pg1_next_clicked));
	if (m_fplautoroutebuttonpg2zoomin)
		m_fplautoroutebuttonpg2zoomin->signal_clicked().connect(sigc::mem_fun(*this, &FPLAutoRouteDialog::pg2_zoomin_clicked));
	if (m_fplautoroutebuttonpg2zoomout)
		m_fplautoroutebuttonpg2zoomout->signal_clicked().connect(sigc::mem_fun(*this, &FPLAutoRouteDialog::pg2_zoomout_clicked));
	if (m_fplautoroutebuttonpg2back)
		m_fplautoroutebuttonpg2back->signal_clicked().connect(sigc::mem_fun(*this, &FPLAutoRouteDialog::pg2_back_clicked));
	if (m_fplautoroutebuttonpg2close)
		m_fplautoroutebuttonpg2close->signal_clicked().connect(sigc::mem_fun(*this, &FPLAutoRouteDialog::pg1_close_clicked));
	if (m_fplautoroutebuttonpg2copy)
		m_fplautoroutebuttonpg2copy->signal_clicked().connect(sigc::mem_fun(*this, &FPLAutoRouteDialog::pg1_copy_clicked));
	if (m_fplautoroutetogglebuttonpg2run)
		m_connpg2run = m_fplautoroutetogglebuttonpg2run->signal_clicked().connect(sigc::mem_fun(*this, &FPLAutoRouteDialog::pg2_run_clicked));
	if (m_fplautoroutecheckbuttondepifr)
		m_fplautoroutecheckbuttondepifr->signal_toggled().connect(sigc::mem_fun(*this, &FPLAutoRouteDialog::ifr_toggled));
	if (m_fplautoroutecheckbuttondestifr)
		m_fplautoroutecheckbuttondestifr->signal_toggled().connect(sigc::mem_fun(*this, &FPLAutoRouteDialog::ifr_toggled));
	if (m_fplautoroutecheckbuttondepifr)
		m_fplautoroutecheckbuttondepifr->set_active(true);
	if (m_fplautoroutecheckbuttondestifr)
		m_fplautoroutecheckbuttondestifr->set_active(true);
	m_sidstarlimit[0][0] = m_sidstarlimit[1][0] = 20;
	m_sidstarlimit[0][1] = m_sidstarlimit[1][1] = 40;
	if (m_fplautoroutespinbuttonsidlimit)
		m_fplautoroutespinbuttonsidlimit->set_value(m_sidstarlimit[0][m_depifr]);
	if (m_fplautoroutespinbuttonstarlimit)
		m_fplautoroutespinbuttonstarlimit->set_value(m_sidstarlimit[1][m_destifr]);
	for (unsigned int i = 0; i < sizeof(m_ptchanged)/sizeof(m_ptchanged[0]); ++i)
		m_ptchanged[i] = false;
	if (m_fplautorouteentrydepicao) {
		m_fplautorouteentrydepicao->signal_changed().connect(sigc::bind(sigc::mem_fun(*this, &FPLAutoRouteDialog::pt_changed), 0));
		m_fplautorouteentrydepicao->signal_focus_out_event().connect(sigc::mem_fun(*this, &FPLAutoRouteDialog::pt_focusout));
	}
	if (m_fplautorouteentrydesticao) {
		m_fplautorouteentrydesticao->signal_changed().connect(sigc::bind(sigc::mem_fun(*this, &FPLAutoRouteDialog::pt_changed), 1));
		m_fplautorouteentrydesticao->signal_focus_out_event().connect(sigc::mem_fun(*this, &FPLAutoRouteDialog::pt_focusout));
	}
	if (m_fplautorouteentrysidicao) {
		m_fplautorouteentrysidicao->signal_changed().connect(sigc::bind(sigc::mem_fun(*this, &FPLAutoRouteDialog::pt_changed), 2));
		m_fplautorouteentrysidicao->signal_focus_out_event().connect(sigc::mem_fun(*this, &FPLAutoRouteDialog::pt_focusout));
	}
	if (m_fplautorouteentrystaricao) {
		m_fplautorouteentrystaricao->signal_changed().connect(sigc::bind(sigc::mem_fun(*this, &FPLAutoRouteDialog::pt_changed), 3));
		m_fplautorouteentrystaricao->signal_focus_out_event().connect(sigc::mem_fun(*this, &FPLAutoRouteDialog::pt_focusout));
	}
	if (m_fplautoroutedrawingarea)
                m_fplautoroutedrawingarea->set_route(m_proxy.get_fplan());
 	m_proxy.signal_recvcmd().connect(sigc::mem_fun(*this, &FPLAutoRouteDialog::recvcmd));
	m_sidend.set_invalid();
	m_starstart.set_invalid();
        set_flags(MapRenderer::drawflags_route | MapRenderer::drawflags_route_labels | MapRenderer::drawflags_navaids |
		  /*MapRenderer::drawflags_waypoints | */MapRenderer::drawflags_airports);
	if (m_fplautoroutetextviewpg2graph) {
	    	Glib::RefPtr<Gtk::TextBuffer> buf(m_fplautoroutetextviewpg2graph->get_buffer());
		if (buf) {
			m_txtgraphfpl = buf->create_tag();
			m_txtgraphrule = buf->create_tag();
			m_txtgraphruledesc = buf->create_tag();
			m_txtgraphruleoprgoal = buf->create_tag();
			m_txtgraphchange = buf->create_tag();
			m_txtgraphfpl->property_weight() = 600;
			m_txtgraphfpl->property_weight_set() = true;
			m_txtgraphfpl->property_style() = Pango::STYLE_NORMAL;
			m_txtgraphfpl->property_style_set() = true;
			m_txtgraphrule->property_weight() = 400;
			m_txtgraphrule->property_weight_set() = true;
			m_txtgraphrule->property_style() = Pango::STYLE_NORMAL;
			m_txtgraphrule->property_style_set() = true;
			m_txtgraphruledesc->property_weight() = 300;
			m_txtgraphruledesc->property_weight_set() = true;
			m_txtgraphruledesc->property_style() = Pango::STYLE_NORMAL;
			m_txtgraphruledesc->property_style_set() = true;
			m_txtgraphruleoprgoal->property_weight() = 300;
			m_txtgraphruleoprgoal->property_weight_set() = true;
			m_txtgraphruleoprgoal->property_style() = Pango::STYLE_ITALIC;
			m_txtgraphruleoprgoal->property_style_set() = true;
			m_txtgraphchange->property_weight() = 400;
			m_txtgraphchange->property_weight_set() = true;
			m_txtgraphchange->property_style() = Pango::STYLE_ITALIC;
			m_txtgraphchange->property_style_set() = true;
		}
	}
}

FlightDeckWindow::FPLAutoRouteDialog::~FPLAutoRouteDialog()
{
}

bool FlightDeckWindow::FPLAutoRouteDialog::on_delete_event(GdkEventAny* event)
{
        close();
        return true;
}

void FlightDeckWindow::FPLAutoRouteDialog::shift_clicked(void)
{
	m_shiftstate = (m_shiftstate == shiftstate_once) ? shiftstate_none : shiftstate_once;
	set_shiftstate();
}

void FlightDeckWindow::FPLAutoRouteDialog::capslock_clicked(void)
{
	m_shiftstate = (m_shiftstate == shiftstate_lock) ? shiftstate_none : shiftstate_lock;
	set_shiftstate();
}

void FlightDeckWindow::FPLAutoRouteDialog::set_shiftstate(void)
{
	m_connshiftleft.block();
	m_connshiftright.block();
	m_conncapslock.block();
	if (m_fplautoroutetogglebuttonshiftleft)
		m_fplautoroutetogglebuttonshiftleft->set_active(m_shiftstate == shiftstate_once);
	if (m_fplautoroutetogglebuttonshiftright)
		m_fplautoroutetogglebuttonshiftright->set_active(m_shiftstate == shiftstate_once);
	if (m_fplautoroutetogglebuttoncapslock)
		m_fplautoroutetogglebuttoncapslock->set_active(m_shiftstate == shiftstate_lock);
	m_connshiftleft.unblock();
	m_connshiftright.unblock();
	m_conncapslock.unblock();
}

void FlightDeckWindow::FPLAutoRouteDialog::key_clicked(unsigned int i)
{
 	if (i >= nrkeybuttons)
		return;
	guint keyval((m_shiftstate != shiftstate_none) ? keybuttons[i].key_shift : keybuttons[i].key);
	if (false)
		std::cerr << "key_clicked: " << keyval << std::endl;
        Gtk::Widget *w(get_focus());
        if (!w)
                return;
	KeyboardDialog::send_key(w, ((m_shiftstate == shiftstate_once) ? GDK_SHIFT_MASK : 0) |
				 ((m_shiftstate == shiftstate_lock) ? GDK_LOCK_MASK : 0), keyval, keybuttons[i].hwkeycode);
	if (m_shiftstate == shiftstate_once) {
		m_shiftstate = shiftstate_none;
		set_shiftstate();
	}
}

void FlightDeckWindow::FPLAutoRouteDialog::set_flags(MapRenderer::DrawFlags flags)
{
        m_flags = flags;
        if (m_fplautoroutedrawingarea)
                *m_fplautoroutedrawingarea = m_flags;
}

void FlightDeckWindow::FPLAutoRouteDialog::set_map_scale(float scale)
{
        m_mapscale = scale;
        if (!m_fplautoroutedrawingarea)
                return;
	m_fplautoroutedrawingarea->set_map_scale(m_mapscale);
	m_mapscale = m_fplautoroutedrawingarea->get_map_scale();
}

void FlightDeckWindow::FPLAutoRouteDialog::pg2_zoomin_clicked(void)
{
        if (!m_fplautoroutedrawingarea)
                return;
	m_fplautoroutedrawingarea->zoom_in();
	m_mapscale = m_fplautoroutedrawingarea->get_map_scale();
}

void FlightDeckWindow::FPLAutoRouteDialog::pg2_zoomout_clicked(void)
{
        if (!m_fplautoroutedrawingarea)
                return;
	m_fplautoroutedrawingarea->zoom_out();
	m_mapscale = m_fplautoroutedrawingarea->get_map_scale();
}

void FlightDeckWindow::FPLAutoRouteDialog::open(Sensors *sensors)
{
	m_sensors = sensors;
	if (!m_sensors) {
		close();
		return;
	}
	if (m_fplautoroutedrawingarea)
                m_fplautoroutedrawingarea->set_engine(*m_sensors->get_engine());
	m_departure = m_destination = AirportsDb::Airport();
       	m_sidend = m_starstart = Point::invalid;
	if (m_fplautoroutecheckbuttondepifr)
		m_fplautoroutecheckbuttondepifr->set_active(true);
	if (m_fplautoroutecheckbuttondestifr)
		m_fplautoroutecheckbuttondestifr->set_active(true);
	if (m_fplautorouteentrydestname)
		m_fplautorouteentrydestname->set_text("");
	if (m_fplautorouteentrydepname)
		m_fplautorouteentrydepname->set_text("");
	if (m_fplautorouteentrysidname)
		m_fplautorouteentrysidname->set_text("");
	if (m_fplautorouteentrystarname)
		m_fplautorouteentrystarname->set_text("");
	if (!m_fplautoroutenotebook) {
		close();
		return;
	}
	if (m_fplautorouteentrydepicao)
		m_fplautorouteentrydepicao->set_text("");
	if (m_fplautorouteentrydepname)
		m_fplautorouteentrydepname->set_text("");
	if (m_fplautorouteentrydesticao)
		m_fplautorouteentrydesticao->set_text("");
	if (m_fplautorouteentrydestname)
		m_fplautorouteentrydestname->set_text("");
	if (m_fplautoroutecheckbuttondepifr)
		m_fplautoroutecheckbuttondepifr->set_active(true);
	if (m_fplautoroutecheckbuttondestifr)
		m_fplautoroutecheckbuttondestifr->set_active(true);
	if (m_fplautorouteentrysidicao)
		m_fplautorouteentrysidicao->set_text("");
	if (m_fplautorouteentrysidname)
		m_fplautorouteentrysidname->set_text("");
	if (m_fplautorouteentrystaricao)
		m_fplautorouteentrystaricao->set_text("");
	if (m_fplautorouteentrystarname)
		m_fplautorouteentrystarname->set_text("");
	m_fplautoroutenotebook->set_current_page(0);
	m_proxy.start_autorouter();
	m_proxy.clear_fplan();
	try {
		if (m_firstrun) {
			m_firstrun = false;
			{
				CFMUAutorouteProxy::Command cmd("departure");
				cmd.set_option("cmdseq", "init");
				m_proxy.sendcmd(cmd);
			}
			{
				CFMUAutorouteProxy::Command cmd("destination");
				cmd.set_option("cmdseq", "init");
				m_proxy.sendcmd(cmd);
			}
			{
				CFMUAutorouteProxy::Command cmd("enroute");
				cmd.set_option("cmdseq", "init");
				m_proxy.sendcmd(cmd);
			}
			{
				CFMUAutorouteProxy::Command cmd("levels");
				cmd.set_option("cmdseq", "init");
				m_proxy.sendcmd(cmd);
			}
			{
				CFMUAutorouteProxy::Command cmd("tfr");
				cmd.set_option("cmdseq", "init");
				m_proxy.sendcmd(cmd);
			}
			if (m_fplautoroutecheckbuttontfr)
				m_fplautoroutecheckbuttontfr->set_sensitive(false);
			{
				CFMUAutorouteProxy::Command cmd("preload");
				cmd.set_option("cmdseq", "init");
				m_proxy.sendcmd(cmd);
			}
		}
	} catch (const std::exception& e) {
		append_log(std::string("Exception: ") + e.what() + "\n");
	}
	{
		const FPlanRoute& route(m_sensors->get_route());
		for (unsigned int i = 0; i < 2; ++i) {
			if (route.get_nrwpt() <= i)
				break;
			IcaoFlightPlan::FindCoord f(*m_sensors->get_engine());
			const FPlanWaypoint& wpt(route[(route.get_nrwpt() - 1) & -i]);
			AirportsDb::Airport el;
			bool elok(false);
			if (f.find(wpt.get_coord(), IcaoFlightPlan::FindCoord::flag_airport, 0.1f))
				elok = f.get_airport(el);
			if (!elok || !el.is_valid())
				continue;
			if (i) {
				m_destination = el;
				if (m_destination.is_valid()) {
					if (m_fplautorouteentrydesticao)
						m_fplautorouteentrydesticao->set_text(m_destination.get_icao());
					if (m_fplautorouteentrydestname)
						m_fplautorouteentrydestname->set_text(m_destination.get_name());
				}
				if (m_fplautoroutecheckbuttondestifr)
					m_fplautoroutecheckbuttondestifr->set_active(!!(wpt.get_flags() & FPlanWaypoint::altflag_ifr));
			} else {
				m_departure = el;
				if (m_departure.is_valid()) {
					if (m_fplautorouteentrydepicao)
						m_fplautorouteentrydepicao->set_text(m_departure.get_icao());
					if (m_fplautorouteentrydepname)
						m_fplautorouteentrydepname->set_text(m_departure.get_name());
				}
				if (m_fplautoroutecheckbuttondepifr)
					m_fplautoroutecheckbuttondepifr->set_active(!!(wpt.get_flags() & FPlanWaypoint::altflag_ifr));
			}
		}
		if (route.get_nrwpt() >= 3) {
			for (unsigned int i = 0; i < 2; ++i) {
				Point pt;
				if (i) {
					unsigned int j(route.get_nrwpt() - 2);
					while (j > 1) {
						if (route[j - 1].get_pathcode() != FPlanWaypoint::pathcode_star)
							break;
						--j;
					}
					pt = route[j].get_coord();
				} else {
					unsigned int j(1);
					while (j + 2 < route.get_nrwpt()) {
						if (route[j].get_pathcode() != FPlanWaypoint::pathcode_sid)
							break;
						++j;
					}
					pt = route[j].get_coord();
				}
				if (route.get_nrwpt() == 3) {
					bool ok(pt.simple_distance_rel(route[0].get_coord()) <
						pt.simple_distance_rel(route[route.get_nrwpt() - 1].get_coord()));
					ok ^= i;
					if (!ok)
						continue;
				}
				Glib::ustring pticao, ptname;
				IcaoFlightPlan::FindCoord f(*m_sensors->get_engine());
				if (f.find(pt, IcaoFlightPlan::FindCoord::flag_navaid | IcaoFlightPlan::FindCoord::flag_waypoint, 0.1f)) {
					pt = Point::invalid;
					NavaidsDb::element_t nav;
					WaypointsDb::element_t wpt;
					if (f.get_navaid(nav)) {
						pt = nav.get_coord();
						ptname = nav.get_navaid_typename() + " " + nav.get_name();
						pticao = nav.get_icao();
					} else if (f.get_waypoint(wpt)) {
						pt = wpt.get_coord();
						ptname = "INT " + wpt.get_name();
						pticao = wpt.get_name();
					}
				} else {
					pt = Point::invalid;
				}
				if (!i) {
					m_sidend = pt;
					if (m_fplautorouteentrysidicao)
						m_fplautorouteentrysidicao->set_text(pticao);
					if (m_fplautorouteentrysidname)
						m_fplautorouteentrysidname->set_text(ptname);
				} else {
					m_starstart = pt;
					if (m_fplautorouteentrystaricao)
						m_fplautorouteentrystaricao->set_text(pticao);
					if (m_fplautorouteentrystarname)
						m_fplautorouteentrystarname->set_text(ptname);
				}
			}
		}
	}
	for (unsigned int i = 0; i < sizeof(m_ptchanged)/sizeof(m_ptchanged[0]); ++i)
		m_ptchanged[i] = false;
	pt_focusout();
	initial_fplan();
	show();
	if (m_fplautoroutedrawingarea) {
                m_fplautoroutedrawingarea->show();
		m_fplautoroutedrawingarea->set_renderer(VectorMapArea::renderer_vmap);
                *m_fplautoroutedrawingarea = m_flags;
                m_fplautoroutedrawingarea->set_map_scale(m_mapscale);
		m_fplautoroutedrawingarea->set_route(m_proxy.get_fplan());
 	}
	if (m_fplautoroutetextviewpg2route) {
	    	Glib::RefPtr<Gtk::TextBuffer> buf(m_fplautoroutetextviewpg2route->get_buffer());
		if (buf)
			buf->set_text("");
	}
	if (m_fplautoroutetextviewpg2results) {
	    	Glib::RefPtr<Gtk::TextBuffer> buf(m_fplautoroutetextviewpg2results->get_buffer());
		if (buf)
			buf->set_text("");
	}
	if (m_fplautoroutetextviewpg2log) {
	    	Glib::RefPtr<Gtk::TextBuffer> buf(m_fplautoroutetextviewpg2log->get_buffer());
		if (buf)
			buf->set_text("");
	}
	if (m_fplautoroutetextviewpg2graph) {
	    	Glib::RefPtr<Gtk::TextBuffer> buf(m_fplautoroutetextviewpg2graph->get_buffer());
		if (buf)
			buf->set_text("");
	}
	if (m_fplautoroutetextviewpg2comm) {
	    	Glib::RefPtr<Gtk::TextBuffer> buf(m_fplautoroutetextviewpg2comm->get_buffer());
		if (buf)
			buf->set_text("");
	}
	if (m_fplautoroutetextviewpg2debug) {
	    	Glib::RefPtr<Gtk::TextBuffer> buf(m_fplautoroutetextviewpg2debug->get_buffer());
		if (buf)
			buf->set_text("");
	}
}

void FlightDeckWindow::FPLAutoRouteDialog::close(void)
{
	if (m_fplautoroutedrawingarea)
                m_fplautoroutedrawingarea->hide();
	m_sensors = 0;
	hide();
	m_departure = m_destination = AirportsDb::Airport();
	m_proxy.stop_autorouter();
	m_proxy.clear_fplan();
	if (m_fplautoroutetextviewpg2route) {
	    	Glib::RefPtr<Gtk::TextBuffer> buf(m_fplautoroutetextviewpg2route->get_buffer());
		if (buf)
			buf->set_text("");
	}
	if (m_fplautoroutetextviewpg2results) {
	    	Glib::RefPtr<Gtk::TextBuffer> buf(m_fplautoroutetextviewpg2results->get_buffer());
		if (buf)
			buf->set_text("");
	}
	if (m_fplautoroutetextviewpg2log) {
	    	Glib::RefPtr<Gtk::TextBuffer> buf(m_fplautoroutetextviewpg2log->get_buffer());
		if (buf)
			buf->set_text("");
	}
	if (m_fplautoroutetextviewpg2graph) {
	    	Glib::RefPtr<Gtk::TextBuffer> buf(m_fplautoroutetextviewpg2graph->get_buffer());
		if (buf)
			buf->set_text("");
	}
	if (m_fplautoroutetextviewpg2comm) {
	    	Glib::RefPtr<Gtk::TextBuffer> buf(m_fplautoroutetextviewpg2comm->get_buffer());
		if (buf)
			buf->set_text("");
	}
	if (m_fplautoroutetextviewpg2debug) {
	    	Glib::RefPtr<Gtk::TextBuffer> buf(m_fplautoroutetextviewpg2debug->get_buffer());
		if (buf)
			buf->set_text("");
	}
}

void FlightDeckWindow::FPLAutoRouteDialog::append_log(const Glib::ustring& text)
{
	if (!m_fplautoroutetextviewpg2log)
		return;
	Glib::RefPtr<Gtk::TextBuffer> buf(m_fplautoroutetextviewpg2log->get_buffer());
	if (!buf)
		return;
	buf->insert(buf->end(), text);
	Gtk::TextBuffer::iterator it(buf->end());
	m_fplautoroutetextviewpg2log->scroll_to(it);
}

void FlightDeckWindow::FPLAutoRouteDialog::initial_fplan(void)
{
	m_proxy.clear_fplan();
	if (m_departure.is_valid()) {
		Engine::DbObject::insert(m_proxy.get_fplan(), m_departure);
		if (m_depifr) {
			FPlanWaypoint& wpt(m_proxy.get_fplan()[0]);
			wpt.set_flags(wpt.get_flags() | FPlanWaypoint::altflag_ifr);
		}
	}
	int alt(100);
	if (m_fplautoroutespinbuttonbaselevel)
		alt = m_fplautoroutespinbuttonbaselevel->get_value();
	alt *= 100;
	if (!m_sidend.is_invalid()) {
		FPlanWaypoint wpt;
		wpt.set_coord(m_sidend);
		wpt.set_name("SID");
		if (m_depifr)
			wpt.set_flags(wpt.get_flags() | FPlanWaypoint::altflag_ifr);
		wpt.set_flags(wpt.get_flags() | FPlanWaypoint::altflag_standard);
		wpt.set_altitude(alt);
		m_proxy.get_fplan().insert_wpt(~0U, wpt);
	}
	if (!m_starstart.is_invalid()) {
		FPlanWaypoint wpt;
		wpt.set_coord(m_starstart);
		wpt.set_name("STAR");
		if (m_destifr)
			wpt.set_flags(wpt.get_flags() | FPlanWaypoint::altflag_ifr);
		wpt.set_flags(wpt.get_flags() | FPlanWaypoint::altflag_standard);
		wpt.set_altitude(alt);
		m_proxy.get_fplan().insert_wpt(~0U, wpt);
	}
	if (m_destination.is_valid()) {
		Engine::DbObject::insert(m_proxy.get_fplan(), m_destination);
		if (m_destifr) {
			FPlanWaypoint& wpt(m_proxy.get_fplan()[m_proxy.get_fplan().get_nrwpt() - 1]);
			wpt.set_flags(wpt.get_flags() | FPlanWaypoint::altflag_ifr);
		}
	}
	if (m_fplautoroutedrawingarea)
		m_fplautoroutedrawingarea->redraw_route();
}

Rect FlightDeckWindow::FPLAutoRouteDialog::get_bbox(void) const
{
	if (!m_departure.is_valid() && !m_destination.is_valid())
		return Rect();
	if (!m_departure.is_valid())
		return Rect(m_destination.get_coord(), m_destination.get_coord());
	if (!m_destination.is_valid())
		return Rect(m_departure.get_coord(), m_departure.get_coord());
	Point pt1(m_departure.get_coord());
	Point pt2(m_destination.get_coord() - pt1);
	return Rect(Point(std::min(pt2.get_lon(), (Point::coord_t)0), std::min(pt2.get_lat(), (Point::coord_t)0)) + pt1,
		    Point(std::max(pt2.get_lon(), (Point::coord_t)0), std::max(pt2.get_lat(), (Point::coord_t)0)) + pt1);
}

void FlightDeckWindow::FPLAutoRouteDialog::recvcmd(const CFMUAutorouteProxy::Command& cmd)
{
	if (false)
		std::cerr << "Autorouter: " << cmd.get_cmdstring() << std::endl;
	if (m_fplautoroutetextviewpg2comm) {
	    	Glib::RefPtr<Gtk::TextBuffer> buf(m_fplautoroutetextviewpg2comm->get_buffer());
		if (buf) {
			buf->insert(buf->end(), "# " + cmd.get_cmdstring() + "\n");
			Gtk::TextBuffer::iterator it(buf->end());
			m_fplautoroutetextviewpg2comm->scroll_to(it);
		}
	}
	if (cmd.get_cmdname() == "autoroute") {
		std::pair<const std::string&,bool> status(cmd.get_option("status"));
		if (status.second) {
			if (status.first == "stopping" &&
			    m_fplautoroutetogglebuttonpg2run && m_fplautoroutetogglebuttonpg2run->get_active())
				m_fplautoroutetogglebuttonpg2run->set_active(false);
		}
		return;
	}
	if (cmd.get_cmdname() == "proxy") {
		std::pair<const std::string&,bool> state(cmd.get_option("state"));
		if (state.second) {
			if ((state.first == "stopped" || state.first == "error") &&
			    m_fplautoroutetogglebuttonpg2run && m_fplautoroutetogglebuttonpg2run->get_active())
				m_fplautoroutetogglebuttonpg2run->set_active(false);
		}
		return;
	}
	if (cmd.get_cmdname() == "log") {
		std::pair<const std::string&,bool> item(cmd.get_option("item"));
		std::pair<const std::string&,bool> text(cmd.get_option("text"));
		if (!item.second || !text.second)
			return;
		if (item.first == "fplvalidation") {
			if (!m_fplautoroutetextviewpg2results)
				return;
			Glib::RefPtr<Gtk::TextBuffer> buf(m_fplautoroutetextviewpg2results->get_buffer());
			if (!buf)
				return;
			if (m_fplvalclear)
				buf->set_text("");
			buf->insert(buf->end(), text.first + "\n");
			Gtk::TextBuffer::iterator it(buf->end());
			m_fplautoroutetextviewpg2results->scroll_to(it);
			m_fplvalclear = text.first.empty();
			return;
		}
		if (item.first == "graphchange") {
			if (m_fplautoroutetextviewpg2graph) {
				Glib::RefPtr<Gtk::TextBuffer> buf(m_fplautoroutetextviewpg2graph->get_buffer());
				if (buf) {
					buf->insert_with_tag(buf->end(), "  " + text.first + "\n", m_txtgraphchange);
					Gtk::TextBuffer::iterator it(buf->end());
					m_fplautoroutetextviewpg2graph->scroll_to(it);
				}
			}
			return;
		}
		if (item.first == "graphrule") {
			if (m_fplautoroutetextviewpg2graph) {
				Glib::RefPtr<Gtk::TextBuffer> buf(m_fplautoroutetextviewpg2graph->get_buffer());
				if (buf) {
					buf->insert_with_tag(buf->end(), text.first + "\n", m_txtgraphrule);
					Gtk::TextBuffer::iterator it(buf->end());
					m_fplautoroutetextviewpg2graph->scroll_to(it);
				}
			}
			return;
		}
		if (item.first == "graphruledesc") {
			if (m_fplautoroutetextviewpg2graph) {
				Glib::RefPtr<Gtk::TextBuffer> buf(m_fplautoroutetextviewpg2graph->get_buffer());
				if (buf) {
					buf->insert_with_tag(buf->end(), text.first + "\n", m_txtgraphruledesc);
					Gtk::TextBuffer::iterator it(buf->end());
					m_fplautoroutetextviewpg2graph->scroll_to(it);
				}
			}
			return;
		}
		if (item.first == "graphruleoprgoal") {
			if (m_fplautoroutetextviewpg2graph) {
				Glib::RefPtr<Gtk::TextBuffer> buf(m_fplautoroutetextviewpg2graph->get_buffer());
				if (buf) {
					buf->insert_with_tag(buf->end(), text.first + "\n", m_txtgraphruleoprgoal);
					Gtk::TextBuffer::iterator it(buf->end());
					m_fplautoroutetextviewpg2graph->scroll_to(it);
				}
			}
			return;
		}
		if (item.first.substr(0, 5) == "debug") {
			if (m_fplautoroutetextviewpg2debug) {
				Glib::RefPtr<Gtk::TextBuffer> buf(m_fplautoroutetextviewpg2debug->get_buffer());
				if (buf) {
					buf->insert(buf->end(), text.first + "\n");
					Gtk::TextBuffer::iterator it(buf->end());
					m_fplautoroutetextviewpg2debug->scroll_to(it);
				}
			}
			return;
		}
		append_log(text.first + "\n");
		return;
	}
	if (cmd.get_cmdname() == "fplend") {
		std::pair<const std::string&,bool> fpl(cmd.get_option("fpl"));
		if (fpl.second && m_fplautoroutetextviewpg2route) {
			Glib::RefPtr<Gtk::TextBuffer> buf(m_fplautoroutetextviewpg2route->get_buffer());
			if (buf)
				buf->set_text(fpl.first);
		}
		if (m_fplautoroutetextviewpg2graph) {
			Glib::RefPtr<Gtk::TextBuffer> buf(m_fplautoroutetextviewpg2graph->get_buffer());
			if (buf) {
				buf->insert_with_tag(buf->end(), fpl.first + "\n", m_txtgraphfpl);
				Gtk::TextBuffer::iterator it(buf->end());
				m_fplautoroutetextviewpg2graph->scroll_to(it);
			}
		}
		std::pair<double,bool> gcdist(cmd.get_option_float("gcdist"));
		std::pair<double,bool> routedist(cmd.get_option_float("routedist"));
		std::pair<unsigned long,bool> mintime(cmd.get_option_uint("mintime"));
		std::pair<unsigned long,bool> routetime(cmd.get_option_uint("routetime"));
		std::pair<double,bool> minfuel(cmd.get_option_float("minfuel"));
		std::pair<double,bool> routefuel(cmd.get_option_float("routefuel"));
		std::pair<const std::string&,bool> iter(cmd.get_option("iteration"));
		std::pair<const std::string&,bool> iterlocal(cmd.get_option("localiteration"));
		std::pair<const std::string&,bool> iterremote(cmd.get_option("remoteiteration"));
		if (gcdist.second && m_fplautorouteentrypg2gcdist)
			m_fplautorouteentrypg2gcdist->set_text(Conversions::dist_str(gcdist.first));
		if (routedist.second && m_fplautorouteentrypg2routedist)
			m_fplautorouteentrypg2routedist->set_text(Conversions::dist_str(routedist.first));
		if (mintime.second && m_fplautorouteentrypg2mintime)
			m_fplautorouteentrypg2mintime->set_text(Conversions::time_str(mintime.first));
		if (routetime.second && m_fplautorouteentrypg2routetime)
			m_fplautorouteentrypg2routetime->set_text(Conversions::time_str(routetime.first));
		if (minfuel.second && m_fplautorouteentrypg2minfuel)
			m_fplautorouteentrypg2minfuel->set_text(Conversions::dist_str(minfuel.first));
		if (routefuel.second && m_fplautorouteentrypg2routefuel)
			m_fplautorouteentrypg2routefuel->set_text(Conversions::dist_str(routefuel.first));
		if (m_fplautorouteentrypg2iteration) {
			std::ostringstream oss;
			if (iter.second)
				oss << iter.first;
			if (iterlocal.second && iterremote.second) {
				if (iter.second)
					oss << ' ';
				oss << '(' << iterlocal.first << '/' << iterremote.first << ')';
			}
			m_fplautorouteentrypg2iteration->set_text(oss.str());
		}
		if (m_fplautoroutedrawingarea)
			m_fplautoroutedrawingarea->redraw_route();
		return;
	}
	if (cmd.get_cmdname() == "departure") {
		std::pair<const std::string&,bool> cmdseq(cmd.get_option("cmdseq"));
		std::pair<const std::string&,bool> icao(cmd.get_option("icao"));
		std::pair<const std::string&,bool> name(cmd.get_option("name"));
		std::string icao_name;
		if (icao.second && !icao.first.empty())
			icao_name = icao.first;
		if (icao.second && !icao.first.empty() && name.second && !name.first.empty())
			icao_name += " ";
		if (name.second && !name.first.empty())
			icao_name += name.first;
		if (m_fplautorouteentrypg2dep)
			m_fplautorouteentrypg2dep->set_text(icao_name);
		if (!cmdseq.second || cmdseq.first != "init")
			return;
		std::pair<double,bool> sidlimit(cmd.get_option_float("sidlimit"));
		if (sidlimit.second) {
			m_sidstarlimit[0][1] = sidlimit.first;
			if (m_depifr && m_fplautoroutespinbuttonsidlimit)
				m_fplautoroutespinbuttonsidlimit->set_value(m_sidstarlimit[0][m_depifr]);
		}
		return;
	}
	if (cmd.get_cmdname() == "destination") {
		std::pair<const std::string&,bool> cmdseq(cmd.get_option("cmdseq"));
		std::pair<const std::string&,bool> icao(cmd.get_option("icao"));
		std::pair<const std::string&,bool> name(cmd.get_option("name"));
		std::string icao_name;
		if (icao.second && !icao.first.empty())
			icao_name = icao.first;
		if (icao.second && !icao.first.empty() && name.second && !name.first.empty())
			icao_name += " ";
		if (name.second && !name.first.empty())
			icao_name += name.first;
		if (m_fplautorouteentrypg2dest)
			m_fplautorouteentrypg2dest->set_text(icao_name);
		if (!cmdseq.second || cmdseq.first != "init")
			return;
		std::pair<double,bool> starlimit(cmd.get_option_float("starlimit"));
		if (starlimit.second) {
			m_sidstarlimit[1][1] = starlimit.first;
			if (m_destifr && m_fplautoroutespinbuttonstarlimit)
				m_fplautoroutespinbuttonstarlimit->set_value(m_sidstarlimit[1][m_destifr]);
		}
		return;
	}
	if (cmd.get_cmdname() == "enroute") {
		std::pair<const std::string&,bool> cmdseq(cmd.get_option("cmdseq"));
		if (!cmdseq.second || cmdseq.first != "init")
			return;
		std::pair<double,bool> dctlimit(cmd.get_option_float("dctlimit"));
		if (dctlimit.second && m_fplautoroutespinbuttondctlimit)
			m_fplautoroutespinbuttondctlimit->set_value(dctlimit.first);
		std::pair<double,bool> dctpenalty(cmd.get_option_float("dctpenalty"));
		if (dctpenalty.second && m_fplautoroutespinbuttondctpenalty)
			m_fplautoroutespinbuttondctpenalty->set_value(dctpenalty.first);
		std::pair<double,bool> vfraspclimit(cmd.get_option_float("vfraspclimit"));
		if (vfraspclimit.second && m_fplautoroutespinbuttonvfraspclimit)
			m_fplautoroutespinbuttonvfraspclimit->set_value(vfraspclimit.first);
		std::pair<unsigned long,bool> forceenrouteifr(cmd.get_option_uint("forceenrouteifr"));
		if (forceenrouteifr.second && m_fplautoroutecheckbuttonforceenrouteifr)
			m_fplautoroutecheckbuttonforceenrouteifr->set_active(!!forceenrouteifr.first);
		std::pair<unsigned long,bool> honourawylevels(cmd.get_option_uint("honourawylevels"));
		if (honourawylevels.second && m_fplautoroutecheckbuttonawylevels)
			m_fplautoroutecheckbuttonawylevels->set_active(!!honourawylevels.first);
		return;
	}
	if (cmd.get_cmdname() == "tfr") {
		std::pair<unsigned long,bool> available(cmd.get_option_uint("available"));
		std::pair<unsigned long,bool> enabled(cmd.get_option_uint("enabled"));
		std::pair<const std::string&,bool> trace(cmd.get_option("trace"));
		std::pair<const std::string&,bool> disable(cmd.get_option("disable"));
		if (available.second && m_fplautoroutecheckbuttontfr)
			m_fplautoroutecheckbuttontfr->set_sensitive(!!available.first);
		if (enabled.second && m_fplautoroutecheckbuttontfr)
			m_fplautoroutecheckbuttontfr->set_active(!!enabled.first);
		if (trace.second && m_fplautorouteentrytracerules)
			m_fplautorouteentrytracerules->set_text(trace.first);
		if (disable.second && m_fplautorouteentrydisablerules)
			m_fplautorouteentrydisablerules->set_text(disable.first);
		return;
	}
	if (cmd.get_cmdname() == "levels") {
		std::pair<const std::string&,bool> cmdseq(cmd.get_option("cmdseq"));
		if (!cmdseq.second || cmdseq.first != "init")
			return;
		std::pair<long,bool> base(cmd.get_option_int("base"));
		if (base.second && m_fplautoroutespinbuttonbaselevel)
			m_fplautoroutespinbuttonbaselevel->set_value(base.first);
		std::pair<long,bool> top(cmd.get_option_int("top"));
		if (top.second && m_fplautoroutespinbuttontoplevel)
			m_fplautoroutespinbuttontoplevel->set_value(top.first);
		return;
	}
}

void FlightDeckWindow::FPLAutoRouteDialog::pg1_close_clicked(void)
{
	close();
}

void FlightDeckWindow::FPLAutoRouteDialog::pg1_copy_clicked(void)
{
	const FPlanRoute& route2(m_proxy.get_fplan());
       	if (!m_sensors || !route2.get_nrwpt()) {
		close();
		return;
	}
	// read airway graph to add intermediate non-flightplannable waypoints
	if (false) {
		Engine::AirwayGraphResult::Graph areagraph;
		IcaoFlightPlan::FindCoord f(*m_sensors->get_engine());
		f.find_area(route2.get_bbox().oversize_nmi(200.f),
			    Engine::AirwayGraphResult::Vertex::typemask_awyendpoints |
			    Engine::AirwayGraphResult::Vertex::typemask_airway);
		f.get_airwaygraph(areagraph);
		areagraph.delete_disconnected_vertices();
	}
	FPlanRoute& route(m_sensors->get_route());
	while (route.get_nrwpt() > 0)
		m_sensors->nav_delete_wpt(0);
	time_t toffs(route2[0].get_time_unix());
	{
		time_t curt(time(0));
		if (toffs >= curt) {
			toffs = 0;
		} else {
			curt /= 24 * 60 * 60;
			curt *= 24 * 60 * 60;
			curt += toffs % (24 * 60 * 60);
			toffs = curt - toffs;
			if (toffs < 0)
				toffs += 24 * 60 * 60;
			struct tm utm;
			gmtime_r(&toffs, &utm);
			utm.tm_hour = utm.tm_min = utm.tm_sec = 0;
			toffs = timegm(&utm);
		}
	}
	typedef std::map<Glib::ustring, Engine::AirwayGraphResult::Graph> awygraph_t;
	awygraph_t awygraph;
	for (unsigned int i = 0; i < route2.get_nrwpt(); ++i) {
		FPlanWaypoint wpt(route2[i]);
		wpt.set_time_unix(wpt.get_time_unix() + toffs);
		m_sensors->nav_insert_wpt(~0U, wpt);
		if (!i) {
			if (!(wpt.get_flags() & FPlanWaypoint::altflag_ifr))
				continue;
			if (false)
				std::cerr << "SID: " << wpt.get_icao() << " dep " << m_departure.get_icao() << std::endl;
			if (wpt.get_icao() != m_departure.get_icao())
				continue;
			const FPlanWaypoint& wpt1(route2[i + 1]);
			for (unsigned int j = 0, k = m_departure.get_nrvfrrte(); j < k; ++j) {
				const AirportsDb::Airport::VFRRoute& rte(m_departure.get_vfrrte(j));
				if (rte.size() < 2 || rte.get_name().empty())
					continue;
				unsigned int m(rte.size()-1);
				const AirportsDb::Airport::VFRRoute::VFRRoutePoint& rtept(rte[m]);
				if (rtept.get_pathcode() != AirportsDb::Airport::VFRRoute::VFRRoutePoint::pathcode_sid)
					continue;
				const AirportsDb::Airport::VFRRoute::VFRRoutePoint& rtepta(rte[0]);
				if (false)
					std::cerr << "SID: " << rte.get_name()
						  << " aprt dist " << rtepta.get_coord().spheric_distance_nmi_dbl(m_departure.get_coord())
						  << " end dist " << rtept.get_coord().spheric_distance_nmi_dbl(wpt1.get_coord()) << std::endl;
				if (rtept.get_coord().spheric_distance_nmi_dbl(wpt1.get_coord()) > 0.1)
					continue;
				if (rtepta.get_coord().spheric_distance_nmi_dbl(m_departure.get_coord()) > 3.0)
					continue;
				{
					FPlanRoute& route(m_sensors->get_route());
					FPlanWaypoint& wpt(route[route.get_nrwpt() - 1]);
					wpt.set_pathname(rte.get_name());
					wpt.set_pathcode(rtepta.get_fplan_pathcode());					
				}
				for (unsigned int l = 1; l < m; ++l) {
					FPlanWaypoint wpt2;
					const AirportsDb::Airport::VFRRoute::VFRRoutePoint& vfrpt(rte[l]);
					//wpt2.set_icao(m_departure.get_icao());
					wpt2.set_icao("");
					wpt2.set_name(vfrpt.get_name());
					wpt2.set_coord(vfrpt.get_coord());
					wpt2.set_flags(wpt1.get_flags() & ~FPlanWaypoint::altflag_standard);
					wpt2.set_altitude(vfrpt.get_altitude());
					if (vfrpt.get_altitude() >= wpt1.get_altitude()) {
						wpt2.set_flags(wpt1.get_flags());
						wpt2.set_altitude(wpt1.get_altitude());
					}
					wpt2.set_frequency(0);
					Glib::ustring note(vfrpt.get_altcode_string());
					if (vfrpt.is_compulsory())
						note += " Compulsory";
					note += "\n" + vfrpt.get_pathcode_string();
					wpt2.set_note(note);
					wpt2.set_pathcode(vfrpt.get_fplan_pathcode());
					wpt2.set_pathname(rte.get_name());
					m_sensors->nav_insert_wpt(~0U, wpt2);
				}
				break;
			}
			continue;
		}
		if (i + 2 == route2.get_nrwpt()) {
			const FPlanWaypoint& wpt1(route2[i + 1]);
			if (!(wpt1.get_flags() & FPlanWaypoint::altflag_ifr))
				continue;
			if (false)
				std::cerr << "STAR: " << wpt1.get_icao() << " dest " << m_destination.get_icao() << std::endl;
			if (wpt1.get_icao() != m_destination.get_icao())
				continue;
			for (unsigned int j = 0, k = m_destination.get_nrvfrrte(); j < k; ++j) {
				const AirportsDb::Airport::VFRRoute& rte(m_destination.get_vfrrte(j));
				if (rte.size() < 2 || rte.get_name().empty())
					continue;
				unsigned int m(rte.size()-1);
				const AirportsDb::Airport::VFRRoute::VFRRoutePoint& rtept(rte[0]);
				if (rtept.get_pathcode() != AirportsDb::Airport::VFRRoute::VFRRoutePoint::pathcode_star)
					continue;
				const AirportsDb::Airport::VFRRoute::VFRRoutePoint& rtepta(rte[m]);
				if (false)
					std::cerr << "STAR: " << rte.get_name()
						  << " aprt dist " << rtepta.get_coord().spheric_distance_nmi_dbl(m_departure.get_coord())
						  << " end dist " << rtept.get_coord().spheric_distance_nmi_dbl(wpt1.get_coord()) << std::endl;
				if (rtept.get_coord().spheric_distance_nmi_dbl(wpt.get_coord()) > 0.1)
					continue;
				if (rtepta.get_coord().spheric_distance_nmi_dbl(m_destination.get_coord()) > 3.0)
					continue;
				{
					FPlanRoute& route(m_sensors->get_route());
					FPlanWaypoint& wpt(route[route.get_nrwpt() - 1]);
					wpt.set_pathname(rte.get_name());
					wpt.set_pathcode(rtepta.get_fplan_pathcode());					
				}
				for (unsigned int l = 1; l < m; ++l) {
					FPlanWaypoint wpt2;
					const AirportsDb::Airport::VFRRoute::VFRRoutePoint& vfrpt(rte[l]);
					//wpt2.set_icao(m_destination.get_icao());
					wpt2.set_icao("");
					wpt2.set_name(vfrpt.get_name());
					wpt2.set_coord(vfrpt.get_coord());
					wpt2.set_flags(wpt.get_flags() & ~FPlanWaypoint::altflag_standard);
					wpt2.set_altitude(vfrpt.get_altitude());
					if (vfrpt.get_altitude() >= wpt.get_altitude()) {
						wpt2.set_flags(wpt.get_flags());
						wpt2.set_altitude(wpt.get_altitude());
					}
					wpt2.set_frequency(0);
					Glib::ustring note(vfrpt.get_altcode_string());
					if (vfrpt.is_compulsory())
						note += " Compulsory";
					note += "\n" + vfrpt.get_pathcode_string();
					wpt2.set_note(note);
					wpt2.set_pathcode(vfrpt.get_fplan_pathcode());
					wpt2.set_pathname(rte.get_name());
					m_sensors->nav_insert_wpt(~0U, wpt2);
				}
				break;
			}
			continue;
		}
		// try to find an intermediate airway point
		if (wpt.get_pathcode() == FPlanWaypoint::pathcode_airway && !wpt.get_pathname().empty()) {
			if (!(wpt.get_flags() & FPlanWaypoint::altflag_ifr))
				continue;
			awygraph_t::iterator agi(awygraph.find(wpt.get_pathname()));
			if (agi == awygraph.end()) {
				std::pair<awygraph_t::iterator,bool> ins(awygraph.insert(std::make_pair(wpt.get_pathname(),
													Engine::AirwayGraphResult::Graph())));
				agi = ins.first;
				IcaoFlightPlan::FindCoord f(*m_sensors->get_engine());
				if (false)
					std::cerr << "Loading airway graph " << wpt.get_pathname() << std::endl;
				f.find_airway(wpt.get_pathname());
				f.get_airwaygraph(ins.first->second);
			}
			Engine::AirwayGraphResult::Graph& agraph(agi->second);
			Engine::AirwayGraphResult::Graph::vertex_descriptor u0;
			if (!agraph.find_nearest(u0, wpt.get_coord(), wpt.get_pathname(),
						 Engine::AirwayGraphResult::Vertex::typemask_navaid |
						 Engine::AirwayGraphResult::Vertex::typemask_intersection |
						 Engine::AirwayGraphResult::Vertex::typemask_undefined,
						 true, false))
				continue;
			const FPlanWaypoint& wpt1(route2[i + 1]);
			const Engine::AirwayGraphResult::Vertex& v0(agraph[u0]);
			if (v0.get_ident() != wpt.get_ident() ||
			    wpt.get_coord().spheric_distance_nmi(v0.get_coord()) > 0.1)
				continue;
			Engine::AirwayGraphResult::Graph::vertex_descriptor u1;
			if (!agraph.find_nearest(u1, wpt1.get_coord(), wpt.get_pathname(),
						 Engine::AirwayGraphResult::Vertex::typemask_navaid |
						 Engine::AirwayGraphResult::Vertex::typemask_intersection |
						 Engine::AirwayGraphResult::Vertex::typemask_undefined,
						 false, true))
				continue;
			if (u0 == u1)
				continue;
			const Engine::AirwayGraphResult::Vertex& v1(agraph[u1]);
			if (v1.get_ident() != wpt1.get_ident() ||
			    wpt1.get_coord().spheric_distance_nmi(v1.get_coord()) > 0.1)
				continue;
			std::vector<Engine::AirwayGraphResult::Graph::vertex_descriptor> predecessors;
			agraph.shortest_paths(u0, predecessors);
			Engine::AirwayGraphResult::Graph::vertex_descriptor uu(predecessors[u1]);
			if (uu == u1)
				continue;
			unsigned int inspt(route.get_nrwpt());
			for (; uu != u0; uu = predecessors[uu]) {
				const Engine::AirwayGraphResult::Vertex& vv(agraph[uu]);
				FPlanWaypoint wpt2;
				if (!vv.set(wpt2)) {
					wpt2.set_name(vv.get_ident());
					wpt2.set_coord(vv.get_coord());
				}
				wpt2.set_pathcode(wpt.get_pathcode());
				wpt2.set_pathname(wpt.get_pathname());
				wpt2.set_time(wpt.get_time());
				wpt2.set_altitude(wpt.get_altitude());
				wpt2.set_flags(wpt.get_flags());
				wpt2.set_winddir(wpt.get_winddir());
				wpt2.set_windspeed(wpt.get_windspeed());
				wpt2.set_qff(wpt.get_qff());
				wpt2.set_isaoffset(wpt.get_isaoffset());
				m_sensors->nav_insert_wpt(inspt, wpt2);
				if (false)
					std::cerr << "Inserting " << wpt.get_name() << std::endl;
			}
                }
	}
	//route.set_time_offblock_unix(route2[0].get_time_unix() + toffs - 5 * 60);
	//route.set_time_onblock_unix(route2[route2.get_nrwpt() - 1].get_time_unix() + toffs + 5 * 60);
	//m_sensors->nav_fplan_modified();
	m_sensors->nav_fplan_setdeparture(route2[0].get_time_unix() + toffs);
	close();
}

void FlightDeckWindow::FPLAutoRouteDialog::pg1_next_clicked(void)
{
	if (!m_fplautoroutenotebook) {
		close();
		return;
	}
	if (m_fplautorouteentrypg2dep)
		m_fplautorouteentrypg2dep->set_text("");
	if (m_fplautorouteentrypg2dest)
		m_fplautorouteentrypg2dest->set_text("");
	if (m_fplautorouteentrypg2gcdist)
		m_fplautorouteentrypg2gcdist->set_text("");
	if (m_fplautorouteentrypg2routedist)
		m_fplautorouteentrypg2routedist->set_text("");
	if (m_fplautorouteentrypg2mintime)
		m_fplautorouteentrypg2mintime->set_text("");
	if (m_fplautorouteentrypg2routetime)
		m_fplautorouteentrypg2routetime->set_text("");
	if (m_fplautorouteentrypg2minfuel)
		m_fplautorouteentrypg2minfuel->set_text("");
	if (m_fplautorouteentrypg2routefuel)
		m_fplautorouteentrypg2routefuel->set_text("");
	if (m_fplautorouteentrypg2iteration)
		m_fplautorouteentrypg2iteration->set_text("");
	m_fplautoroutenotebook->set_current_page(1);
	if (m_fplautoroutenotebookpg2)
		m_fplautoroutenotebookpg2->set_current_page(0);
      	try {
		if (m_sensors) {
			CFMUAutorouteProxy::Command cmd("aircraft");
			if (false && !m_sensors->get_aircraftfile().empty())
				cmd.set_option("file", m_sensors->get_aircraftfile());
			else
				cmd.set_option("data", m_sensors->get_aircraft().save_string());
			m_proxy.sendcmd(cmd);
		}
		{
			CFMUAutorouteProxy::Command cmd("clear");
			m_proxy.sendcmd(cmd);
		}
		if (m_departure.is_valid()) {
			CFMUAutorouteProxy::Command cmd("departure");
			cmd.set_option("icao", m_departure.get_icao());
			cmd.set_option("name", m_departure.get_name());
			cmd.set_option("coord", m_departure.get_coord());
			cmd.set_option(m_depifr ? "ifr" : "vfr", "");
			cmd.set_option("sid", m_sidend);
			if (m_fplautoroutespinbuttonsidlimit)
				cmd.set_option("sidlimit", m_fplautoroutespinbuttonsidlimit->get_value());
			m_proxy.sendcmd(cmd);
		}
		if (m_destination.is_valid()) {
			CFMUAutorouteProxy::Command cmd("destination");
			cmd.set_option("icao", m_destination.get_icao());
			cmd.set_option("name", m_destination.get_name());
			cmd.set_option("coord", m_destination.get_coord());
			cmd.set_option(m_destifr ? "ifr" : "vfr", "");
			cmd.set_option("star", m_starstart);
			if (m_fplautoroutespinbuttonstarlimit)
				cmd.set_option("starlimit", m_fplautoroutespinbuttonstarlimit->get_value());
			m_proxy.sendcmd(cmd);
		}
		{
			CFMUAutorouteProxy::Command cmd("enroute");
			if (m_fplautoroutespinbuttondctlimit)
				cmd.set_option("dctlimit", m_fplautoroutespinbuttondctlimit->get_value());
			if (m_fplautoroutespinbuttondctpenalty)
				cmd.set_option("dctpenalty", m_fplautoroutespinbuttondctpenalty->get_value());
			if (m_fplautoroutespinbuttonvfraspclimit)
				cmd.set_option("vfraspclimit", m_fplautoroutespinbuttonvfraspclimit->get_value());
			if (m_fplautoroutecheckbuttonforceenrouteifr)
				cmd.set_option("forceenrouteifr", (int)m_fplautoroutecheckbuttonforceenrouteifr->get_active());
			if (m_fplautoroutecheckbuttonawylevels)
				cmd.set_option("honourawylevels", (int)m_fplautoroutecheckbuttonawylevels->get_active());
			m_proxy.sendcmd(cmd);
		}
		{
			CFMUAutorouteProxy::Command cmd("tfr");
			if (m_fplautoroutecheckbuttontfr)
				cmd.set_option("enabled", (int)m_fplautoroutecheckbuttontfr->get_active());
			if (m_fplautorouteentrytracerules)
				cmd.set_option("trace", m_fplautorouteentrytracerules->get_text());
			if (m_fplautorouteentrydisablerules)
				cmd.set_option("disable", m_fplautorouteentrydisablerules->get_text());
			m_proxy.sendcmd(cmd);
		}
		if (m_fplautorouteentryexclude) {
			std::string txt;
			{
				Glib::ustring txt1(m_fplautorouteentryexclude->get_text());
				txt = txt1.uppercase();
			}
			std::string txtr;
			bool first(true);
			for (std::string::size_type n(0); n < txt.size(); ) {
				if (!std::isalnum(txt[n])) {
					++n;
					continue;
				}
				std::string id, typ;
				int base(-1), top(-1);
				while (n < txt.size() && std::isalnum(txt[n]))
					id += txt[n++];
				if (n < txt.size() && txt[n] == '/') {
					++n;
					while (n < txt.size() && std::isalnum(txt[n]))
						typ += txt[n++];
				}
				if (n + 1 < txt.size() && txt[n] == ':' && txt[n + 1] == 'F') {
					n += 2;
					base = 0;
					while (n < txt.size() && std::isdigit(txt[n]))
						base = base * 10 + txt[n++] - '0';
					if (n < txt.size() && txt[n] == '.') {
						++n;
						while (n < txt.size() && txt[n] == '.')
							++n;
						if (n < txt.size() && txt[n] == 'F') {
							++n;
							top = 0;
							while (n < txt.size() && std::isdigit(txt[n]))
								top = top * 10 + txt[n++] - '0';
						}
					}
				}
				if (typ.empty() || std::isdigit(typ[0])) {
					IcaoFlightPlan::FindCoord f(*m_sensors->get_engine());
					if (!f.find(id, id, IcaoFlightPlan::FindCoord::flag_navaid | IcaoFlightPlan::FindCoord::flag_waypoint,
						    m_departure.get_coord().halfway(m_destination.get_coord())))
						continue;
					Point pt;
					if (!f.get_coord(pt))
						continue;
					unsigned int miles(1);
					if (!typ.empty()) {
						std::istringstream iss(typ);
						iss >> miles;
					}
					Rect bbox(pt, pt);
					bbox = bbox.oversize_nmi(miles);
					std::ostringstream oss;
					oss << id;
					if (miles != 1 || !typ.empty())
						oss << '/' << miles;
					if (base >= 0) {
						oss << ":F" << std::setw(3) << std::setfill('0') << base;
						if (top >= 0)
							oss << "..F" << std::setw(3) << std::setfill('0') << top;
					}
					if (!txtr.empty())
						txtr += " ";
					txtr += oss.str();
					CFMUAutorouteProxy::Command cmd("exclude");
					if (first) {
						cmd.set_option("clear", 1);
						first = false;
					}
					cmd.set_option("sw", bbox.get_southwest());
					cmd.set_option("ne", bbox.get_northeast());
					if (base >= 0) {
						cmd.set_option("base", base);
						cmd.set_option("top", (top >= 0) ? top : base);
					}
					m_proxy.sendcmd(cmd);
					continue;
				}
				std::ostringstream oss;
				oss << id << '/' << typ;
				if (base >= 0) {
					oss << ":F" << std::setw(3) << std::setfill('0') << base;
					if (top >= 0)
						oss << "..F" << std::setw(3) << std::setfill('0') << top;
				}
				if (!txtr.empty())
					txtr += " ";
				txtr += oss.str();
				CFMUAutorouteProxy::Command cmd("exclude");
				if (first) {
					cmd.set_option("clear", 1);
					first = false;
				}
				cmd.set_option("aspcid", id);
				cmd.set_option("aspctype", typ);
				if (base >= 0) {
					cmd.set_option("base", base);
					cmd.set_option("top", (top >= 0) ? top : base);
				}
				m_proxy.sendcmd(cmd);
			}
			m_fplautorouteentryexclude->set_text(txtr);
		}
		{
			CFMUAutorouteProxy::Command cmd("levels");
			if (m_fplautoroutespinbuttonbaselevel)
				cmd.set_option("base", m_fplautoroutespinbuttonbaselevel->get_value());
			if (m_fplautoroutespinbuttontoplevel)
				cmd.set_option("top", m_fplautoroutespinbuttontoplevel->get_value());
			m_proxy.sendcmd(cmd);
		}
		{
			CFMUAutorouteProxy::Command cmd("atmosphere");
			if (m_sensors)
				cmd.set_option("qnh", m_sensors->get_altimeter_qnh());
			if (m_sensors)
				cmd.set_option("isa", m_sensors->get_altimeter_tempoffs());
			m_proxy.sendcmd(cmd);
		}
		{
			CFMUAutorouteProxy::Command cmd("cruise");
			if (m_sensors)
				cmd.set_option("rpm", m_sensors->get_acft_cruise_rpm());
			if (m_sensors)
				cmd.set_option("mp", m_sensors->get_acft_cruise_mp());
			m_proxy.sendcmd(cmd);
		}
		{
			CFMUAutorouteProxy::Command cmd("optimization");
			if (m_fplautoroutecomboboxopttarget) {
				int x(m_fplautoroutecomboboxopttarget->get_active_row_number());
				cmd.set_option("target", x ? "fuel" : "time");
			}
			m_proxy.sendcmd(cmd);
		}
		{
			CFMUAutorouteProxy::Command cmd("preload");
			m_proxy.sendcmd(cmd);
		}
	} catch (const std::exception& e) {
		append_log(std::string("Exception: ") + e.what() + "\n");
	}
	{
		Point pt(get_bbox().boxcenter());
		int alt(50);
		if (m_fplautoroutespinbuttonbaselevel)
				alt = m_fplautoroutespinbuttonbaselevel->get_value();
		alt *= 100;
		if (m_fplautoroutedrawingarea) {
			Glib::TimeVal tv;
			tv.assign_current_time();		
 			m_fplautoroutedrawingarea->set_center(pt, alt, 0, tv.tv_sec);
			m_fplautoroutedrawingarea->unset_cursor();
		}
	}
}

void FlightDeckWindow::FPLAutoRouteDialog::pg2_back_clicked(void)
{
	if (!m_fplautoroutenotebook) {
		close();
		return;
	}
	m_fplautoroutenotebook->set_current_page(0);
	if (m_fplautoroutetogglebuttonpg2run)
		m_fplautoroutetogglebuttonpg2run->set_active(false);
	try {
		CFMUAutorouteProxy::Command cmd("clear");
		m_proxy.sendcmd(cmd);
	} catch (const std::exception& e) {
		append_log(std::string("Exception: ") + e.what() + "\n");
	}
}

void FlightDeckWindow::FPLAutoRouteDialog::pg2_run_clicked(void)
{
	if (!m_fplautoroutetogglebuttonpg2run)
		return;
	try {
		CFMUAutorouteProxy::Command cmd(m_fplautoroutetogglebuttonpg2run->get_active() ? "start" : "stop");
		m_proxy.sendcmd(cmd);
	} catch (const std::exception& e) {
		append_log(std::string("Exception: ") + e.what() + "\n");
	}
}

void FlightDeckWindow::FPLAutoRouteDialog::ifr_toggled(void)
{

	if (m_fplautoroutespinbuttonsidlimit)
		m_sidstarlimit[0][m_depifr] = m_fplautoroutespinbuttonsidlimit->get_value();
	if (m_fplautoroutespinbuttonstarlimit)
		m_sidstarlimit[1][m_destifr] = m_fplautoroutespinbuttonstarlimit->get_value();
	if (m_fplautoroutecheckbuttondepifr)
		m_depifr = m_fplautoroutecheckbuttondepifr->get_active();
	if (m_fplautoroutecheckbuttondestifr)
		m_destifr = m_fplautoroutecheckbuttondestifr->get_active();
	if (m_fplautoroutespinbuttonsidlimit)
		m_fplautoroutespinbuttonsidlimit->set_value(m_sidstarlimit[0][m_depifr]);
	if (m_fplautoroutespinbuttonstarlimit)
		m_fplautoroutespinbuttonstarlimit->set_value(m_sidstarlimit[1][m_destifr]);
	if (m_fplautoroutelabelsid)
		m_fplautoroutelabelsid->set_text(m_depifr ? "SID End" : "Joining");
	if (m_fplautoroutelabelstar)
		m_fplautoroutelabelstar->set_text(m_destifr ? "STAR Begin" : "Leaving");
	initial_fplan();
}

void FlightDeckWindow::FPLAutoRouteDialog::pt_changed(unsigned int i)
{
	if (i >= sizeof(m_ptchanged)/sizeof(m_ptchanged[0]))
		return;
	m_ptchanged[i] = true;
}

bool FlightDeckWindow::FPLAutoRouteDialog::pt_focusout(GdkEventFocus *event)
{
	for (unsigned int i = 0; i < sizeof(m_ptchanged)/sizeof(m_ptchanged[0]); ++i) {
		if (!m_ptchanged[i])
			continue;
		m_ptchanged[i] = false;
		if (i < 2) {
			// find airport
			Glib::ustring arptname;
			if (i) {
				if (m_fplautorouteentrydesticao)
					arptname = m_fplautorouteentrydesticao->get_text();
			} else {
				if (m_fplautorouteentrydepicao)
					arptname = m_fplautorouteentrydepicao->get_text();
			}
			AirportsDb::Airport el;
			if (!arptname.empty()) {
				IcaoFlightPlan::FindCoord f(*m_sensors->get_engine());
				if (f.find(arptname, "", IcaoFlightPlan::FindCoord::flag_airport))
					if (!f.get_airport(el))
						el = AirportsDb::Airport();
			}
			if (i) {
				m_destination = el;
				if (m_fplautorouteentrydestname) {
					if (m_destination.is_valid())
						m_fplautorouteentrydestname->set_text(m_destination.get_name());
					else
						m_fplautorouteentrydestname->set_text("");
				}
			} else {
				m_departure = el;
				if (m_fplautorouteentrydepname) {
					if (m_departure.is_valid())
						m_fplautorouteentrydepname->set_text(m_departure.get_name());
					else
						m_fplautorouteentrydepname->set_text("");
				}
			}
			continue;
		}
		// find navaid or intersection
		Glib::ustring ptname;
		Point pt;
		if (i == 2) {
			if (m_fplautorouteentrysidicao)
				ptname = m_fplautorouteentrysidicao->get_text();
			if (m_departure.is_valid())
				pt = m_departure.get_coord();
			else if (m_destination.is_valid())
				pt = m_destination.get_coord();
		} else {
			if (m_fplautorouteentrystaricao)
				ptname = m_fplautorouteentrystaricao->get_text();
			if (m_destination.is_valid())
				pt = m_destination.get_coord();
			else if (m_departure.is_valid())
				pt = m_departure.get_coord();
		}
		if (ptname.empty()) {
			pt = Point::invalid;
			ptname.clear();
		} else {
			IcaoFlightPlan::FindCoord f(*m_sensors->get_engine());
			if (f.find(ptname, ptname, IcaoFlightPlan::FindCoord::flag_navaid | IcaoFlightPlan::FindCoord::flag_waypoint, pt)) {
				pt = Point::invalid;
				ptname.clear();
				NavaidsDb::element_t nav;
				WaypointsDb::element_t wpt;
				if (f.get_navaid(nav)) {
					pt = nav.get_coord();
					ptname = nav.get_navaid_typename() + " " + nav.get_name();
				} else if (f.get_waypoint(wpt)) {
					pt = wpt.get_coord();
					ptname = "INT " + wpt.get_name();
				}
			} else {
				pt = Point::invalid;
				ptname.clear();
			}
		}
		if (i == 2) {
			m_sidend = pt;
			if (m_fplautorouteentrysidname)
				m_fplautorouteentrysidname->set_text(ptname);
		} else {
			m_starstart = pt;
			if (m_fplautorouteentrystarname)
				m_fplautorouteentrystarname->set_text(ptname);
		}
	}
	{
		bool ok(m_departure.is_valid() && m_destination.is_valid());
		if (m_fplautoroutebuttonpg1copy)
			m_fplautoroutebuttonpg1copy->set_sensitive(ok);
		if (m_fplautoroutebuttonpg1next)
			m_fplautoroutebuttonpg1next->set_sensitive(ok);
	}
	initial_fplan();
	return false;
}
