/***************************************************************************
 *   Copyright (C) 2012 by Thomas Sailer                                   *
 *   t.sailer@alumni.ethz.ch                                               *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "sysdeps.h"
#include "engine.h"

#include <iomanip>

class Engine::DbObjectAirport : public DbObject {
  public:
	DbObjectAirport(const AirportsDb::Airport& el) : m_element(el) {}
	virtual Point get_coord(void) const { return m_element.get_coord(); }
	virtual const Glib::ustring& get_ident(void) const { return m_element.get_icao(); }
	virtual const Glib::ustring& get_icao(void) const { return m_element.get_icao(); }
	virtual const Glib::ustring& get_name(void) const { return m_element.get_name(); }
	virtual bool is_airport(void) const { return true; }
	using DbObject::get;
	virtual bool get(AirportsDb::Airport& el) const { el = m_element; return true; }
	using DbObject::set;
	virtual bool set(FPlanWaypoint& el) const { set(el, m_element); return true; }
	using DbObject::insert;
	virtual unsigned int insert(FPlanRoute& route, uint32_t nr = ~0U) const { return insert(route, m_element, nr); }

  protected:
	AirportsDb::Airport m_element;
};

class Engine::DbObjectNavaid : public DbObject {
  public:
	DbObjectNavaid(const NavaidsDb::Navaid& el) : m_element(el) {}
	virtual Point get_coord(void) const { return m_element.get_coord(); }
	virtual const Glib::ustring& get_ident(void) const { return m_element.get_icao(); }
	virtual const Glib::ustring& get_icao(void) const { return m_element.get_icao(); }
	virtual const Glib::ustring& get_name(void) const { return m_element.get_name(); }
	virtual bool is_navaid(void) const { return true; }
	using DbObject::get;
	virtual bool get(NavaidsDb::Navaid& el) const { el = m_element; return true; }
	using DbObject::set;
	virtual bool set(FPlanWaypoint& el) const { set(el, m_element); return true; }
	using DbObject::insert;
	virtual unsigned int insert(FPlanRoute& route, uint32_t nr = ~0U) const { return insert(route, m_element, nr); }

  protected:
	NavaidsDb::Navaid m_element;
};

class Engine::DbObjectWaypoint : public DbObject {
  public:
	DbObjectWaypoint(const WaypointsDb::Waypoint& el) : m_element(el) {}
	virtual Point get_coord(void) const { return m_element.get_coord(); }
	virtual const Glib::ustring& get_ident(void) const { return m_element.get_name(); }
	virtual const Glib::ustring& get_icao(void) const { return m_element.get_icao(); }
	virtual const Glib::ustring& get_name(void) const { return m_element.get_name(); }
	virtual bool is_intersection(void) const { return true; }
	using DbObject::get;
	virtual bool get(WaypointsDb::Waypoint& el) const { el = m_element; return true; }
	using DbObject::set;
	virtual bool set(FPlanWaypoint& el) const { set(el, m_element); return true; }
	using DbObject::insert;
	virtual unsigned int insert(FPlanRoute& route, uint32_t nr = ~0U) const { return insert(route, m_element, nr); }

  protected:
	WaypointsDb::Waypoint m_element;
};

class Engine::DbObjectMapelement : public DbObject {
  public:
	DbObjectMapelement(const MapelementsDb::Mapelement& el) : m_element(el) {}
	virtual Point get_coord(void) const { return m_element.get_coord(); }
	virtual const Glib::ustring& get_ident(void) const { return m_element.get_name(); }
	virtual const Glib::ustring& get_name(void) const { return m_element.get_name(); }
	virtual bool is_mapelement(void) const { return true; }
	using DbObject::get;
	virtual bool get(MapelementsDb::Mapelement& el) const { el = m_element; return true; }
	using DbObject::set;
	virtual bool set(FPlanWaypoint& el) const { set(el, m_element); return true; }
	using DbObject::insert;
	virtual unsigned int insert(FPlanRoute& route, uint32_t nr = ~0U) const { return insert(route, m_element, nr); }

  protected:
	MapelementsDb::Mapelement m_element;
};

class Engine::DbObjectFPlanWaypoint : public DbObject {
  public:
	DbObjectFPlanWaypoint(const FPlanWaypoint& el) : m_element(el) {}
	virtual Point get_coord(void) const { return m_element.get_coord(); }
	virtual const Glib::ustring& get_ident(void) const { return m_element.get_icao().empty() ? m_element.get_name() : m_element.get_icao(); }
	virtual const Glib::ustring& get_icao(void) const { return m_element.get_icao(); }
	virtual const Glib::ustring& get_name(void) const { return m_element.get_name(); }
	virtual bool is_fplwaypoint(void) const { return true; }
	using DbObject::get;
	virtual bool get(FPlanWaypoint& el) const { el = m_element; return true; }
	using DbObject::set;
	virtual bool set(FPlanWaypoint& el) const { set(el, m_element); return true; }
	using DbObject::insert;
	virtual unsigned int insert(FPlanRoute& route, uint32_t nr = ~0U) const { return insert(route, m_element, nr); }

  protected:
	FPlanWaypoint m_element;
};

const Glib::ustring Engine::DbObject::empty;

Engine::DbObject::DbObject(void)
	: m_refcount(1)
{
}

Engine::DbObject::~DbObject()
{
}

void Engine::DbObject::reference(void) const
{
        ++m_refcount;
}

void Engine::DbObject::unreference(void) const
{
        if (--m_refcount)
                return;
        delete this;
}


Glib::RefPtr<Engine::DbObject> Engine::DbObject::create(const AirportsDb::Airport& el)
{
	return Glib::RefPtr<Engine::DbObject>(new DbObjectAirport(el));
}

Glib::RefPtr<Engine::DbObject> Engine::DbObject::create(const NavaidsDb::Navaid& el)
{
	return Glib::RefPtr<Engine::DbObject>(new DbObjectNavaid(el));
}

Glib::RefPtr<Engine::DbObject> Engine::DbObject::create(const WaypointsDb::Waypoint& el)
{
	return Glib::RefPtr<Engine::DbObject>(new DbObjectWaypoint(el));
}

Glib::RefPtr<Engine::DbObject> Engine::DbObject::create(const MapelementsDb::Mapelement& el)
{
	return Glib::RefPtr<Engine::DbObject>(new DbObjectMapelement(el));
}

Glib::RefPtr<Engine::DbObject> Engine::DbObject::create(const FPlanWaypoint& el)
{
	return Glib::RefPtr<Engine::DbObject>(new DbObjectFPlanWaypoint(el));
}

void Engine::DbObject::set(FPlanWaypoint& wpt, const AirportsDb::Airport& el)
{
	wpt.set_icao(el.get_icao());
	wpt.set_name(el.get_name());
	wpt.set_pathcode(FPlanWaypoint::pathcode_none);
	wpt.set_pathname("");
	wpt.set_coord(el.get_coord());
	wpt.set_altitude(el.get_elevation());
	wpt.set_flags(wpt.get_flags() & ~FPlanWaypoint::altflag_standard);
	wpt.set_frequency(0);
	Glib::ustring note("Airport Type: " + el.get_type_string());
	if (el.get_nrrwy() > 0)
		note += "\n\nRunways:";
	for (unsigned int nr(0); nr < el.get_nrrwy(); ++nr) {
		const AirportsDb::Airport::Runway& rwy(el.get_rwy(nr));
		if (rwy.get_flags() & AirportsDb::Airport::Runway::flag_le_usable) {
			std::ostringstream oss;
			oss << "\n  " << std::setw(4) << rwy.get_ident_le() << ' '
			    << std::setw(4) << Conversions::track_str(rwy.get_le_hdg() * FPlanLeg::to_deg) << ' '
			    << std::fixed << std::setw(4) << std::setprecision(0) << (rwy.get_length() * Point::ft_to_m_dbl) << 'x'
			    << std::fixed << std::setprecision(0) << (rwy.get_width() * Point::ft_to_m_dbl) << ' '
			    << std::fixed << std::setw(4) << std::setprecision(0) << (rwy.get_le_tda() * Point::ft_to_m_dbl) << ' '
			    << std::fixed << std::setw(4) << std::setprecision(0) << (rwy.get_le_lda() * Point::ft_to_m_dbl) << ' '
			    << std::setw(4) << Conversions::alt_str(rwy.get_le_elev(), 0) << ' ' << rwy.get_surface();
			note += oss.str();
			Glib::ustring lights;
			for (unsigned int i = 0; i < AirportsDb::Airport::Runway::nr_lights; ++i) {
				AirportsDb::Airport::Runway::light_t l(rwy.get_le_light(i));
				if (l == AirportsDb::Airport::Runway::light_off)
					continue;
				if (!lights.empty())
					lights += "; ";
				lights += AirportsDb::Airport::Runway::get_light_string(l);
			}
			note += " " + lights;
		}
		if (rwy.get_flags() & AirportsDb::Airport::Runway::flag_he_usable) {
			std::ostringstream oss;
			oss << "\n  " << std::setw(4) << rwy.get_ident_he() << ' '
			    << std::setw(4) << Conversions::track_str(rwy.get_he_hdg() * FPlanLeg::to_deg) << ' '
			    << std::fixed << std::setw(4) << std::setprecision(0) << (rwy.get_length() * Point::ft_to_m_dbl) << 'x'
			    << std::fixed << std::setprecision(0) << (rwy.get_width() * Point::ft_to_m_dbl) << ' '
			    << std::fixed << std::setw(4) << std::setprecision(0) << (rwy.get_he_tda() * Point::ft_to_m_dbl) << ' '
			    << std::fixed << std::setw(4) << std::setprecision(0) << (rwy.get_he_lda() * Point::ft_to_m_dbl) << ' '
			    << std::setw(4) << Conversions::alt_str(rwy.get_he_elev(), 0) << ' ' << rwy.get_surface();
			note += oss.str();
			Glib::ustring lights;
			for (unsigned int i = 0; i < AirportsDb::Airport::Runway::nr_lights; ++i) {
				AirportsDb::Airport::Runway::light_t l(rwy.get_he_light(i));
				if (l == AirportsDb::Airport::Runway::light_off)
					continue;
				if (!lights.empty())
					lights += "; ";
				lights += AirportsDb::Airport::Runway::get_light_string(l);
			}
			note += " " + lights;
		}
	}
	if (el.get_nrhelipads() > 0)
		note += "\n\nHelipads:";
	for (unsigned int nr(0); nr < el.get_nrhelipads(); ++nr) {
		const AirportsDb::Airport::Helipad& helipad(el.get_helipad(nr));
		if (!(helipad.get_flags() & AirportsDb::Airport::Helipad::flag_usable))
			continue;
		 std::ostringstream oss;
		 oss << "\n  " << std::setw(4) << helipad.get_ident() << ' '
		     << std::setw(4) << Conversions::track_str(helipad.get_hdg() * FPlanLeg::to_deg) << ' '
		     << std::fixed << std::setw(4) << std::setprecision(0) << (helipad.get_length() * Point::ft_to_m_dbl) << 'x'
		     << std::fixed << std::setprecision(0) << (helipad.get_width() * Point::ft_to_m_dbl) << ' '
		     << std::setw(4) << Conversions::alt_str(helipad.get_elev(), 0);
		 note += oss.str();
	}
	if (el.get_nrcomm() > 0) {
		note += "\n\nComm:";
		unsigned int prio(0);
		Glib::ustring twr("twr");
		twr = twr.casefold();
		for (unsigned int nr(0); nr < el.get_nrcomm(); ++nr) {
			const AirportsDb::Airport::Comm& comm(el.get_comm(nr));
			note += "\n  " + comm.get_name() + " " + Conversions::freq_str(comm[0]) + " " + Conversions::freq_str(comm[1]) + " " +
				Conversions::freq_str(comm[2]) + " " + Conversions::freq_str(comm[3]) + " " + Conversions::freq_str(comm[4]) + " " +
				comm.get_sector() + " " + comm.get_oprhours();
			if (!comm[0])
				continue;
			unsigned int p(1);
			Glib::ustring name(comm.get_name().casefold());
			if (name == twr)
				p = 3;
			else if (name.find(twr) != Glib::ustring::npos)
				p = 2;
			if (p <= prio)
				continue;
			wpt.set_frequency(comm[0]);
		}
	}
	if (!el.get_vfrrmk().empty())
		note += "\n\n" + el.get_vfrrmk();
	wpt.set_note(note);
}

void Engine::DbObject::set(FPlanWaypoint& wpt, const NavaidsDb::Navaid& el)
{
	wpt.set_flags(wpt.get_flags() & ~FPlanWaypoint::altflag_standard);
	wpt.set_icao(el.get_icao());
	wpt.set_name(el.get_name());
	wpt.set_pathcode(FPlanWaypoint::pathcode_none);
	wpt.set_pathname("");
	wpt.set_coord(el.get_coord());
	wpt.set_altitude(el.get_elev());
	wpt.set_frequency(el.get_frequency());
	std::ostringstream range;
	range << el.get_range();
	Glib::ustring note(el.get_navaid_typename() + " Range " + range.str() + " ");
	if (el.is_inservice())
		note += "in";
	else
		note += "out of";
	note += " service";
	wpt.set_note(note);
}

void Engine::DbObject::set(FPlanWaypoint& wpt, const WaypointsDb::Waypoint& el)
{
	wpt.set_icao(el.get_icao());
	wpt.set_name(el.get_name());
	wpt.set_pathcode(FPlanWaypoint::pathcode_none);
	wpt.set_pathname("");
	wpt.set_coord(el.get_coord());
	wpt.set_note(el.get_usage_name());
}

void Engine::DbObject::set(FPlanWaypoint& wpt, const MapelementsDb::Mapelement& el)
{
	wpt.set_name(el.get_name());
	wpt.set_coord(el.get_coord());
	wpt.set_note(el.get_typename());
	wpt.set_pathcode(FPlanWaypoint::pathcode_none);
	wpt.set_pathname("");
}

void Engine::DbObject::set(FPlanWaypoint& wpt, const FPlanWaypoint& el)
{
	wpt = el;
}

FPlanWaypoint Engine::DbObject::insert_prepare(FPlanRoute& route, uint32_t nr)
{
	FPlanWaypoint wpt;
	{
		time_t t(time(0));
		wpt.set_time_unix(t);
                wpt.set_save_time_unix(t);
	}
	wpt.set_flags(0);
	wpt.set_altitude(0);
        unsigned int nrwpt(route.get_nrwpt());
	if (!nrwpt)
		return wpt;
	if (nr >= 1 && nr <= nrwpt) {
		const FPlanWaypoint& wpt1(route[nr - 1]);
		wpt.set_flags(wpt1.get_flags() & (FPlanWaypoint::altflag_ifr | FPlanWaypoint::altflag_standard));
		wpt.set_altitude(wpt1.get_altitude());
	}
	return wpt;
}

unsigned int Engine::DbObject::insert(FPlanRoute& route, const AirportsDb::Airport& el, uint32_t nr)
{
	FPlanWaypoint wpt(insert_prepare(route, nr));
	set(wpt, el);
	route.insert_wpt(nr, wpt);
	return 1;
}

unsigned int Engine::DbObject::insert(FPlanRoute& route, unsigned int vfrrtenr, const AirportsDb::Airport& el, uint32_t nr)
{
	if (vfrrtenr >= el.get_nrvfrrte())
		return insert(route, el, nr);
	const AirportsDb::Airport::VFRRoute& vfrrte(el.get_vfrrte(vfrrtenr));
	FPlanWaypoint wpt(insert_prepare(route, nr));
	if (!vfrrte.size()) {
		set(wpt, el);
		wpt.set_pathcode(FPlanWaypoint::pathcode_vfrtransition);
		wpt.set_pathname(vfrrte.get_name());
		route.insert_wpt(nr, wpt);
		return 1;
	}
	nr = std::max(nr, static_cast<uint32_t>(route.get_nrwpt()));
	for (unsigned int nrr(0); nrr < vfrrte.size(); ++nrr) {
		const AirportsDb::Airport::VFRRoute::VFRRoutePoint& vfrpt(vfrrte[nrr]);
		if (vfrpt.is_at_airport()) {
			set(wpt, el);
		} else {
			wpt.set_icao(el.get_icao());
			wpt.set_name(vfrpt.get_name());
			wpt.set_coord(vfrpt.get_coord());
			wpt.set_altitude(vfrpt.get_altitude());
			wpt.set_frequency(0);
			Glib::ustring note(vfrpt.get_altcode_string());
			if (vfrpt.is_compulsory())
				note += " Compulsory";
			note += "\n" + vfrpt.get_pathcode_string();
			wpt.set_note(note);
		}
		switch (vfrpt.get_pathcode()) {
		case AirportsDb::Airport::VFRRoute::VFRRoutePoint::pathcode_arrival:
			wpt.set_pathcode(FPlanWaypoint::pathcode_vfrarrival);
			break;
							
		case AirportsDb::Airport::VFRRoute::VFRRoutePoint::pathcode_departure:
			wpt.set_pathcode(FPlanWaypoint::pathcode_vfrdeparture);
			break;
							
		case AirportsDb::Airport::VFRRoute::VFRRoutePoint::pathcode_vfrtransition:
			wpt.set_pathcode(FPlanWaypoint::pathcode_vfrtransition);
			break;
							
		case AirportsDb::Airport::VFRRoute::VFRRoutePoint::pathcode_holding:
		case AirportsDb::Airport::VFRRoute::VFRRoutePoint::pathcode_trafficpattern:
		case AirportsDb::Airport::VFRRoute::VFRRoutePoint::pathcode_other:
		case AirportsDb::Airport::VFRRoute::VFRRoutePoint::pathcode_invalid:
		default:
			wpt.set_pathcode(FPlanWaypoint::pathcode_none);
			break;
		}
		wpt.set_pathname(vfrrte.get_name());
		route.insert_wpt(nr, wpt);
		++nr;
	}
	return vfrrte.size();
}

unsigned int Engine::DbObject::insert(FPlanRoute& route, unsigned int vfrrtenr, unsigned int vfrrteptnr, const AirportsDb::Airport& el, uint32_t nr)
{
	if (vfrrtenr >= el.get_nrvfrrte())
		return insert(route, el, nr);
	const AirportsDb::Airport::VFRRoute& vfrrte(el.get_vfrrte(vfrrtenr));
	if (vfrrteptnr >= vfrrte.size())
		return insert(route, vfrrtenr, el, nr);
	FPlanWaypoint wpt(insert_prepare(route, nr));
	const AirportsDb::Airport::VFRRoute::VFRRoutePoint& vfrpt(vfrrte[vfrrteptnr]);
	if (vfrpt.is_at_airport()) {
		set(wpt, el);
	} else {
		wpt.set_icao(el.get_icao());
		wpt.set_name(vfrpt.get_name());
		wpt.set_coord(vfrpt.get_coord());
		wpt.set_altitude(vfrpt.get_altitude());
		wpt.set_frequency(0);
		Glib::ustring note(vfrpt.get_altcode_string());
		if (vfrpt.is_compulsory())
			note += " Compulsory";
		note += "\n" + vfrpt.get_pathcode_string();
		wpt.set_note(note);
	}
	switch (vfrpt.get_pathcode()) {
	case AirportsDb::Airport::VFRRoute::VFRRoutePoint::pathcode_arrival:
		wpt.set_pathcode(FPlanWaypoint::pathcode_vfrarrival);
		break;

	case AirportsDb::Airport::VFRRoute::VFRRoutePoint::pathcode_departure:
		wpt.set_pathcode(FPlanWaypoint::pathcode_vfrdeparture);
		break;

	case AirportsDb::Airport::VFRRoute::VFRRoutePoint::pathcode_vfrtransition:
		wpt.set_pathcode(FPlanWaypoint::pathcode_vfrtransition);
		break;

	case AirportsDb::Airport::VFRRoute::VFRRoutePoint::pathcode_holding:
	case AirportsDb::Airport::VFRRoute::VFRRoutePoint::pathcode_trafficpattern:
	case AirportsDb::Airport::VFRRoute::VFRRoutePoint::pathcode_other:
	case AirportsDb::Airport::VFRRoute::VFRRoutePoint::pathcode_invalid:
	default:
		wpt.set_pathcode(FPlanWaypoint::pathcode_none);
		break;
	}
	wpt.set_pathname(vfrrte.get_name());
	route.insert_wpt(nr, wpt);
	return 1;
}

unsigned int Engine::DbObject::insert(FPlanRoute& route, const NavaidsDb::Navaid& el, uint32_t nr)
{
	FPlanWaypoint wpt(insert_prepare(route, nr));
	set(wpt, el);
	route.insert_wpt(nr, wpt);
	return 1;
}

unsigned int Engine::DbObject::insert(FPlanRoute& route, const WaypointsDb::Waypoint& el, uint32_t nr)
{
	FPlanWaypoint wpt(insert_prepare(route, nr));
	set(wpt, el);
	route.insert_wpt(nr, wpt);
	return 1;
}

unsigned int Engine::DbObject::insert(FPlanRoute& route, const MapelementsDb::Mapelement& el, uint32_t nr)
{
	FPlanWaypoint wpt(insert_prepare(route, nr));
	set(wpt, el);
	route.insert_wpt(nr, wpt);
	return 1;
}

unsigned int Engine::DbObject::insert(FPlanRoute& route, const FPlanWaypoint& el, uint32_t nr)
{
	FPlanWaypoint wpt(insert_prepare(route, nr));
	set(wpt, el);
	route.insert_wpt(nr, wpt);
	return 1;
}
