#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <iostream>
#include <iomanip>

#include <boost/graph/dijkstra_shortest_paths.hpp>
#include <boost/graph/johnson_all_pairs_shortest.hpp>
#include <boost/graph/floyd_warshall_shortest.hpp>
#include <boost/graph/filtered_graph.hpp>

#include "cfmuautoroute.hh"

const bool CFMUAutoroute::lgraphrestartedgescan;
const bool CFMUAutoroute::lgraphawyvertdct;

bool CFMUAutoroute::LSuperVertex::is_airport(void) const
{
	if (!m_obj)
		return false;
	return m_obj->is_airport();
}

bool CFMUAutoroute::LSuperVertex::is_navaid(void) const
{
	if (!m_obj)
		return false;
	return m_obj->is_navaid();
}

bool CFMUAutoroute::LSuperVertex::is_intersection(void) const
{
	if (!m_obj)
		return false;
	return m_obj->is_intersection();
}

bool CFMUAutoroute::LSuperVertex::is_mapelement(void) const
{
	if (!m_obj)
		return false;
	return m_obj->is_mapelement();
}

bool CFMUAutoroute::LSuperVertex::get(AirportsDb::Airport& el) const
{
	if (!m_obj)
		return false;
	return m_obj->get(el);
}

bool CFMUAutoroute::LSuperVertex::get(NavaidsDb::Navaid& el) const
{
	if (!m_obj)
		return false;
	return m_obj->get(el);
}

bool CFMUAutoroute::LSuperVertex::get(WaypointsDb::Waypoint& el) const
{
	if (!m_obj)
		return false;
	return m_obj->get(el);
}

bool CFMUAutoroute::LSuperVertex::get(MapelementsDb::Mapelement& el) const
{
	if (!m_obj)
		return false;
	return m_obj->get(el);
}

bool CFMUAutoroute::LSuperVertex::set(FPlanWaypoint& el) const
{
	if (!m_obj)
		return false;
	return m_obj->set(el);
}

unsigned int CFMUAutoroute::LSuperVertex::insert(FPlanRoute& route, uint32_t nr) const
{
	if (!m_obj)
		return false;
	return m_obj->insert(route, nr);
}

const Glib::ustring& CFMUAutoroute::LSuperVertex::get_ident(void) const
{
	if (get_object())
		return get_object()->get_ident();
	return Engine::DbObject::empty;
}

bool CFMUAutoroute::LSuperVertex::is_ident_valid(void) const
{
	const Glib::ustring& id(get_ident());
	return id.size() >= 2 && !Engine::AirwayGraphResult::Vertex::is_ident_numeric(id);
}

const double CFMUAutoroute::LEdge::maxmetric = std::numeric_limits<double>::max();

bool CFMUAutoroute::LEdge::is_match(lgraphairwayindex_t ident, lgraphairwayindex_t id)
{
	switch (id) {
	case airwaymatchall:
		return true;

	case airwaymatchnone:
		return false;

	case airwaymatchawy:
		return ident < airwaydct;

	case airwaymatchdctawy:
		return ident <= airwaydct;

	case airwaymatchdctawysidstar:
		return ident <= airwaysid;

	case airwaymatchdctawysidstarlevel:
		return ident <= airwaylevelup;

	case airwaymatchsidstar:
		return ident >= airwaystar && ident <= airwaysid;

	case airwaymatchlevel:
		return ident >= airwayleveldown && ident <= airwaylevelup;

	default:
		return ident == id;
	}
}

CFMUAutoroute::LVertexEdge::LVertexEdge(LGraph::edge_descriptor e, const LGraph& g)
{
	const LEdge& ee(g[e]);
	m_vertex = boost::target(e, g);
	m_edge = ee.get_ident();
}

bool CFMUAutoroute::LVertexEdge::operator<(const LVertexEdge& x) const
{
	if (get_vertex() < x.get_vertex())
		return true;
	if (get_vertex() > x.get_vertex())
		return false;
	return get_edge() < x.get_edge();
}

bool CFMUAutoroute::LRoute::operator<(const LRoute& x) const
{
	if (get_metric() < x.get_metric())
		return true;
	if (get_metric() > x.get_metric())
		return false;
	for (size_type i(0), n(std::min(size(), x.size())); i < n; ++i) {
		if ((*this)[i] < x[i])
			return true;
		if ((*this)[i] > x[i])
			return false;
	}
	return size() < x.size();
}

bool CFMUAutoroute::LRoute::is_repeated_nodes(const LGraph& g, const LSuperVertices& supervertices) const
{
	typedef std::set<unsigned int> nmap_t;
	nmap_t nmap;
	typedef std::set<Glib::ustring> imap_t;
	imap_t imap;
	for (const_iterator i(begin()), e(end()); i != e;) {
		const LVertex& v(g[i->get_vertex()]);
		unsigned int superindex0(v.get_superindex());
		{
			std::pair<nmap_t::iterator,bool> x(nmap.insert(superindex0));
			if (!x.second)
				return true;
		}
		if (!supervertices.empty()) {
			const LSuperVertex& super0(supervertices[superindex0]);
			std::pair<imap_t::iterator,bool> x(imap.insert(super0.get_ident()));
			if (!x.second)
				return true;
		}
		++i;
		while (i != e) {
			const LVertex& v1(g[i->get_vertex()]);
			if (v1.get_superindex() != superindex0)
				break;
			++i;
		}
		if (false) {
			if (i == e)
				break;
			if (!supervertices.empty()) {
				const LVertex& v1(g[i->get_vertex()]);
				const LSuperVertex& super0(supervertices[v.get_superindex()]);
				const LSuperVertex& super1(supervertices[v1.get_superindex()]);
				if (super0.get_ident() == super1.get_ident())
					return true;
			}
		}
	}
	return false;
}

bool CFMUAutoroute::LRoute::is_existing(const LGraph& g) const
{
	for (LRoute::size_type i(1); i < this->size(); ++i) {
		LGraph::edge_descriptor e;
		bool ok;
		boost::tie(e, ok) = this->get_edge(i, g);
		if (!ok)
			return false;
	}
	return true;
}

std::pair<CFMUAutoroute::LGraph::edge_descriptor,bool> CFMUAutoroute::LRoute::get_edge(const LVertexEdge& u, const LVertexEdge& v, const LGraph& g)
{
	return get_edge(u.get_vertex(), v, g);
}

std::pair<CFMUAutoroute::LGraph::edge_descriptor,bool> CFMUAutoroute::LRoute::get_edge(LGraph::vertex_descriptor u, const LVertexEdge& v, const LGraph& g)
{
	LGraph::out_edge_iterator e0, e1;
	for (boost::tie(e0, e1) = boost::out_edges(u, g); e0 != e1; ++e0) {
		if (boost::target(*e0, g) != v.get_vertex())
			continue;
		const LEdge& ee(g[*e0]);
		if (ee.get_ident() != v.get_edge())
			continue;
		return std::pair<LGraph::edge_descriptor,bool>(*e0, true);
	}
	return std::pair<LGraph::edge_descriptor,bool>(LGraph::edge_descriptor(), false);
}

std::pair<CFMUAutoroute::LGraph::edge_descriptor,bool> CFMUAutoroute::LRoute::get_edge(size_type i, const LGraph& g) const
{
	if (i < 1 || i >= size())
		return std::pair<LGraph::edge_descriptor,bool>(LGraph::edge_descriptor(), false);
	return get_edge((*this)[i - 1], (*this)[i], g);
}

CFMUAutoroute::LEdgeAdd::LEdgeAdd(LGraph::vertex_descriptor v0, LGraph::vertex_descriptor v1, double metric)
	: m_v0(v0), m_v1(v1), m_metric(metric)
{
}

bool CFMUAutoroute::LEdgeAdd::operator<(const LEdgeAdd& x) const
{
	if (get_v0() < x.get_v0())
		return true;
	if (get_v0() > x.get_v0())
		return false;
	return get_v1() < x.get_v1();
}

void CFMUAutoroute::clearlgraph(void)
{
	m_lgraph.clear();
	m_lsupervertices.clear();
	m_lgraphairwaytable.clear();
	m_lgraphairwayrevtable.clear();
	m_solutionvertices.clear();
	m_superdep = m_superdest = 0;
	lgraphmodified();
}

CFMUAutoroute::lgraphairwayindex_t CFMUAutoroute::lgraphairway2index(const Glib::ustring& awy, bool create)
{
	Glib::ustring awyu(awy.uppercase());
	if (awyu.empty())
		return airwaydct;
	if (create) {
		std::pair<lgraphairwayrevtable_t::iterator,bool> ins(m_lgraphairwayrevtable.insert(std::make_pair(awyu, m_lgraphairwaytable.size())));
		if (ins.second)
			m_lgraphairwaytable.push_back(awyu);
		return ins.first->second;
	}
	lgraphairwayrevtable_t::const_iterator i(m_lgraphairwayrevtable.find(awyu));
	if (i == m_lgraphairwayrevtable.end())
		return airwaymatchnone;
	return i->second;
}

const Glib::ustring& CFMUAutoroute::lgraphindex2airway(lgraphairwayindex_t idx, bool specialnames)
{
	switch (idx) {
	case airwaydct:
		if (specialnames) {
			static const Glib::ustring r("DCT");
			return r;
		}
		return Engine::DbObject::empty;

	case airwaysid:
		if (specialnames) {
			static const Glib::ustring r("SID");
			return r;
		}
		return Engine::DbObject::empty;

	case airwaystar:
		if (specialnames) {
			static const Glib::ustring r("STAR");
			return r;
		}
		return Engine::DbObject::empty;

	case airwaylevelup:
		if (specialnames) {
			static const Glib::ustring r("LEVEL+");
			return r;
		}
		return Engine::DbObject::empty;

	case airwayleveldown:
		if (specialnames) {
			static const Glib::ustring r("LEVEL-");
			return r;
		}
		return Engine::DbObject::empty;

	case airwaymatchall:
	{
		static const Glib::ustring r("[MATCHALL]");
		return r;
	}

	case airwaymatchnone:
	{
		static const Glib::ustring r("[MATCHALL]");
		return r;
	}

	case airwaymatchdctawy:
	{
		static const Glib::ustring r("[MATCHDCTAWY]");
		return r;
	}

	case airwaymatchdctawysidstar:
	{
		static const Glib::ustring r("[MATCHDCTAWYSIDSTAR]");
		return r;
	}

	case airwaymatchdctawysidstarlevel:
	{
		static const Glib::ustring r("[MATCHDCTAWYSIDSTARLEVEL]");
		return r;
	}

	case airwaymatchsidstar:
	{
		static const Glib::ustring r("[MATCHDCTAWYSIDSTAR]");
		return r;
	}

	case airwaymatchlevel:
	{
		static const Glib::ustring r("[MATCHDCTAWYLEVEL]");
		return r;
	}

	default:
		break;
	}
	if (idx < m_lgraphairwaytable.size())
		return m_lgraphairwaytable[idx];
	{
		static const Glib::ustring r("[??]");
		return r;
	}
}

unsigned int CFMUAutoroute::lgraphaddvertex(Engine::DbObject::ptr_t obj, bool supernode, bool levelchg)
{
	if (!obj)
		throw std::runtime_error("LGraph Vertex: object null");
	unsigned int nr(1);
	if (supernode)
		nr = m_performance.size();
	if (!nr)
		return ~0U;
	unsigned int idx(m_lsupervertices.size());
	LGraph::vertex_descriptor v0, v1;
	for (unsigned int i = 0; i < nr; ++i) {
		LGraph::vertex_descriptor vn(boost::add_vertex(LVertex(idx), m_lgraph));
		if (!i) {
			v0 = v1 = vn;
			continue;
		}
		if (vn != v1 + 1)
			throw std::runtime_error("LGraph Vertex: nonconsecutive descriptors");
		v1 = vn;
	}
	m_lsupervertices.push_back(LSuperVertex(obj, v0, v1));
	if (!supernode || !levelchg)
		return idx;
	for (unsigned int i = 1; i < nr; ++i) {
		LGraph::vertex_descriptor u0(v0 + (i - 1)), u1(v0 + i);
		const Performance& perf(m_performance[i - 1]);
		boost::add_edge(u0, u1, LEdge(airwaylevelup, is_optfuel() ? perf.get_fuelpenalty() : perf.get_timepenalty()), m_lgraph);
		boost::add_edge(u1, u0, LEdge(airwayleveldown, 0), m_lgraph);
	}
	return idx;
}

bool CFMUAutoroute::lgraphload(const Rect& bbox)
{
	typedef Engine::AirwayGraphResult::Graph Graph;
	typedef Engine::AirwayGraphResult::Vertex Vertex;
	typedef Engine::AirwayGraphResult::Edge Edge;

	Rect bbox1(bbox.oversize_nmi(100.f));
	if (true) {
		std::ostringstream oss;
		oss << "Loading Airway graph: [" << bbox1.get_southwest().get_lat_str2() << ' ' << bbox1.get_southwest().get_lon_str2()
		    << ' ' << bbox1.get_northeast().get_lat_str2() << ' ' << bbox1.get_northeast().get_lon_str2() << ']';
		m_signal_log(log_debug0, oss.str());
	}
	Graph g;
	opendb();
	g.clear();
	{
		NavaidsDb::elementvector_t ev(m_navaiddb.find_by_rect(bbox1, ~0, NavaidsDb::Navaid::subtables_none));
		for (NavaidsDb::elementvector_t::iterator i(ev.begin()); i != ev.end();) {
			switch (i->get_navaid_type()) {
			case NavaidsDb::Navaid::navaid_vor:
			case NavaidsDb::Navaid::navaid_vordme:
			case NavaidsDb::Navaid::navaid_ndb:
			case NavaidsDb::Navaid::navaid_ndbdme:
			case NavaidsDb::Navaid::navaid_vortac:
				++i;
				break;

			default:
				i = ev.erase(i);
				break;
			}
		}
		g.add(ev);
	}
	{
		WaypointsDb::elementvector_t ev(m_waypointdb.find_by_rect(bbox1, ~0, WaypointsDb::Waypoint::subtables_none));
		for (WaypointsDb::elementvector_t::iterator i(ev.begin()); i != ev.end();) {
			switch (i->get_type()) {
			case WaypointsDb::Waypoint::type_icao:
				++i;
				break;

                        case WaypointsDb::Waypoint::type_other:
				if (i->get_icao().size() == 5) {
					++i;
					break;
				}
				i = ev.erase(i);
                                break;

			default:
				i = ev.erase(i);
				break;
			}
		}
		g.add(ev);
	}
	g.add(m_airwaydb.find_by_area(bbox1, ~0, AirwaysDb::Airway::subtables_all));
	if (true) {
		{
			std::ostringstream oss;
			oss << "Airway area graph: [" << bbox1.get_southwest().get_lat_str2() << ' ' << bbox1.get_southwest().get_lon_str2()
			    << ' ' << bbox1.get_northeast().get_lat_str2() << ' ' << bbox1.get_northeast().get_lon_str2() << "], "
			    << boost::num_vertices(g) << " vertices, " << boost::num_edges(g) << " edges";
			m_signal_log(log_debug0, oss.str());
		}
		unsigned int counts[4];
		for (unsigned int i = 0; i < sizeof(counts)/sizeof(counts[0]); ++i)
			counts[i] = 0;
		Graph::vertex_iterator u0, u1;
		for (boost::tie(u0, u1) = boost::vertices(g); u0 != u1; ++u0) {
			const Vertex& uu(g[*u0]);
			unsigned int i(uu.get_type());
			if (i < sizeof(counts)/sizeof(counts[0]))
				++counts[i];
		}
		{
			std::ostringstream oss;
			oss << "Airway area graph: " << counts[Vertex::type_airport] << " airports, " << counts[Vertex::type_navaid] << " navaids, "
			    << counts[Vertex::type_intersection] << " intersections, " << counts[Vertex::type_mapelement] << " mapelements";
			m_signal_log(log_debug0, oss.str());
		}
	}
	typedef std::map<Graph::vertex_descriptor,unsigned int> vmap_t;
	vmap_t vmap;
	// create vertices
	Graph::vertex_iterator u0, u1;
	for (boost::tie(u0, u1) = boost::vertices(g); u0 != u1; ++u0) {
		const Vertex& uu(g[*u0]);
		if (uu.get_type() == Vertex::type_intersection && uu.is_ident_numeric())
			continue;
		Engine::DbObject::ptr_t obj(uu.get_object());
		if (!obj) {
			{
				WaypointsDb::Waypoint el;
				el.set_icao("");
				el.set_name(uu.get_ident());
				el.set_coord(uu.get_coord());
                                el.set_usage(WaypointsDb::Waypoint::usage_user);
                                el.set_type(WaypointsDb::Waypoint::type_other);
				Engine::DbObject::ptr_t p(Engine::DbObject::create(el));
				obj.swap(p);
			}
			if (true) {
				std::ostringstream oss;
				oss << "lgraphload: creating artificial object for vertex " << uu.get_ident();
				m_signal_log(log_debug0, oss.str());
			}
		}
		if (false && !boost::out_degree(*u0, g) && !boost::in_degree(*u0, g))
			continue;
		unsigned int super(lgraphaddvertex(obj, true, true));
		vmap.insert(std::make_pair(*u0, super));
	}
	// create edges
	Graph::edge_iterator e0, e1;
	for (boost::tie(e0, e1) = boost::edges(g); e0 != e1; ++e0) {
		Edge& ee(g[*e0]);
		// FIXME: once we have better base and top level information, this heuristic should be removed
		if (false) {
			const Glib::ustring& id(ee.get_ident());
			if (id.size() >= 3 && (id[0] == 'U' || id[0] == 'u') && std::isalpha(id[1]) && std::isdigit(id[2]))
				continue;
		}
		unsigned int superidx0, superidx1;
		{
			vmap_t::const_iterator i(vmap.find(boost::source(*e0, g)));
			if (i == vmap.end())
				continue;
			superidx0 = i->second;
			i = vmap.find(boost::target(*e0, g));
			if (i == vmap.end())
				continue;
			superidx1 = i->second;
		}
		lgraphairwayindex_t awyname(lgraphairway2index(ee.get_ident(), true));
		const LSuperVertex& super0(m_lsupervertices[superidx0]);
		const LSuperVertex& super1(m_lsupervertices[superidx1]);
		if (superidx0 == superidx1 || super0.get_v0() == super1.get_v0()) {
			std::ostringstream oss;
			oss << "Routing Graph: refusing to add self-edge " << super0.get_ident();
			m_signal_log(log_debug0, oss.str());
			continue;
		}
		int minaltitude(std::numeric_limits<int>::min());
		int maxaltitude(std::numeric_limits<int>::max());
		if (ee.get_corridor5_elev() != AirwaysDb::Airway::nodata) {
			minaltitude = ee.get_corridor5_elev();
			if (minaltitude > 5000)
				minaltitude += 1000;
			minaltitude += 1000;
		}
		int mindctaltitude(minaltitude);
		if (get_honour_awy_levels()) {
			int mi(ee.get_base_level()), ma(ee.get_top_level());
			mi *= 100;
			ma *= 100;
			minaltitude = std::max(minaltitude, mi);
			maxaltitude = std::min(maxaltitude, ma);
		}
		TrafficFlowRestrictions::DctParameters dctrule(super0.get_ident(), 
							       super0.get_coord(), super1.get_ident(), super1.get_coord(),
							       mindctaltitude, m_performance[m_performance.size()-1].get_altitude()+1);
		bool dctcalc(!get_tfr_enabled());
		for (unsigned int pi = 0; pi < m_performance.size(); ++pi) {
			const Performance& perf(m_performance[pi]);
			if (perf.get_altitude() < mindctaltitude)
				continue;
			//double dist(ee.get_dist());
			double dist(super0.get_coord().spheric_distance_nmi_dbl(super1.get_coord()));
			double x(dist * perf.get_secpernmi());
			if (is_optfuel())
				x *= perf.get_fuelpersec();
			LGraph::vertex_descriptor v0(super0.get_v0() + pi);
			LGraph::vertex_descriptor v1(super1.get_v0() + pi);
			if (v0 > super0.get_v1() || v1 > super1.get_v1())
				throw std::runtime_error("CFMUAutoroute::lgraphload: vertex error");
			bool outsideawyvert(perf.get_altitude() < minaltitude || perf.get_altitude() > maxaltitude);
			if (outsideawyvert || awyname == airwaydct) {
				if (!lgraphawyvertdct)
					continue;
				if (!dctcalc) {
					std::vector<TrafficFlowRestrictions::Message> msg;
					m_tfr.check_dct(msg, dctrule);
					for (unsigned int i = 0; i < msg.size(); ++i)
						logmessage(msg[i]);
					dctcalc = true;
				}
			}
			if (outsideawyvert) {
				if (!dctrule.get_alt().is_inside(0, perf.get_altitude()))
					continue;
				LGraph::edge_descriptor e;
				bool ok;
				tie(e, ok) = boost::edge(v0, v1, m_lgraph);
				if (ok)
					continue;
				x *= get_dctpenalty();
				boost::add_edge(v0, v1, LEdge(airwaydct, x), m_lgraph);
				continue;
			}
			if (awyname == airwaydct) {
				if (!dctrule.get_alt().is_inside(0, perf.get_altitude()))
					continue;
				LGraph::edge_descriptor e;
				bool ok;
				tie(e, ok) = boost::edge(v0, v1, m_lgraph);
				if (ok)
					continue;
				x *= get_dctpenalty();
			}
			LGraph::out_edge_iterator e0, e1;
			for (boost::tie(e0, e1) = boost::out_edges(v0, m_lgraph); e0 != e1; ++e0) {
				if (boost::target(*e0, m_lgraph) != v1)
					continue;
				LEdge& ee(m_lgraph[*e0]);
				if (ee.is_match(awyname)) {
					if (true) {
						std::ostringstream oss;
						oss << "Duplicate Airway Segment: " << lgraphvertexname(v0)
						    << ' ' << lgraphindex2airway(ee.get_ident(), true)
						    << ' ' << lgraphvertexname(v1);
						m_signal_log(log_debug0, oss.str());
					}
					break;
				}
				if (!ee.is_match(airwaydct))
					continue;
				ee.set_ident(awyname);
				ee.set_metric(x);
				break;
			}
			if (e0 == e1)
				boost::add_edge(v0, v1, LEdge(awyname, x), m_lgraph);
		}
	}
	{
		std::ostringstream oss;
		oss << "Routing Graph after airway population: " << boost::num_vertices(m_lgraph) << " V, " << boost::num_edges(m_lgraph) << " E";
		m_signal_log(log_debug0, oss.str());
	}
	lgraphmodified();
	return true;
}

void CFMUAutoroute::lgraphkillinvalidsuper(void)
{
	bool work(false);
	for (LSuperVertices::const_iterator si(m_lsupervertices.begin()), se(m_lsupervertices.end()); si != se; ++si) {
		const LSuperVertex& super(*si);
		bool valid(true);
		if (!super.is_ident_valid())
			valid = false;
		// see TrafficFlowRestrictions::load_airway_graph
		if (valid) {
			NavaidsDb::Navaid el;
			if (super.get(el)) {
				switch (el.get_navaid_type()) {
				case NavaidsDb::Navaid::navaid_vor:
				case NavaidsDb::Navaid::navaid_vordme:
				case NavaidsDb::Navaid::navaid_ndb:
				case NavaidsDb::Navaid::navaid_ndbdme:
				case NavaidsDb::Navaid::navaid_vortac:
				case NavaidsDb::Navaid::navaid_dme:
					break;

				default:
					valid = false;
					break;
				}
			}
		}
		if (valid) {
			WaypointsDb::Waypoint el;
			if (super.get(el)) {
				switch (el.get_type()) {
				case WaypointsDb::Waypoint::type_icao:
					break;

				case WaypointsDb::Waypoint::type_other:
					valid = el.get_icao().size() == 5;
					break;

				default:
					valid = false;
					break;
				}
			}
		}
		if (valid)
			continue;
		for (LGraph::vertex_descriptor v0(super.get_v0()); v0 <= super.get_v1(); ++v0) {
			work = work || boost::out_degree(v0, m_lgraph) || boost::in_degree(v0, m_lgraph);
			LGraph::in_edge_iterator e0, z0;
			for(boost::tie(e0, z0) = boost::in_edges(v0, m_lgraph); e0 != z0; ++e0) {
				if (boost::target(*e0, m_lgraph) != v0 || boost::source(*e0, m_lgraph) == v0)
					continue;
				const LEdge& ee0(m_lgraph[*e0]);
				if (!ee0.is_match(airwaymatchawy))
					continue;
				LGraph::out_edge_iterator e1, z1;
				for(boost::tie(e1, z1) = boost::out_edges(v0, m_lgraph); e1 != z1; ++e1) {
					if (boost::source(*e1, m_lgraph) != v0 || boost::target(*e1, m_lgraph) == v0)
						continue;
					const LEdge& ee1(m_lgraph[*e1]);
					if (!ee1.is_match(ee0.get_ident()))
						continue;
					if (boost::source(*e0, m_lgraph) == boost::target(*e1, m_lgraph))
						continue;
					boost::add_edge(boost::source(*e0, m_lgraph), boost::target(*e1, m_lgraph),
							LEdge(ee0.get_ident(), ee0.get_metric() + ee1.get_metric()), m_lgraph);
					if (true) {
						std::ostringstream oss;
						oss << "lgraphkillinvalidsuper: avoiding " << lgraphvertexname(v0)
						    << ": " << lgraphvertexname(boost::source(*e0, m_lgraph))
						    << ' ' << lgraphindex2airway(ee0.get_ident())
						    << ' ' << lgraphvertexname(boost::target(*e1, m_lgraph))
						    << " metric " << ee0.get_metric() + ee1.get_metric();
						m_signal_log(log_debug0, oss.str());
					}
				}
			}
			boost::clear_vertex(v0, m_lgraph);
		}
	}
	if (!work)
		return;
	{
		std::ostringstream oss;
		oss << "Routing Graph after invalid vertex removal: " << boost::num_vertices(m_lgraph) << " V, " << boost::num_edges(m_lgraph) << " E";
		m_signal_log(log_debug0, oss.str());
	}
	lgraphmodified();
}

bool CFMUAutoroute::lgraphcheckintersect(const Point& pt0, const Point& pt1, const Rect& bbox)
{
	return bbox.is_inside(pt0) || bbox.is_inside(pt1) || bbox.is_intersect(pt0, pt1);
}

bool CFMUAutoroute::lgraphcheckintersect(const Point& pt0, const Point& pt1, const Rect& bbox, const MultiPolygonHole& poly)
{
	{
		Rect r(pt0, pt0);
		r.add(pt1);
		if (!r.is_intersect(bbox))
			return false;
	}
	return ((bbox.is_inside(pt0) && poly.windingnumber(pt0)) ||
		(bbox.is_inside(pt1) && poly.windingnumber(pt1)) ||
		poly.is_intersection(pt0, pt1));
}

void CFMUAutoroute::lgraphexcluderegions(void)
{
	bool work(false);
	for (excluderegions_t::const_iterator ei(m_excluderegions.begin()), ee(m_excluderegions.end()); ei != ee; ++ei) {
		AirspacesDb::Airspace aspc;
		Rect bbox(ei->get_bbox());
		if (ei->is_airspace()) {
			AirspacesDb::elementvector_t ev(m_airspacedb.find_by_icao(ei->get_airspace_id(), 0, AirspacesDb::comp_exact, 0, AirspacesDb::Airspace::subtables_all));
			int eadidx(-1);
			for (unsigned int evi = 0; evi < ev.size(); ) {
				if (!ev[evi].is_valid() || ev[evi].get_class_string() != ei->get_airspace_type()) {
					ev.erase(ev.begin() + evi);
					continue;
				}
				if (ev[evi].get_typecode() == AirspacesDb::Airspace::typecode_ead && eadidx == -1)
					eadidx = evi;
				++evi;
			}
			if (eadidx == -1)
				eadidx = 0;
			if (ev.empty()) {
				std::ostringstream oss;
				oss << "Exclude Regions: Airspace " << ei->get_airspace_id() << '/' << ei->get_airspace_type()
				    << " not found";
				m_signal_log(log_debug0, oss.str());
				continue;
			}
			aspc = ev[eadidx];
			bbox = aspc.get_bbox();
			if (ev.size() > 1) {
				std::ostringstream oss;
				oss << "Exclude Regions: Multiple Airspaces " << ei->get_airspace_id() << '/' << ei->get_airspace_type()
				    << " found";
				m_signal_log(log_debug0, oss.str());
			}
		}
		unsigned int pi0(~0U), pi1(~0U);
		for (unsigned int pi = 0; pi < m_performance.size(); ++pi) {
			const Performance& perf(m_performance[pi]);
			if (perf.get_altitude() < 100 * ei->get_minlevel() || perf.get_altitude() > 100 * ei->get_maxlevel())
				continue;
			if (pi0 == ~0U)
				pi0 = pi;
			pi1 = pi;
		}
		{
			std::ostringstream oss;
			oss << "Excluding ";
			if (aspc.is_valid())
				oss << "Airspace " << aspc.get_icao() << '/' << aspc.get_class_string();
			else
				oss << "Region " << bbox;
			if (pi0 != ~0U)
				oss << " F" << std::setw(3) << std::setfill('0') << (m_performance[pi0].get_altitude() / 100)
				    << "..F" << std::setw(3) << std::setfill('0') << ((m_performance[pi1].get_altitude() + 99) / 100);
			m_signal_log(log_debug0, oss.str());
		}
		if (pi0 == ~0U)
			continue;
		for (LSuperVertices::const_iterator s0i(m_lsupervertices.begin()), s0e(m_lsupervertices.end()); s0i != s0e; ++s0i) {
			const LSuperVertex& super0(*s0i);
			if (super0.get_v0() == super0.get_v1())
				continue;
			for (LSuperVertices::const_iterator s1i(s0i + 1); s1i != s0e; ++s1i) {
				const LSuperVertex& super1(*s1i);
				if (super1.get_v0() == super1.get_v1())
					continue;
				int inters(-1);
				for (unsigned int pi = pi0; pi <= pi1; ++pi) {
					for (;;) {
						bool done(true);
						LGraph::out_edge_iterator e0, e1;
						for (tie(e0, e1) = out_edges(super0.get_v0() + pi, m_lgraph); e0 != e1; ) {
							LGraph::out_edge_iterator ex(e0);
							++e0;
							if (target(*ex, m_lgraph) != super1.get_v0() + pi)
								continue;
							if (inters == -1) {
								if (aspc.is_valid())
									inters = lgraphcheckintersect(super0.get_coord(), super1.get_coord(), bbox,
												      aspc.get_polygon());
								else
									inters = lgraphcheckintersect(super0.get_coord(), super1.get_coord(), bbox);
								if (!inters)
									break;
							}
							remove_edge(ex, m_lgraph);
							work = true;
							if (lgraphrestartedgescan) {
								done = false;
								break;
							}
						}
						if (done)
							break;
					}
					if (!inters)
						break;
					for (;;) {
						bool done(true);
						LGraph::out_edge_iterator e0, e1;
						for (tie(e0, e1) = out_edges(super1.get_v0() + pi, m_lgraph); e0 != e1; ) {
							LGraph::out_edge_iterator ex(e0);
							++e0;
							if (target(*ex, m_lgraph) != super0.get_v0() + pi)
								continue;
							if (inters == -1) {
								if (aspc.is_valid())
									inters = lgraphcheckintersect(super0.get_coord(), super1.get_coord(), bbox,
												      aspc.get_polygon());
								else
									inters = lgraphcheckintersect(super0.get_coord(), super1.get_coord(), bbox);
								if (!inters)
									break;
							}
							remove_edge(ex, m_lgraph);
							work = true;
							if (lgraphrestartedgescan) {
								done = false;
								break;
							}
						}
						if (done)
							break;
					}
				}
			}
		}
	}
	{
		std::ostringstream oss;
		oss << "Routing Graph after exclude regions: " << boost::num_vertices(m_lgraph) << " V, " << boost::num_edges(m_lgraph) << " E";
		m_signal_log(log_debug0, oss.str());
	}
	lgraphmodified();
}

#if 0

void CFMUAutoroute::lgraphadddct(void)
{
	if (get_dctlimit() <= 0)
		return;
	std::vector<std::vector<double> > awydist(boost::num_vertices(m_lgraph));
	for (LGraph::vertex_descriptor i = 0; i < boost::num_vertices(m_lgraph); ++i)
		awydist[i].resize(boost::num_vertices(m_lgraph), 0.0);
	if (!boost::johnson_all_pairs_shortest_paths(m_lgraph, awydist, boost::weight_map(boost::get(&LEdge::m_metric, m_lgraph)))) {
	//if (!boost::floyd_warshall_all_pairs_shortest_paths(m_lgraph, awydist, boost::weight_map(boost::get(&LEdge::m_metric, m_lgraph)))) {
		m_signal_log(log_debug0, "DCT: airway distances contain negative cycles, ignoring");
		for (LGraph::vertex_descriptor i = 0; i < boost::num_vertices(m_lgraph); ++i)
			for (LGraph::vertex_descriptor j = 0; j < boost::num_vertices(m_lgraph); ++j)
				awydist[i][j] = std::numeric_limits<double>::max();
	} else if (false) {
		for (LGraph::vertex_descriptor i = 0; i < boost::num_vertices(m_lgraph); ++i)
			for (LGraph::vertex_descriptor j = 0; j < boost::num_vertices(m_lgraph); ++j) {
				double d(awydist[i][j]);
				if (d == std::numeric_limits<double>::max())
					continue;
				const LVertex& v0(m_lgraph[i]);
				const LVertex& v1(m_lgraph[j]);
				const LSuperVertex& super0(m_lsupervertices[v0.get_superindex()]);
				const LSuperVertex& super1(m_lsupervertices[v1.get_superindex()]);
				std::ostringstream oss;
				oss << "DCT: AWY Distance " << super0.get_ident() << "->" << super1.get_ident()
				    << " metric " << std::setprecision(1);
				if (d < 1000.0)
					oss << std::fixed;
				oss << d << "nmi";
				m_signal_log(log_debug0, oss.str());
			}
	}
	typedef std::set<LEdgeAdd> edgeadd_t;
	edgeadd_t ea;
	unsigned int count_distpair(0);
	unsigned int count_awy(0);
	for (LSuperVertices::const_iterator s0i(m_lsupervertices.begin()), s0e(m_lsupervertices.end()); s0i != s0e; ++s0i) {
		const LSuperVertex& super0(*s0i);
		if (!super0.is_ident_valid())
			continue;
		for (LSuperVertices::const_iterator s1i(s0i + 1); s1i != s0e; ++s1i) {
			const LSuperVertex& super1(*s1i);
			if (!super1.is_ident_valid())
				continue;
			double dist(super0.get_coord().spheric_distance_nmi_dbl(super1.get_coord()));
			if (dist > get_dctlimit())
				continue;
			if (super0.get_ident() == super1.get_ident())
				continue;
			++count_distpair;
			int minalt(0);
			if (true) {
				TopoDb30::Profile prof(m_topodb.get_profile(super0.get_coord(), super1.get_coord(), 5));
                                if (prof.empty()) {
                                        std::ostringstream oss;
                                        oss << "DCT: Error computing ground clearance: " << super0.get_ident() << ' '
					    << super1.get_ident() << " dist " << std::setprecision(1) << dist << "nmi";
					m_signal_log(log_debug0, oss.str());
				} else {
					TopoDb30::elev_t el(prof.get_maxelev());
					if (el != TopoDb30::nodata) {
						minalt = prof.get_maxelev() * Point::m_to_ft;
						if (minalt >= 5000)
							minalt += 1000;
						minalt += 1000;
					}
				}
			}
			for (unsigned int pi = 0; pi < m_performance.size(); ++pi) {
				const Performance& perf(m_performance[pi]);
				if (perf.get_altitude() < minalt)
					continue;
				double x(dist * perf.get_secpernmi() * get_dctpenalty());
				if (is_optfuel())
					x *= perf.get_fuelpersec();
				LGraph::vertex_descriptor v0(super0.get_v0() + pi);
				LGraph::vertex_descriptor v1(super1.get_v0() + pi);
				if (v0 > super0.get_v1() || v1 > super1.get_v1())
					throw std::runtime_error("CFMUAutoroute::lgraphadddct: vertex error");
				// omit DCT if the airway distance is not (significantly) longer
				// i.e. prefer airways
				// recompute airway distance
				if (false && awydist[v0][v1] == std::numeric_limits<double>::max()) {
					std::vector<LGraph::vertex_descriptor> predecessors(boost::num_vertices(m_lgraph), 0);
					std::vector<double> distances(boost::num_vertices(m_lgraph), 0);
					dijkstra_shortest_paths(m_lgraph, v0, boost::weight_map(boost::get(&LEdge::m_metric, m_lgraph)).
								predecessor_map(&predecessors[0]).distance_map(&distances[0]));
					if (distances[v1] == std::numeric_limits<double>::max()) {
						double m(0);
						LGraph::vertex_descriptor v(v1);
						while (v != v0) {
							LGraph::vertex_descriptor u(predecessors[v]);
							if (u == v) {
								m = std::numeric_limits<double>::max();
								break;
							}
							LGraph::edge_descriptor e;
							bool ok(false);
							double bestm(std::numeric_limits<double>::max());
							LGraph::out_edge_iterator e0, e1;
							for (boost::tie(e0, e1) = boost::out_edges(u, m_lgraph); e0 != e1; ++e0) {
								if (boost::target(*e0, m_lgraph) != v)
									continue;
								const LEdge& ee(m_lgraph[*e0]);
								if (std::isnan(ee.get_metric()) || std::isinf(ee.get_metric()))
									continue;
								if (ee.get_metric() >= bestm)
									continue;
								bestm = ee.get_metric();
								ok = true;
								e = *e0;
							}
							if (!ok) {
								m = std::numeric_limits<double>::max();
								break;
							}
							m += bestm;
							v = u;
						}
						distances[v1] = m;
					}
					if (distances[v1] != std::numeric_limits<double>::max()) {
						awydist[v0][v1] = distances[v1];
						if (true) {
							Glib::ustring path;
							{
								LGraph::vertex_descriptor v(v1);
								while (v != v0) {
									LGraph::vertex_descriptor u(predecessors[v]);
									if (u == v) {
										path.clear();
										break;
									}
									path = "->" + lgraphvertexname(v) + path;
								}
								path = lgraphvertexname(v0) + path;
							}
							std::ostringstream oss;
							oss << "DCT: " << super0.get_ident() << "->" << super1.get_ident()
							    << " johnson metric error " << awydist[v0][v1] << " path " << path;
							m_signal_log(log_debug0, oss.str());
						}
					}
				}
				if (false && awydist[v1][v0] == std::numeric_limits<double>::max()) {
					std::vector<LGraph::vertex_descriptor> predecessors(boost::num_vertices(m_lgraph), 0);
					std::vector<double> distances(boost::num_vertices(m_lgraph), 0);
					dijkstra_shortest_paths(m_lgraph, v1, boost::weight_map(boost::get(&LEdge::m_metric, m_lgraph)).
								predecessor_map(&predecessors[0]).distance_map(&distances[0]));
					if (distances[v0] == std::numeric_limits<double>::max()) {
						double m(0);
						LGraph::vertex_descriptor v(v0);
						while (v != v1) {
							LGraph::vertex_descriptor u(predecessors[v]);
							if (u == v) {
								m = std::numeric_limits<double>::max();
								break;
							}
							LGraph::edge_descriptor e;
							bool ok(false);
							double bestm(std::numeric_limits<double>::max());
							LGraph::out_edge_iterator e0, e1;
							for (boost::tie(e0, e1) = boost::out_edges(u, m_lgraph); e0 != e1; ++e0) {
								if (boost::target(*e0, m_lgraph) != v)
									continue;
								const LEdge& ee(m_lgraph[*e0]);
								if (std::isnan(ee.get_metric()) || std::isinf(ee.get_metric()))
									continue;
								if (ee.get_metric() >= bestm)
									continue;
								bestm = ee.get_metric();
								ok = true;
								e = *e0;
							}
							if (!ok) {
								m = std::numeric_limits<double>::max();
								break;
							}
							m += bestm;
							v = u;
						}
						distances[v0] = m;
					}
					if (distances[v0] != std::numeric_limits<double>::max()) {
						awydist[v1][v0] = distances[v0];
						if (true) {
							Glib::ustring path;
							{
								LGraph::vertex_descriptor v(v0);
								while (v != v1) {
									LGraph::vertex_descriptor u(predecessors[v]);
									if (u == v) {
										path.clear();
										break;
									}
									path = "->" + lgraphvertexname(v) + path;
								}
								path = lgraphvertexname(v0) + path;
							}
							std::ostringstream oss;
							oss << "DCT: " << super1.get_ident() << "->" << super0.get_ident()
							    << " johnson metric error " << awydist[v1][v0] << " path " << path;
							m_signal_log(log_debug0, oss.str());
						}
					}
				}
				LGraph::edge_descriptor e;
				bool ok;
				tie(e, ok) = boost::edge(v0, v1, m_lgraph);
				if (!ok) {
					if (awydist[v0][v1] <= 1.01 * x)
						++count_awy;
				        else
						ea.insert(LEdgeAdd(v0, v1, x));
					if (false && count_distpair < 50 && pi + 1 == m_performance.size()) {
						std::vector<LGraph::vertex_descriptor> predecessors(boost::num_vertices(m_lgraph), 0);
						dijkstra_shortest_paths(m_lgraph, v0, boost::weight_map(boost::get(&LEdge::m_metric, m_lgraph)).
									predecessor_map(&predecessors[0]));
						double distd(0);
						LGraph::vertex_descriptor v(v1);
						while (v != v0) {
							LGraph::vertex_descriptor u(predecessors[v]);
							if (u == v) {
								distd = std::numeric_limits<double>::max();
								break;
							}
							LGraph::edge_descriptor e;
							bool ok(false);
							double bestm(std::numeric_limits<double>::max());
							LGraph::out_edge_iterator e0, e1;
							for (boost::tie(e0, e1) = boost::out_edges(u, m_lgraph); e0 != e1; ++e0) {
								if (boost::target(*e0, m_lgraph) != v)
									continue;
								const LEdge& ee(m_lgraph[*e0]);
								if (std::isnan(ee.get_metric()) || std::isinf(ee.get_metric()))
									continue;
								if (ee.get_metric() >= bestm)
									continue;
								bestm = ee.get_metric();
								ok = true;
								e = *e0;
							}
							if (!ok) {
								distd = std::numeric_limits<double>::max();
								break;
							}
							distd += bestm;
							v = u;
						}
						std::ostringstream oss;
						oss << "DCT: " << super0.get_ident() << "->" << super1.get_ident()
						    << " F" << ((perf.get_altitude() + 50) / 100) << ": DCT D "
						    << std::setprecision(1) << dist << " M "
						    << std::setprecision(1) << x
						    << " AWY D "
						    << std::setprecision(1) << distd << " M "
						    << std::setprecision(1) << awydist[v0][v1];
						m_signal_log(log_debug0, oss.str());
					}
				}
				tie(e, ok) = boost::edge(v1, v0, m_lgraph);
				if (!ok) {
					if (awydist[v1][v0] <= 1.01 * x)
						++count_awy;
					else
						ea.insert(LEdgeAdd(v1, v0, x));
					if (false && count_distpair < 50 && pi + 1 == m_performance.size()) {
						std::vector<LGraph::vertex_descriptor> predecessors(boost::num_vertices(m_lgraph), 0);
						dijkstra_shortest_paths(m_lgraph, v1, boost::weight_map(boost::get(&LEdge::m_metric, m_lgraph)).
									predecessor_map(&predecessors[0]));
						double distd(0);
						LGraph::vertex_descriptor v(v0);
						while (v != v1) {
							LGraph::vertex_descriptor u(predecessors[v]);
							if (u == v) {
								distd = std::numeric_limits<double>::max();
								break;
							}
							LGraph::edge_descriptor e;
							bool ok(false);
							double bestm(std::numeric_limits<double>::max());
							LGraph::out_edge_iterator e0, e1;
							for (boost::tie(e0, e1) = boost::out_edges(u, m_lgraph); e0 != e1; ++e0) {
								if (boost::target(*e0, m_lgraph) != v)
									continue;
								const LEdge& ee(m_lgraph[*e0]);
								if (std::isnan(ee.get_metric()) || std::isinf(ee.get_metric()))
									continue;
								if (ee.get_metric() >= bestm)
									continue;
								bestm = ee.get_metric();
								ok = true;
								e = *e0;
							}
							if (!ok) {
								distd = std::numeric_limits<double>::max();
								break;
							}
							distd += bestm;
							v = u;
						}
						std::ostringstream oss;
						oss << "DCT: " << super1.get_ident() << "->" << super0.get_ident()
						    << " F" << ((perf.get_altitude() + 50) / 100) << ": DCT D "
						    << std::setprecision(1) << dist << " M "
						    << std::setprecision(1) << x << " AWY D "
						    << std::setprecision(1) << distd << " M "
						    << std::setprecision(1) << awydist[v1][v0];
						m_signal_log(log_debug0, oss.str());
					}
				}
			}
		}
	}
	unsigned int count_dct(0);
	for (edgeadd_t::const_iterator ie(ea.begin()), ee(ea.end()); ie != ee; ++ie) {
		LGraph::edge_descriptor e;
		bool ok;
		tie(e, ok) = boost::add_edge(ie->get_v0(), ie->get_v1(), LEdge(airwaydct, ie->get_metric()), m_lgraph);
		if (ok)
			++count_dct;
	}
	{
		std::ostringstream oss;
		oss << "Routing Graph after DCT population: " << boost::num_vertices(m_lgraph) << " V, " << boost::num_edges(m_lgraph) << " E, "
		    << count_distpair << " vertex pairs, " << count_dct << " DCT edges, " << count_awy << " AWY edges preferred";
		m_signal_log(log_debug0, oss.str());
	}
	lgraphmodified();
}

#elif 0

template <class T>
struct DijkstraCombineLimit
{
	const T m_limit;
	const T m_inf;

	DijkstraCombineLimit() : m_limit(std::numeric_limits<T>::max()), m_inf(std::numeric_limits<T>::max()) { }
	DijkstraCombineLimit(T lim) : m_limit(lim), m_inf(std::numeric_limits<T>::max()) { }
	DijkstraCombineLimit(T lim, T inf) : m_limit(lim), m_inf(inf) { }

	T operator()(const T& a, const T& b) const
	{
		if (a == m_inf)
			return m_inf;
		if (b == m_inf)
			return m_inf;
		double z(a + b);
		if (z > m_limit)
			return m_inf;
		return z;
	}
};

void CFMUAutoroute::lgraphadddct(void)
{
	if (get_dctlimit() <= 0)
		return;
	typedef std::set<LEdgeAdd> edgeadd_t;
	edgeadd_t ea;
	unsigned int count_distpair(0);
	unsigned int count_awy(0);
	for (LSuperVertices::const_iterator s0i(m_lsupervertices.begin()), s0e(m_lsupervertices.end()); s0i != s0e; ++s0i) {
		const LSuperVertex& super0(*s0i);
		if (!super0.is_ident_valid())
			continue;
		for (LSuperVertices::const_iterator s1i(s0i + 1); s1i != s0e; ++s1i) {
			const LSuperVertex& super1(*s1i);
			if (!super1.is_ident_valid())
				continue;
			double dist(super0.get_coord().spheric_distance_nmi_dbl(super1.get_coord()));
			if (dist > get_dctlimit())
				continue;
			if (super0.get_ident() == super1.get_ident())
				continue;
			++count_distpair;
			if (true && !(count_distpair & ((1 << 10) - 1))) {
				std::ostringstream oss;
				oss << "DCT: processed " << count_distpair << " pairs";
				m_signal_log(log_debug0, oss.str());
			}
			int minalt(0);
			if (true) {
				TopoDb30::Profile prof(m_topodb.get_profile(super0.get_coord(), super1.get_coord(), 5));
                                if (prof.empty()) {
                                        std::ostringstream oss;
                                        oss << "DCT: Error computing ground clearance: " << super0.get_ident() << ' '
					    << super1.get_ident() << " dist " << std::setprecision(1) << dist << "nmi";
					m_signal_log(log_debug0, oss.str());
				} else {
					TopoDb30::elev_t el(prof.get_maxelev());
					if (el != TopoDb30::nodata) {
						minalt = prof.get_maxelev() * Point::m_to_ft;
						if (minalt >= 5000)
							minalt += 1000;
						minalt += 1000;
					}
				}
			}
			for (unsigned int pi = 0; pi < m_performance.size(); ++pi) {
				const Performance& perf(m_performance[pi]);
				if (perf.get_altitude() < minalt)
					continue;
				double x(dist * perf.get_secpernmi() * get_dctpenalty());
				if (is_optfuel())
					x *= perf.get_fuelpersec();
				LGraph::vertex_descriptor v0(super0.get_v0() + pi);
				LGraph::vertex_descriptor v1(super1.get_v0() + pi);
				if (v0 > super0.get_v1() || v1 > super1.get_v1())
					throw std::runtime_error("CFMUAutoroute::lgraphadddct: vertex error");
				// omit DCT if the airway distance is not (significantly) longer
				// i.e. prefer airways
				LGraph::edge_descriptor e;
				bool ok;
				tie(e, ok) = boost::edge(v0, v1, m_lgraph);
				if (!ok) {
					std::vector<LGraph::vertex_descriptor> predecessors(boost::num_vertices(m_lgraph));
					std::vector<double> distances(boost::num_vertices(m_lgraph));
					DijkstraCombineLimit<double> cmb(1.1 * x);
					dijkstra_shortest_paths(m_lgraph, v0, boost::weight_map(boost::get(&LEdge::m_metric, m_lgraph)).
								predecessor_map(&predecessors[0]).distance_map(&distances[0]).
								distance_combine(cmb));
					if (distances[v1] <= 1.01 * x)
						++count_awy;
				        else
						ea.insert(LEdgeAdd(v0, v1, x));
				}
				tie(e, ok) = boost::edge(v1, v0, m_lgraph);
				if (!ok) {
					std::vector<LGraph::vertex_descriptor> predecessors(boost::num_vertices(m_lgraph));
					std::vector<double> distances(boost::num_vertices(m_lgraph));
					DijkstraCombineLimit<double> cmb(1.1 * x);
					dijkstra_shortest_paths(m_lgraph, v1, boost::weight_map(boost::get(&LEdge::m_metric, m_lgraph)).
								predecessor_map(&predecessors[0]).distance_map(&distances[0]).
								distance_combine(cmb));
					if (distances[v0] <= 1.01 * x)
						++count_awy;
					else
						ea.insert(LEdgeAdd(v1, v0, x));
				}
			}
		}
	}
	unsigned int count_dct(0);
	for (edgeadd_t::const_iterator ie(ea.begin()), ee(ea.end()); ie != ee; ++ie) {
		LGraph::edge_descriptor e;
		bool ok;
		tie(e, ok) = boost::add_edge(ie->get_v0(), ie->get_v1(), LEdge(airwaydct, ie->get_metric()), m_lgraph);
		if (ok)
			++count_dct;
	}
	{
		std::ostringstream oss;
		oss << "Routing Graph after DCT population: " << boost::num_vertices(m_lgraph) << " V, " << boost::num_edges(m_lgraph) << " E, "
		    << count_distpair << " vertex pairs, " << count_dct << " DCT edges, " << count_awy << " AWY edges preferred";
		m_signal_log(log_debug0, oss.str());
	}
	lgraphmodified();
}

#else

void CFMUAutoroute::lgraphadddct(void)
{
	if (get_dctlimit() <= 0)
		return;
	//const unsigned int awydistsz(m_lsupervertices.size());
	//double awydist[awydistsz][awydistsz];
	std::vector<std::vector<double> > awydist(m_lsupervertices.size());
	for (LGraph::vertex_descriptor i = 0; i < m_lsupervertices.size(); ++i)
		awydist[i].resize(m_lsupervertices.size(), 0.0);
	{
		typedef boost::adjacency_list<boost::listS, boost::vecS, boost::directedS, boost::no_property, boost::property<boost::edge_weight_t, double> > DGraph;
		DGraph dg(m_lsupervertices.size());
		boost::property_map<DGraph, boost::edge_weight_t>::type dgdistmap(boost::get(boost::edge_weight, dg));
		{
			LGraph::edge_iterator e0, e1;
			for (boost::tie(e0, e1) = boost::edges(m_lgraph); e0 != e1; ++e0) {
				const LVertex& uu(m_lgraph[boost::source(*e0, m_lgraph)]);
				const LVertex& vv(m_lgraph[boost::target(*e0, m_lgraph)]);
				DGraph::edge_descriptor e;
				bool ok;
				boost::tie(e, ok) = add_edge((DGraph::vertex_descriptor)uu.get_superindex(), (DGraph::vertex_descriptor)vv.get_superindex(), dg);
				if (!ok)
					continue;
				const LSuperVertex& superu(m_lsupervertices[uu.get_superindex()]);
				const LSuperVertex& superv(m_lsupervertices[vv.get_superindex()]);
				dgdistmap[e] = superu.get_coord().spheric_distance_nmi_dbl(superv.get_coord());
			}
		}
		if (!boost::johnson_all_pairs_shortest_paths(dg, awydist)) {
			m_signal_log(log_debug0, "DCT: airway distances contain negative cycles, ignoring");
			for (LGraph::vertex_descriptor i = 0; i < m_lsupervertices.size(); ++i)
				for (LGraph::vertex_descriptor j = 0; j < m_lsupervertices.size(); ++j)
					awydist[i][j] = std::numeric_limits<double>::max();
		}
	}
	unsigned int count_distpair(0);
	unsigned int count_awy(0);
	unsigned int count_dct(0);
	unsigned int count_dctrules(0);
	unsigned int count_uncached(0);
	unsigned int count_cached(0);

	Rect dctbbox(get_bbox().oversize_nmi(100.f));
	DctCache dctcache("", m_tfr.get_created());
	std::vector<DctCache::pointid_t> dctpointids;
	static const bool have_dctcache(true);
	if (have_dctcache) {
		Glib::TimeVal tv;
		tv.assign_current_time();
		dctcache.load(dctbbox);
		{
			Glib::TimeVal tv1;
			tv1.assign_current_time();
			tv = tv1 - tv;
		}
		{
			std::ostringstream oss;
			oss << "loaded DCT cache, " << dctcache.get_nrpoints() << " points, " << dctcache.get_nrdcts() << " DCTs, area ["
			    << dctbbox.get_southwest().get_lat_str2() << ' ' << dctbbox.get_southwest().get_lon_str2() << ' '
			    << dctbbox.get_northeast().get_lat_str2() << ' ' << dctbbox.get_northeast().get_lon_str2()
			    << "], " << std::fixed << std::setprecision(3) << tv.as_double() << 's';
			m_signal_log(log_normal, oss.str());
		}
		tv.assign_current_time();
		dctpointids.resize(m_lsupervertices.size(), 0);
		for (unsigned int superidx0(0), nrsuper(m_lsupervertices.size()); superidx0 < nrsuper; ++superidx0) {
			const LSuperVertex& super0(m_lsupervertices[superidx0]);
			if (!super0.is_ident_valid())
				continue;
			if (!dctbbox.is_inside(super0.get_coord()))
				continue;
			dctpointids[superidx0] = dctcache.find_point(super0.get_ident(), super0.get_coord());
		}
		{
			Glib::TimeVal tv1;
			tv1.assign_current_time();
			tv = tv1 - tv;
		}
		{
			std::ostringstream oss;
			oss << "allocated DCT cache points, " << dctcache.get_nrpoints() << " points, "
			    << std::fixed << std::setprecision(3) << tv.as_double() << 's';
			m_signal_log(log_normal, oss.str());
		}
	}

	for (unsigned int superidx0(0), nrsuper(m_lsupervertices.size()); superidx0 < nrsuper; ++superidx0) {
		const LSuperVertex& super0(m_lsupervertices[superidx0]);
		if (!super0.is_ident_valid())
			continue;
		if (!dctbbox.is_inside(super0.get_coord()))
			continue;
		for (unsigned int superidx1(superidx0 + 1); superidx1 < nrsuper; ++superidx1) {
			const LSuperVertex& super1(m_lsupervertices[superidx1]);
			if (!super1.is_ident_valid())
				continue;
			if (!dctbbox.is_inside(super1.get_coord()))
				continue;
			double dist(super0.get_coord().spheric_distance_nmi_dbl(super1.get_coord()));
			if (dist > get_dctlimit())
				continue;
			if (super0.get_ident() == super1.get_ident())
				continue;
			++count_distpair;
			if (true && !(count_distpair & ((1 << 12) - 1))) {
				std::ostringstream oss;
				oss << "DCT: processed " << count_distpair << " pairs";
				m_signal_log(log_debug0, oss.str());
			}
			if (dist * 1.01 > std::max(awydist[superidx0][superidx1], awydist[superidx1][superidx0]))
				continue;
			int minalt(0);
			TrafficFlowRestrictions::DctParameters dctrule(super0.get_ident(), super0.get_coord(), super1.get_ident(), super1.get_coord(),
								       m_performance[0].get_altitude(), m_performance[m_performance.size()-1].get_altitude()+1);
			bool dctcachefound(false);
			if (have_dctcache) {
				dctcachefound = true;
				const DctCache::DctAlt *a;
				a = dctcache.find_dct(dctpointids[superidx0], dctpointids[superidx1]);
				if (a)
					dctrule.get_alt(0) = a->get_alt();
				else
					dctcachefound = false;
				a = dctcache.find_dct(dctpointids[superidx1], dctpointids[superidx0]);
				if (a)
					dctrule.get_alt(1) = a->get_alt();
				else
					dctcachefound = false;
				if (dctcachefound)
					++count_cached;
				else
					++count_uncached;
			}
			if (true && !dctcachefound && get_tfr_enabled()) {
				std::vector<TrafficFlowRestrictions::Message> msg;
				m_tfr.check_dct(msg, dctrule, false && (super0.get_ident() == "BERSU" || super1.get_ident() == "BERSU"));
				for (unsigned int i = 0; i < msg.size(); ++i)
					logmessage(msg[i]);
				if (false && (super0.get_ident() == "ZUE" || super0.get_ident() == "ROMIR" || super0.get_ident() == "KUDIS" ||
					     super1.get_ident() == "ZUE" || super1.get_ident() == "ROMIR" || super1.get_ident() == "KUDIS")) {
					std::ostringstream oss;
                                        oss << "DCT: rules: " << super0.get_ident() << ' ' << super1.get_ident()
					    << " alt " << dctrule.get_alt(0).to_str() << " / " << dctrule.get_alt(1).to_str();
					if (dctrule.get_alt().is_empty())
						oss << " (E)";
					m_signal_log(log_debug0, oss.str());
				}
			}
			if (dctrule.get_alt().is_empty()) {
				if (have_dctcache && !dctcachefound) {
					dctcache.add_dct(dctpointids[superidx0], dctpointids[superidx1], dctrule.get_alt(0));
					dctcache.add_dct(dctpointids[superidx1], dctpointids[superidx0], dctrule.get_alt(1));
				}
				count_dctrules += 2 * m_performance.size();
				continue;
			}
			if (true && !dctcachefound) {
				TopoDb30::Profile prof(m_topodb.get_profile(super0.get_coord(), super1.get_coord(), 5));
                                if (prof.empty()) {
                                        std::ostringstream oss;
                                        oss << "DCT: Error computing ground clearance: " << super0.get_ident() << ' '
					    << super1.get_ident() << " dist " << std::setprecision(1) << dist << "nmi";
					m_signal_log(log_debug0, oss.str());
				} else {
					TopoDb30::elev_t el(prof.get_maxelev());
					if (el != TopoDb30::nodata) {
						minalt = prof.get_maxelev() * Point::m_to_ft;
						if (minalt >= 5000)
							minalt += 1000;
						minalt += 1000;
					}
				}
			}
			if (have_dctcache && !dctcachefound) {
				DctCache::set_t::Interval malt(minalt, std::numeric_limits<DctCache::set_t::type_t>::max());
				dctrule.get_alt(0) &= malt;
				dctrule.get_alt(1) &= malt;
				dctcache.add_dct(dctpointids[superidx0], dctpointids[superidx1], dctrule.get_alt(0));
				dctcache.add_dct(dctpointids[superidx1], dctpointids[superidx0], dctrule.get_alt(1));
			}
			for (unsigned int pi = 0; pi < m_performance.size(); ++pi) {
				const Performance& perf(m_performance[pi]);
				if (perf.get_altitude() < minalt)
					continue;
				double x(dist * perf.get_secpernmi() * get_dctpenalty());
				if (is_optfuel())
					x *= perf.get_fuelpersec();
				LGraph::vertex_descriptor v0(super0.get_v0() + pi);
				LGraph::vertex_descriptor v1(super1.get_v0() + pi);
				if (v0 > super0.get_v1() || v1 > super1.get_v1())
					throw std::runtime_error("CFMUAutoroute::lgraphadddct: vertex error");
				if (false && ((super0.get_ident() == "SOSAL" && super1.get_ident() == "RAMOK") ||
					      (super0.get_ident() == "RAMOK" && super1.get_ident() == "SOSAL"))) {
					std::ostringstream oss;
					oss << "DCT: " << super0.get_ident() << ' ' << super1.get_ident()
					    << " alt " << perf.get_altitude() << " dist " << dist << " awy "
					    << awydist[superidx0][superidx1] << " / " << awydist[superidx1][superidx0]
					    << " dctrule " << (int)dctrule.get_alt().is_inside(0, perf.get_altitude())
					    << " / " << (int)dctrule.get_alt().is_inside(1, perf.get_altitude())
					    << " metric " << x;
					m_signal_log(log_debug0, oss.str());
				}
				// omit DCT if the airway distance is not (significantly) longer
				// i.e. prefer airways
				LGraph::edge_descriptor e;
				bool ok;
				if (dctrule.get_alt().is_inside(0, perf.get_altitude())) {
					tie(e, ok) = boost::edge(v0, v1, m_lgraph);
					if (!ok) {
						if (awydist[superidx0][superidx1] <= 1.01 * dist) {
							++count_awy;
						} else {
							tie(e, ok) = boost::add_edge(v0, v1, LEdge(airwaydct, x), m_lgraph);
							if (ok)
								++count_dct;
						}
					}
				} else {
					++count_dctrules;
				}
				if (dctrule.get_alt().is_inside(1, perf.get_altitude())) {
					tie(e, ok) = boost::edge(v1, v0, m_lgraph);
					if (!ok) {
						if (awydist[superidx1][superidx0] <= 1.01 * dist) {
							++count_awy;
						} else {
							tie(e, ok) = boost::add_edge(v1, v0, LEdge(airwaydct, x), m_lgraph);
							if (ok)
								++count_dct;
						}
					}
				} else {
					++count_dctrules;
				}
			}
		}
	}
	{
		std::ostringstream oss;
		oss << "Routing Graph after DCT population: " << boost::num_vertices(m_lgraph) << " V, " << boost::num_edges(m_lgraph) << " E, "
		    << count_distpair << " vertex pairs, " << count_dct << " DCT edges, " << count_awy << " AWY edges preferred, "
		    << count_dctrules << " DCT edges prevented by TFR rules";
		m_signal_log(log_debug0, oss.str());
	}
	if (have_dctcache) {
		Glib::TimeVal tv;
		tv.assign_current_time();
		dctcache.flush();
		{
			Glib::TimeVal tv1;
			tv1.assign_current_time();
			tv = tv1 - tv;
		}
		std::ostringstream oss;
		oss << "DCT cache flush: " << count_uncached << " uncached, " << count_cached << " cached, "
		    << dctcache.get_nrpoints() << " points, " << dctcache.get_nrdcts() << " DCTs, "
		    << std::fixed << std::setprecision(3) << tv.as_double() << 's';
		m_signal_log(log_debug0, oss.str());
	}
	lgraphmodified();
}

#endif

bool CFMUAutoroute::lgraphaddsidstar(void)
{
	if (!get_departure().is_valid() || !get_destination().is_valid())
		return false;
	static const bool duplicate_sidstar_nodes = true;
	unsigned int lastsuper(m_lsupervertices.size());
	m_superdep = lgraphaddvertex(Engine::DbObject::create(get_departure()), false, false);
	m_superdest = lgraphaddvertex(Engine::DbObject::create(get_destination()), false, false);
	const LSuperVertex& sdep(m_lsupervertices[m_superdep]);
	const LSuperVertex& sdest(m_lsupervertices[m_superdest]);
	typedef std::vector<unsigned int> sv_t;
	sv_t sidv, starv;
	{
		unsigned int sidsv(lastsuper), starsv(lastsuper);
		double sidd(std::numeric_limits<double>::max()), stard(sidd);
		sv_t sida, stara;
		for (unsigned int s0i(0); s0i < lastsuper; ++s0i) {
			const LSuperVertex& super0(m_lsupervertices[s0i]);
			double dist(super0.get_coord().spheric_distance_nmi_dbl(sdep.get_coord()));
			if (dist <= get_sidlimit())
				sidv.push_back(s0i);
			dist = super0.get_coord().spheric_distance_nmi_dbl(sdest.get_coord());
			if (dist <= get_starlimit())
				starv.push_back(s0i);
			if (!get_sid().is_invalid()) {
				dist = super0.get_coord().spheric_distance_nmi_dbl(get_sid());
				if (dist < sidd) {
					sidd = dist;
					sidsv = s0i;
				}
			}
			if (!get_star().is_invalid()) {
				dist = super0.get_coord().spheric_distance_nmi_dbl(get_star());
				if (dist < stard) {
					stard = dist;
					starsv = s0i;
				}
			}
			for (unsigned int i = 0, j = get_departure().get_nrvfrrte(); i < j; ++i) {
				const AirportsDb::Airport::VFRRoute& rte(get_departure().get_vfrrte(i));
				if (!rte.size())
					continue;
				const AirportsDb::Airport::VFRRoute::VFRRoutePoint& rtept(rte[rte.size()-1]);
				if (rtept.get_pathcode() != AirportsDb::Airport::VFRRoute::VFRRoutePoint::pathcode_sid)
					continue;
				if (super0.get_coord().spheric_distance_nmi_dbl(rtept.get_coord()) > 0.1)
					continue;
				sida.push_back(s0i);
				break;
			}
			for (unsigned int i = 0, j = get_destination().get_nrvfrrte(); i < j; ++i) {
				const AirportsDb::Airport::VFRRoute& rte(get_destination().get_vfrrte(i));
				if (!rte.size())
					continue;
				const AirportsDb::Airport::VFRRoute::VFRRoutePoint& rtept(rte[0]);
				if (rtept.get_pathcode() != AirportsDb::Airport::VFRRoute::VFRRoutePoint::pathcode_star)
					continue;
				if (super0.get_coord().spheric_distance_nmi_dbl(rtept.get_coord()) > 0.1)
					continue;
				stara.push_back(s0i);
				break;
			}
		}
		if (!sida.empty())
			sida.swap(sidv);
		if (!stara.empty())
			stara.swap(starv);
		if (sidd > 0.1)
			sidsv = lastsuper;
		if (stard > 0.1)
			starsv = lastsuper;
		if (sidsv != lastsuper) {
			sidv.clear();
			sidv.push_back(sidsv);
		} else if (!get_sid().is_invalid()) {
			if (true) {
				std::ostringstream oss;
				oss << "IFR: SID point " << get_sid().get_lat_str2() << ' ' << get_sid().get_lon_str2() << " not found";
				m_signal_log(log_debug0, oss.str());
			}
			return false;
		}
		if (starsv != lastsuper) {
			starv.clear();
			starv.push_back(starsv);
		} else if (!get_star().is_invalid()) {
			if (true) {
				std::ostringstream oss;
				oss << "IFR: STAR point " << get_star().get_lat_str2() << ' ' << get_star().get_lon_str2() << " not found";
				m_signal_log(log_debug0, oss.str());
			}
			return false;
		}
	}
	typedef std::map<unsigned int, unsigned int> nodemap_t;
	nodemap_t nodemap;
	if (duplicate_sidstar_nodes) {
		for (sv_t::const_iterator svi(sidv.begin()), sve(sidv.end()); svi != sve; ++svi) {
			unsigned int superidx(*svi);
			if (nodemap.find(superidx) != nodemap.end())
				continue;
			const LSuperVertex& super(m_lsupervertices[superidx]);
			unsigned int newsuperidx(lgraphaddvertex(super.get_object(), true, false));
			nodemap.insert(nodemap_t::value_type(superidx, newsuperidx));
		}
		for (sv_t::const_iterator svi(starv.begin()), sve(starv.end()); svi != sve; ++svi) {
			unsigned int superidx(*svi);
			if (nodemap.find(superidx) != nodemap.end())
				continue;
			const LSuperVertex& super(m_lsupervertices[superidx]);
			unsigned int newsuperidx(lgraphaddvertex(super.get_object(), true, false));
			nodemap.insert(nodemap_t::value_type(superidx, newsuperidx));
		}
		for (nodemap_t::const_iterator ni1(nodemap.begin()), ne(nodemap.end()); ni1 != ne; ++ni1) {
			const LSuperVertex& super(m_lsupervertices[ni1->first]);
			const LSuperVertex& super0(m_lsupervertices[ni1->second]);
			for (unsigned int pi = 0; pi < m_performance.size(); ++pi) {
				LGraph::out_edge_iterator e0, e1;
				for (boost::tie(e0, e1) = boost::out_edges(super.get_v0() + pi, m_lgraph); e0 != e1; ++e0) {
					const LEdge& e(m_lgraph[*e0]);
					if (e.is_match(airwaymatchlevel))
						continue;
					const LVertex& vv1(m_lgraph[boost::target(*e0, m_lgraph)]);
					if (vv1.get_superindex() == ni1->second)
						continue;
					nodemap_t::const_iterator ni2(nodemap.find(vv1.get_superindex()));
					if (ni2 == ne || ni2 == ni1)
						continue;
					const LSuperVertex& super1(m_lsupervertices[ni2->second]);
					if (super1.get_v0() == super1.get_v1())
						continue;
					boost::add_edge(super0.get_v0() + pi, super1.get_v0() + pi, e, m_lgraph);
					if (true) {
						std::ostringstream oss;
						oss << "SID/STAR: add direct route " << lgraphvertexname(super0.get_v0() + pi) << ' '
						    << lgraphindex2airway(e.get_ident(), true) << ' ' << lgraphvertexname(super1.get_v0() + pi);
						m_signal_log(log_debug0, oss.str());
					}
				}
			}
		}
	}
	for (sv_t::const_iterator svi(sidv.begin()), sve(sidv.end()); svi != sve; ++svi) {
		unsigned int superidx(*svi);
		{
			nodemap_t::const_iterator i(nodemap.find(superidx));
			if (i != nodemap.end()) {
				const LSuperVertex& super(m_lsupervertices[superidx]);
				unsigned int newsuperidx(i->second);
				const LSuperVertex& newsuper(m_lsupervertices[newsuperidx]);
				for (unsigned int pi = 0; pi < m_performance.size(); ++pi) {
					LGraph::out_edge_iterator e0, e1;
					for (boost::tie(e0, e1) = boost::out_edges(super.get_v0() + pi, m_lgraph); e0 != e1; ++e0) {
						const LEdge& e(m_lgraph[*e0]);
						if (e.is_match(airwaymatchlevel))
							continue;
						boost::add_edge(newsuper.get_v0() + pi, boost::target(*e0, m_lgraph), e, m_lgraph);
					}
				}
				superidx = newsuperidx;
			}
		}
		const LSuperVertex& super(m_lsupervertices[superidx]);
		const Performance& perf(m_performance[0]);
		double x(super.get_coord().spheric_distance_nmi_dbl(sdep.get_coord()) * perf.get_secpernmi());
		if (is_optfuel())
			x *= perf.get_fuelpersec();
		for (unsigned int pi = 0; pi < m_performance.size(); ++pi) {
			const Performance& perf(m_performance[pi]);
			boost::add_edge(sdep.get_v0(), super.get_v0() + pi, LEdge(airwaysid, x), m_lgraph);
			if (is_optfuel())
				x += perf.get_fuelpenalty();
			else
				x += perf.get_timepenalty();
		}
		lgraphkilllevelchg(super);
	}
	for (sv_t::const_iterator svi(starv.begin()), sve(starv.end()); svi != sve; ++svi) {
		unsigned int superidx(*svi);
		{
			nodemap_t::const_iterator i(nodemap.find(superidx));
			if (i != nodemap.end()) {
				const LSuperVertex& super(m_lsupervertices[superidx]);
				unsigned int newsuperidx(i->second);
				const LSuperVertex& newsuper(m_lsupervertices[newsuperidx]);
				for (unsigned int pi = 0; pi < m_performance.size(); ++pi) {
					LGraph::in_edge_iterator e0, e1;
					for (boost::tie(e0, e1) = boost::in_edges(super.get_v0() + pi, m_lgraph); e0 != e1; ++e0) {
						const LEdge& e(m_lgraph[*e0]);
						if (e.is_match(airwaymatchlevel))
							continue;
						boost::add_edge(boost::source(*e0, m_lgraph), newsuper.get_v0() + pi, e, m_lgraph);
					}
				}
				superidx = newsuperidx;
			}
		}
		const LSuperVertex& super(m_lsupervertices[superidx]);
		const Performance& perf(m_performance[0]);
		double x(super.get_coord().spheric_distance_nmi_dbl(sdest.get_coord()) * perf.get_secpernmi());
		if (is_optfuel())
			x *= perf.get_fuelpersec();
		for (unsigned int pi = 0; pi < m_performance.size(); ++pi) {
			const Performance& perf(m_performance[pi]);
			boost::add_edge(super.get_v0() + pi, sdest.get_v0(), LEdge(airwaystar, x), m_lgraph);
			if (is_optfuel())
				x += perf.get_fuelpenalty();
			else
				x += perf.get_timepenalty();
		}
		lgraphkilllevelchg(super);
	}
	{
		std::ostringstream oss;
		oss << "Routing Graph after SID/STAR population: " << boost::num_vertices(m_lgraph) << " V, " << boost::num_edges(m_lgraph) << " E";
		m_signal_log(log_debug0, oss.str());
	}
	lgraphmodified();
	return true;
}

void CFMUAutoroute::lgraphupdatemetric(LRoute& r)
{
	if (r.size() < 2) {
		r.set_metric(std::numeric_limits<double>::quiet_NaN());
		return;
	}
	double m(0);
	for (LRoute::size_type i(1); i < r.size(); ++i) {
		LGraph::edge_descriptor e;
		bool ok;
		tie(e, ok) = r.get_edge(i, m_lgraph);
		if (!ok) {
			r.set_metric(std::numeric_limits<double>::infinity());
			const LVertex& v0(m_lgraph[r[i-1].get_vertex()]);
			const LVertex& v1(m_lgraph[r[i].get_vertex()]);
			const LSuperVertex& sv0(m_lsupervertices[v0.get_superindex()]);
			const LSuperVertex& sv1(m_lsupervertices[v1.get_superindex()]);
			std::ostringstream oss;
			oss << "Route Update Metric: Leg " << sv0.get_ident()
			    << "--" << lgraphindex2airway(r[i].get_edge(), true) << "->"
			    << sv1.get_ident() << " does not exist";
			m_signal_log(log_debug0, oss.str());
			return;
		}
		const LEdge& ee(m_lgraph[e]);
		m += ee.get_metric();
	}
	r.set_metric(m);
}

void CFMUAutoroute::lgraphclearcurrent(void)
{
	while (m_route.get_nrwpt())
		m_route.delete_wpt(0);
	m_routetime = 0;
	m_routefuel = 0;
	m_validationresponse.clear();
	m_solutionvertices.clear();
	{
		LGraph::edge_iterator e0, e1;
		for (tie(e0, e1) = edges(m_lgraph); e0 != e1; ++e0) {
			LEdge& ee(m_lgraph[*e0]);
			ee.set_solution(false);
		}
	}
}

bool CFMUAutoroute::lgraphsetcurrent(const LRoute& r)
{
	lgraphclearcurrent();
	if (r.size() < 2)
		return false;
	{
		LGraph::vertex_descriptor v(r.back().get_vertex());
		m_solutionvertices.insert(v);
		{
			const LVertex& vv(m_lgraph[v]);
			const LSuperVertex& sv(m_lsupervertices[vv.get_superindex()]);
			if (!sv.insert(m_route, 0)) {
				if (true) {
					std::ostringstream oss;
					oss << "Cannot insert " << sv.get_ident() << " into fpl";
					m_signal_log(log_debug0, oss.str());
				}
				lgraphclearcurrent();
				return false;
			}
			if (get_destination_ifr())
				m_route[0].set_flags(m_route[0].get_flags() | FPlanWaypoint::altflag_ifr);
			else
				m_route[0].set_flags(m_route[0].get_flags() & ~FPlanWaypoint::altflag_ifr);
		}
	}
	for (LRoute::size_type i(r.size()); i > 1; ) {
		--i;
		LGraph::vertex_descriptor u(r[i-1].get_vertex()), v(r[i].get_vertex());
		m_solutionvertices.insert(u);
		const LVertex& uu(m_lgraph[u]);
		const LVertex& vv(m_lgraph[v]);
		LGraph::edge_descriptor e;
		{
			bool ok;
			tie(e, ok) = r.get_edge(i, m_lgraph);
			if (!ok) {
				if (true) {
					std::ostringstream oss;
					oss << "k shortest path: Solution Edge " << lgraphvertexname(u)
					    << "--" << lgraphindex2airway(r[i].get_edge(), true) << "->"
					    << lgraphvertexname(v) << " not found";
					m_signal_log(log_debug0, oss.str());
				}
				lgraphclearcurrent();
				return false;
			}
		}
		LEdge& ee(m_lgraph[e]);
		ee.set_solution(true);
		if (false) {
			const LSuperVertex& su(m_lsupervertices[uu.get_superindex()]);
			const LSuperVertex& sv(m_lsupervertices[vv.get_superindex()]);
			const LEdge& ee(m_lgraph[e]);
			std::ostringstream oss;
			oss << "Solution Edge: " << su.get_ident() << " (" << m_performance[u - su.get_v0()].get_altitude() << ") "
			    << su.get_coord().get_lat_str2() << ' ' << su.get_coord().get_lon_str2()
			    << " --" << lgraphindex2airway(ee.get_ident()) << "-> "
			    << sv.get_ident() << " (" << m_performance[v - sv.get_v0()].get_altitude() << ") "
			    << sv.get_coord().get_lat_str2() << ' ' << sv.get_coord().get_lon_str2();
			m_signal_log(log_debug0, oss.str());
		}
		if (uu.get_superindex() == vv.get_superindex())
			continue;
		const LSuperVertex& su(m_lsupervertices[uu.get_superindex()]);
		const LSuperVertex& sv(m_lsupervertices[vv.get_superindex()]);
		if (!su.insert(m_route, 0)) {
			if (true) {
				std::ostringstream oss;
				oss << "Cannot insert " << su.get_ident() << " into fpl";
				m_signal_log(log_debug0, oss.str());
			}
			lgraphclearcurrent();
			return false;
		}
		m_route[0].set_pathcode(FPlanWaypoint::pathcode_none);
		if (i <= 1)
			break;
		m_route[0].set_flags(m_route[0].get_flags() | (FPlanWaypoint::altflag_ifr | FPlanWaypoint::altflag_standard));
		unsigned int pi(u - su.get_v0());
		if (pi >= m_performance.size()) {
			if (true)
				m_signal_log(log_debug0, "Level graph structure error");
			lgraphclearcurrent();
			return false;
		}
		m_route[0].set_altitude(m_performance[pi].get_altitude());
		if (ee.is_match(airwaydct)) {
			m_route[0].set_pathcode(FPlanWaypoint::pathcode_directto);
		} else if (ee.is_match(airwaymatchawy)) {
			m_route[0].set_pathname(lgraphindex2airway(ee.get_ident()));
			m_route[0].set_pathcode(FPlanWaypoint::pathcode_airway);
		}
	}
	if (get_departure_ifr())
		m_route[0].set_flags(m_route[0].get_flags() | FPlanWaypoint::altflag_ifr);
	else
		m_route[0].set_flags(m_route[0].get_flags() & ~FPlanWaypoint::altflag_ifr);
	if (false && m_route.get_nrwpt() >= 4) {
		FPlanWaypoint& wptstar(m_route[m_route.get_nrwpt() - 2]);
		const FPlanWaypoint& wpt(m_route[m_route.get_nrwpt() - 3]);
		wptstar.set_altitude(wpt.get_altitude());
		uint16_t flags(wptstar.get_flags());
		flags ^= (flags ^ wpt.get_flags()) & FPlanWaypoint::altflag_standard;
		wptstar.set_flags(flags);
	}
	updatefpl();
	return true;
}

void CFMUAutoroute::lgraphmodified(void)
{
	m_solutiontree.clear();
	m_solutionpool.clear();
}

bool CFMUAutoroute::lgraphisinsolutiontree(const LRoute& r) const
{
	const LGraphSolutionTreeNode *tree(&m_solutiontree);
	for (LRoute::size_type i(0); i < r.size(); ++i) {
		LGraphSolutionTreeNode::const_iterator ti(tree->find(LVertexEdge(r[i])));
		if (ti == tree->end())
			return false;
		tree = &ti->second;
	}
	return true;
}

bool CFMUAutoroute::lgraphinsertsolutiontree(const LRoute& r)
{
	bool intree(true);
	LGraphSolutionTreeNode *tree(&m_solutiontree);
	for (LRoute::const_iterator ri(r.begin()), re(r.end()); ri != re; ++ri) {
		std::pair<LGraphSolutionTreeNode::iterator,bool> ins(tree->insert(LGraphSolutionTreeNode::value_type(LVertexEdge(*ri), LGraphSolutionTreeNode())));
		tree = &ins.first->second;
		intree = intree && !ins.second;
	}
	return intree;
}

std::pair<CFMUAutoroute::LGraph::edge_descriptor,bool> CFMUAutoroute::lgraphfindbestedge(LGraph::vertex_descriptor u, LGraph::vertex_descriptor v, bool solutiononly) const
{
	std::pair<LGraph::edge_descriptor,bool> r;
	r.second = false;
	double bestm(std::numeric_limits<double>::max());
	LGraph::out_edge_iterator e0, e1;
	for (boost::tie(e0, e1) = boost::out_edges(u, m_lgraph); e0 != e1; ++e0) {
		if (boost::target(*e0, m_lgraph) != v)
			continue;
		const LEdge& ee(m_lgraph[*e0]);
		if (solutiononly && !ee.is_solution())
			continue;
		if (std::isnan(ee.get_metric()) || std::isinf(ee.get_metric()))
			continue;
		if (ee.get_metric() >= bestm)
			continue;
		bestm = ee.get_metric();
		r.second = true;
		r.first = *e0;
	}
	return r;
}

CFMUAutoroute::LRoute CFMUAutoroute::lgraphdijkstraroute(LGraph::vertex_descriptor u, LGraph::vertex_descriptor v, const std::vector<LGraph::vertex_descriptor>& predecessors) const
{
	LRoute r;
	if (u == v)
		return r;
	while (v != u) {
		LGraph::vertex_descriptor uu(predecessors[v]);
		if (uu == v) {
			r.clear();
			return r;
		}
		std::pair<LGraph::edge_descriptor,bool> e(lgraphfindbestedge(uu, v, false));
		if (!e.second) {
			r.clear();
			return r;
		}
		r.insert(r.begin(), LVertexEdge(e.first, m_lgraph));
		v = uu;
	}
	r.insert(r.begin(), LVertexEdge(u));
	return r;	
}

std::string CFMUAutoroute::lgraphprint(const LRoute& r)
{
	std::ostringstream oss;
	for (LRoute::const_iterator ri(r.begin()), re(r.end()); ri != re; ++ri) {
		if (ri->get_edge() != airwaymatchnone)
			oss << ' ' << lgraphindex2airway(ri->get_edge(), true);
		oss << ' ' << lgraphvertexname(ri->get_vertex());
	}
	if (oss.str().empty())
		return std::string();
	return oss.str().substr(1);
}

class CFMUAutoroute::DijkstraCore {
public:
	typedef enum {
		white,
		gray,
		black
	} color_t;

	class DijkstraQueue {
	public:
		DijkstraQueue(LGraph::vertex_descriptor v = 0, double d = 0) : m_dist(d), m_vertex(v) {}
		double get_dist(void) const { return m_dist; }
		LGraph::vertex_descriptor get_vertex(void) const { return m_vertex; }
		bool operator<(const DijkstraQueue& x) const {
			if (get_dist() < x.get_dist())
				return true;
			if (x.get_dist() < get_dist())
				return false;
			return get_vertex() < x.get_vertex();
		}

	protected:
		double m_dist;
		LGraph::vertex_descriptor m_vertex;
	};

	DijkstraCore(const LGraph& g, const LSuperVertices& s);
	void init(void);
	void set_source(LGraph::vertex_descriptor u);
	void set_source_nodist(LGraph::vertex_descriptor u);
	void rebuild_queue(void);
	void run(lgraphairwayindex_t awy = airwaymatchall, bool solutionedgeonly = false);
	void run_once(lgraphairwayindex_t awy = airwaymatchall, bool solutionedgeonly = false);
	LRoute get_route(LGraph::vertex_descriptor v);
	void mark_all_white(void);
	void mark_path(LGraph::vertex_descriptor v);
	void mark_white_infinite(void);
	void copy_gray_paths(const DijkstraCore& x);
	void swap(DijkstraCore& x);

	const LGraph& m_lgraph;
	const LSuperVertices& m_lsupervertices;
	typedef std::vector<color_t> colormap_t;
	colormap_t m_colormap;
	typedef std::vector<LGraph::vertex_descriptor> predecessors_t;
	predecessors_t m_predecessors;
	typedef std::vector<LGraph::edge_descriptor> edges_t;
	edges_t m_edges; // best edge inbound to this vertex
	typedef std::vector<double> distances_t;
	distances_t m_distances;
	typedef std::set<DijkstraQueue> queue_t;
	queue_t m_queue;
};

CFMUAutoroute::DijkstraCore::DijkstraCore(const LGraph& g, const LSuperVertices& s)
	: m_lgraph(g), m_lsupervertices(s)
{
	init();
}

void CFMUAutoroute::DijkstraCore::init(void)
{
	m_colormap.clear();
	m_colormap.resize(boost::num_vertices(m_lgraph), white);
	m_predecessors.clear();
	m_predecessors.resize(boost::num_vertices(m_lgraph));
	m_edges.clear();
	m_edges.resize(boost::num_vertices(m_lgraph));
	m_distances.clear();
	m_distances.resize(boost::num_vertices(m_lgraph), std::numeric_limits<double>::max());
	for (unsigned int i = 0, n = boost::num_vertices(m_lgraph); i < n; ++i)
		m_predecessors[i] = i;
}

void CFMUAutoroute::DijkstraCore::set_source(LGraph::vertex_descriptor u)
{
	m_colormap[u] = gray;
	m_distances[u] = 0;
	m_queue.insert(DijkstraQueue(u, m_distances[u]));
}

void CFMUAutoroute::DijkstraCore::set_source_nodist(LGraph::vertex_descriptor u)
{
	m_colormap[u] = gray;
	m_queue.insert(DijkstraQueue(u, m_distances[u]));
}

void CFMUAutoroute::DijkstraCore::rebuild_queue(void)
{
	m_queue.clear();
	for (unsigned int i = 0, n = boost::num_vertices(m_lgraph); i < n; ++i) {
		if (m_colormap[i] != gray)
			continue;
		m_queue.insert(DijkstraQueue(i, m_distances[i]));
	}
}

void CFMUAutoroute::DijkstraCore::run_once(lgraphairwayindex_t awy, bool solutionedgeonly)
{
	LGraph::vertex_descriptor u(m_queue.begin()->get_vertex());
	m_queue.erase(m_queue.begin());
	LGraph::out_edge_iterator e0, e1;
	for (boost::tie(e0, e1) = boost::out_edges(u, m_lgraph); e0 != e1; ++e0) {
		LGraph::vertex_descriptor v(boost::target(*e0, m_lgraph));
		if (m_colormap[v] == black)
			continue;
		const LEdge& ee(m_lgraph[*e0]);
		if (!ee.is_match(awy))
			continue;
		if (solutionedgeonly && !ee.is_solution())
			continue;
		double newmetric(m_distances[u] + ee.get_metric());
		if (newmetric >= m_distances[v])
			continue;
		// check whether new edge does not violate some flight plan requirements
		// this violates the optimality principle, hence this dijkstra implementation
		// is not strictly optimal. However, we do not want to incur the cost of a full
		// global optimization. The SID/STAR levelchange restrictions are harmless, though.
		if (m_predecessors[u] != u) {
			const LEdge& eep(m_lgraph[m_edges[u]]);
			// check SID is not immediately followed by a level change
			if (eep.is_match(airwaysid) && ee.is_match(airwaymatchlevel))
				continue;
			// check STAR does not immediately follow a level change
			if (eep.is_match(airwaymatchlevel) && ee.is_match(airwaystar))
				continue;
		}
		// check the flight plan does not contain duplicate nodes
		if (!ee.is_match(airwaymatchlevel)) {
			const LVertex& vv(m_lgraph[v]);
			const LSuperVertex& superv(m_lsupervertices[vv.get_superindex()]);
			LGraph::vertex_descriptor w(u);
			LGraph::edge_descriptor e(*e0);
			bool ok(true);
			for (;;) {
				const LEdge& ee(m_lgraph[e]);
				if (!ee.is_match(airwaymatchlevel)) {
					const LVertex& ww(m_lgraph[w]);
					const LSuperVertex& superw(m_lsupervertices[ww.get_superindex()]);
					if (superw.get_ident() == superv.get_ident()) {
						ok = false;
						break;
					}
				}
				if (w == m_predecessors[w])
					break;
				e = m_edges[w];
				w = m_predecessors[w];
			}
			if (!ok)
				continue;
		}
		// edge is a new better edge
		if (m_colormap[v] == gray) {
			queue_t::iterator i(m_queue.find(DijkstraQueue(v, m_distances[v])));
			if (i == m_queue.end()) {
				std::ostringstream oss;
				oss << "lgraphdijkstra: gray node not found in m_queue";
				throw std::runtime_error(oss.str());
			} else {
				m_queue.erase(i);
			}
		}
		m_distances[v] = newmetric;
		m_predecessors[v] = u;
		m_edges[v] = *e0;
		m_colormap[v] = gray;
		m_queue.insert(DijkstraQueue(v, m_distances[v]));
	}
	m_colormap[u] = black;
}

void CFMUAutoroute::DijkstraCore::run(lgraphairwayindex_t awy, bool solutionedgeonly)
{
	// DIJKSTRA(G, s, w)
	//   for each vertex u in V (This loop is not run in dijkstra_shortest_paths_no_init)
	//     d[u] := infinity 
	//     p[u] := u 
	//     color[u] := WHITE
	//   end for
	//   color[s] := GRAY 
	//   d[s] := 0 
	//   INSERT(Q, s)
	//   while (Q != Ø)
	//     u := EXTRACT-MIN(Q)
	//     S := S U { u }
	//     for each vertex v in Adj[u]
	//       if (w(u,v) + d[u] < d[v])
	//         d[v] := w(u,v) + d[u]
	//         p[v] := u 
	//         if (color[v] = WHITE) 
	//           color[v] := GRAY
	//           INSERT(Q, v) 
	//         else if (color[v] = GRAY)
	//           DECREASE-KEY(Q, v)
	//       else
	//         ...
	//     end for
	//     color[u] := BLACK
	//   end while
	//   return (d, p)
	while (!m_queue.empty())
		run_once(awy, solutionedgeonly);
}

CFMUAutoroute::LRoute CFMUAutoroute::DijkstraCore::get_route(LGraph::vertex_descriptor v)
{
	LRoute r;
	{
		LGraph::vertex_descriptor w(v);
		for (;;) {
			LGraph::vertex_descriptor u(m_predecessors[w]);
			if (u == w) {
				r.insert(r.begin(), LVertexEdge(w));
				break;
			}
			r.insert(r.begin(), LVertexEdge(m_edges[w], m_lgraph));
			w = u;
		}
	}
	if (r.size() <= 1)
		r.clear();
	return r;
}

void CFMUAutoroute::DijkstraCore::mark_all_white(void)
{
	for (unsigned int i = 0, n = boost::num_vertices(m_lgraph); i < n; ++i)
		m_colormap[i] = white;
}

void CFMUAutoroute::DijkstraCore::mark_path(LGraph::vertex_descriptor v)
{
	if (m_predecessors[v] == v)
		return;
	m_colormap[v] = gray;
	for (;;) {
		LGraph::vertex_descriptor u(m_predecessors[v]);
		if (u == v)
			break;
		if (m_colormap[u] == white)
			m_colormap[u] = black;
		v = u;
	}
}

void CFMUAutoroute::DijkstraCore::mark_white_infinite(void)
{
	for (unsigned int i = 0, n = boost::num_vertices(m_lgraph); i < n; ++i)
		if (m_colormap[i] == white)
			m_distances[i] = std::numeric_limits<double>::max();
}

void CFMUAutoroute::DijkstraCore::copy_gray_paths(const DijkstraCore& x)
{
	if (boost::num_vertices(m_lgraph) != boost::num_vertices(x.m_lgraph))
		throw std::runtime_error("copy_gray_paths: invalid graphs");
	for (unsigned int i = 0, n = boost::num_vertices(m_lgraph); i < n; ++i) {
		if (x.m_colormap[i] != gray)
			continue;
		if (x.m_predecessors[i] == i)
			continue;
		LGraph::vertex_descriptor v(i);
		for (;;) {
			LGraph::vertex_descriptor u(x.m_predecessors[v]);
			if (!(x.m_distances[v] < m_distances[v]))
				break;
			m_predecessors[v] = u;
			m_distances[v] = x.m_distances[v];
			m_edges[v] = x.m_edges[v];
			if (m_colormap[v] == white)
				m_colormap[v] = black;
			if (u == v)
				break;
			v = u;
		}
		m_colormap[i] = gray;
	}
}

void CFMUAutoroute::DijkstraCore::swap(DijkstraCore& x)
{
	if (boost::num_vertices(m_lgraph) != boost::num_vertices(x.m_lgraph))
		throw std::runtime_error("swap: invalid graphs");
	m_colormap.swap(x.m_colormap);
	m_predecessors.swap(x.m_predecessors);
	m_edges.swap(x.m_edges);
	m_distances.swap(x.m_distances);
	m_queue.swap(x.m_queue);
}

CFMUAutoroute::LRoute CFMUAutoroute::lgraphdijkstra(LGraph::vertex_descriptor u, LGraph::vertex_descriptor v, const LRoute& baseroute, bool solutionedgeonly, LGMandatory mandatory)
{
	DijkstraCore d(m_lgraph, m_lsupervertices);
	d.m_distances[u] = 0;
	// start at specific route
	if (!baseroute.empty()) {
		LRoute::size_type i(1), n(baseroute.size());
		if (baseroute[0].get_vertex() != u) {
			d.m_distances[baseroute[0].get_vertex()] = 0;
			for (; i < n; ++i) {
				LGraph::vertex_descriptor vv(baseroute[i].get_vertex());
				LGraph::vertex_descriptor uu(baseroute[i - 1].get_vertex());
				d.m_predecessors[vv] = uu;
				std::pair<LGraph::edge_descriptor,bool> ee(LRoute::get_edge(baseroute[i - 1], baseroute[i], m_lgraph));
				if (!ee.second) {
					std::ostringstream oss;
					oss << "lgraphdijkstra: base route edge not found";
					m_signal_log(log_debug0, oss.str());
					return LRoute();
				}
				d.m_edges[vv] = ee.first;
				d.m_colormap[uu] = DijkstraCore::black;
				const LEdge& eee(m_lgraph[ee.first]);
				d.m_distances[vv] = d.m_distances[uu] + eee.get_metric();
				if (vv == u)
					break;
			}
			if (i >= n) {
				std::ostringstream oss;
				oss << "lgraphdijkstra: start vertex not found in base route";
				m_signal_log(log_debug0, oss.str());
			}
		}
	}
	d.set_source_nodist(u);
	for (;;) {
		try {
			d.run(airwaymatchall, solutionedgeonly);
		} catch (const std::exception& e) {
			m_signal_log(log_debug0, e.what());
			return LRoute();
		}
		if (mandatory.empty())
			break;
		// find the mandatory rule that is closest to the source
		unsigned int mrule(~0U);
		{
			double bestdist(std::numeric_limits<double>::max());
			for (unsigned int mr = 0; mr < mandatory.size(); ++mr) {
				const LGMandatoryAlternative& alt(mandatory[mr]);
				for (unsigned int ma = 0; ma < alt.size(); ++ma) {
					const LGMandatorySequence& seq(alt[ma]);
					if (seq.empty())
						continue;
					double dist(std::numeric_limits<double>::max());
					for (LGraph::vertex_descriptor v = seq[0].get_v0(); v <= seq[0].get_v1(); ++v) {
						if (v == d.m_predecessors[v])
							continue;
						if (!(d.m_distances[v] < dist) || !(d.m_distances[v] < bestdist))
							continue;
						LGraph::out_edge_iterator e0, e1;
						for (boost::tie(e0, e1) = boost::out_edges(v, m_lgraph); e0 != e1; ++e0) {
							const LEdge& ee(m_lgraph[*e0]);
							if (!ee.is_match(seq[0].get_airway()))
								continue;
							break;
						}
						if (e0 == e1)
							continue;
						dist = d.m_distances[v];
					}
					if (!(dist < bestdist))
						continue;
					bestdist = dist;
					mrule = mr;
				}
			}
		}
		if (mrule >= mandatory.size()) {
			std::ostringstream oss;
			oss << "lgraphdijkstra: cannot satisfy mandatory rules";
			m_signal_log(log_debug0, oss.str());
			return LRoute();
		}
		// remove all but paths to beginning of mandatory alternatives from dijkstra results
		d.mark_all_white();
		{
			const LGMandatoryAlternative& alt(mandatory[mrule]);
			for (unsigned int ma = 0; ma < alt.size(); ++ma) {
				const LGMandatorySequence& seq(alt[ma]);
				if (seq.empty())
					continue;
				for (LGraph::vertex_descriptor v = seq[0].get_v0(); v <= seq[0].get_v1(); ++v) {
					d.mark_path(v);
					if (true && d.m_predecessors[v] != v) {
						const LVertex& vv(m_lgraph[v]);
						const LSuperVertex& sv(m_lsupervertices[vv.get_superindex()]);
						std::ostringstream oss;
						oss << "lgraphdijkstra: mandatory rule " << mrule << " alt " << ma << " marking " << sv.get_ident();
						unsigned int pi(v - sv.get_v0());
						if (sv.get_v0() != sv.get_v1() && pi < m_performance.size())
							oss << " F" << std::setw(3) << std::setfill('0') << (m_performance[pi].get_altitude() / 100);
						oss << " dist " << d.m_distances[v];
						m_signal_log(log_debug0, oss.str());
					}
				}
			}
		}
		d.mark_white_infinite();
		{
			const LGMandatoryAlternative& alt(mandatory[mrule]);
			DijkstraCore dr(m_lgraph, m_lsupervertices);
			for (unsigned int ma = 0; ma < alt.size(); ++ma) {
				const LGMandatorySequence& seq(alt[ma]);
				if (seq.empty())
					continue;
				DijkstraCore ds(d);
				for (unsigned int mp = 0; ; ++mp) {
					const LGMandatoryPoint& pt(seq[mp]);
					ds.mark_all_white();
					for (LGraph::vertex_descriptor v = pt.get_v0(); v <= pt.get_v1(); ++v) {
						ds.mark_path(v);
						if (true && ds.m_predecessors[v] != v) {
							const LVertex& vv(m_lgraph[v]);
							const LSuperVertex& sv(m_lsupervertices[vv.get_superindex()]);
							std::ostringstream oss;
							oss << "lgraphdijkstra: mandatory rule " << mrule << " alt " << ma << " point " << mp << " marking " << sv.get_ident();
							unsigned int pi(v - sv.get_v0());
							if (sv.get_v0() != sv.get_v1() && pi < m_performance.size())
								oss << " F" << std::setw(3) << std::setfill('0') << (m_performance[pi].get_altitude() / 100);
							oss << " dist " << ds.m_distances[v];
							m_signal_log(log_debug0, oss.str());
						}
					}
					if (mp + 1 >= seq.size())
						break;
					ds.mark_white_infinite();
					ds.rebuild_queue();
					if (pt.get_airway() == airwaydct) {
						int minalt(-1);
						const LVertex& vv(m_lgraph[seq[mp + 1].get_v0()]);
						const LSuperVertex& sv(m_lsupervertices[vv.get_superindex()]);
						for (LGraph::vertex_descriptor u = pt.get_v0(); u <= pt.get_v1(); ++u) {
							if (ds.m_predecessors[u] == u)
								continue;
							unsigned int pi(u - pt.get_v0());
							LGraph::vertex_descriptor v(sv.get_v0() + pi);
							if (v > sv.get_v1())
								continue;
							LGraph::out_edge_iterator e0, e1;
							for (boost::tie(e0, e1) = boost::out_edges(u, m_lgraph); e0 != e1; ++e0) {
								if (boost::target(*e0, m_lgraph) != v)
									continue;
								const LEdge& ee(m_lgraph[*e0]);
								if (!ee.is_match(airwaydct))
									continue;
								if (solutionedgeonly && !ee.is_solution())
									break;
								ds.m_predecessors[v] = u;
								ds.m_distances[v] = ds.m_distances[u] + ee.get_metric();
								ds.m_colormap[v] = DijkstraCore::gray;
								ds.m_edges[v] = *e0;
								break;
							}
							if (e0 != e1)
								continue;
							if (pi >= m_performance.size())
								continue;
							const Performance& perf(m_performance[pi]);
							const LVertex& uu(m_lgraph[u]);
							const LSuperVertex& su(m_lsupervertices[uu.get_superindex()]);
							double dist(su.get_coord().spheric_distance_nmi_dbl(sv.get_coord()));
							if (minalt < 0) {
								TopoDb30::Profile prof(m_topodb.get_profile(su.get_coord(), sv.get_coord(), 5));
								if (prof.empty()) {
									std::ostringstream oss;
									oss << "lgraphdijkstra: DCT: Error computing ground clearance: " << su.get_ident() << ' '
									    << sv.get_ident() << " dist " << std::setprecision(1) << dist << "nmi";
									m_signal_log(log_debug0, oss.str());
								} else {
									TopoDb30::elev_t el(prof.get_maxelev());
									if (el != TopoDb30::nodata) {
										minalt = prof.get_maxelev() * Point::m_to_ft;
										minalt = std::max(minalt, 0);
										if (minalt >= 5000)
											minalt += 1000;
										minalt += 1000;
									}
								}
							}
							if (perf.get_altitude() < minalt)
								continue;
							double x(dist * perf.get_secpernmi() * get_dctpenalty());
							if (is_optfuel())
								x *= perf.get_fuelpersec();
							LGraph::edge_descriptor e;
							bool ok;
							tie(e, ok) = boost::add_edge(u, v, LEdge(airwaydct, x), m_lgraph);
							if (ok) {
								ds.m_predecessors[v] = u;
								ds.m_distances[v] = ds.m_distances[u] + x;
								ds.m_colormap[v] = DijkstraCore::gray;
								ds.m_edges[v] = e;
							}
						}
					} else {
						ds.run(pt.get_airway(), solutionedgeonly);
					}
				}
				dr.copy_gray_paths(ds);
				if (true) {
					for (unsigned int i = 0, n = boost::num_vertices(m_lgraph); i < n; ++i) {
						if (ds.m_colormap[i] != DijkstraCore::gray)
							continue;
						LRoute r(ds.get_route(i));
						if (r.empty())
							continue;
						lgraphupdatemetric(r);
						std::ostringstream oss;
						oss << "lgraphdijkstra: mandatory rule " << mrule << " alt " << ma << " route " << lgraphprint(r) << " metric " << r.get_metric();
						m_signal_log(log_debug0, oss.str());
					}
				}
			}
			d.swap(dr);
			d.rebuild_queue();
			if (true) {
				for (unsigned int i = 0, n = boost::num_vertices(m_lgraph); i < n; ++i) {
					if (d.m_colormap[i] != DijkstraCore::gray)
						continue;
					LRoute r(d.get_route(i));
					if (r.empty())
						continue;
					lgraphupdatemetric(r);
					std::ostringstream oss;
					oss << "lgraphdijkstra: mandatory rule " << mrule << " route " << lgraphprint(r) << " metric " << r.get_metric();
					m_signal_log(log_debug0, oss.str());
				}
			}
		}
		mandatory.erase(mandatory.begin() + mrule);
		if (d.m_queue.empty()) {
			std::ostringstream oss;
			oss << "lgraphdijkstra: no route after mandatory processing";
			m_signal_log(log_debug0, oss.str());
			return LRoute();
		}
	}
	LRoute r(d.get_route(v));
	if (r.empty())
		return r;
	lgraphupdatemetric(r);
	if (r[0].get_vertex() != (baseroute.empty() ? u : baseroute[0].get_vertex())) {
			const LVertex& v0(m_lgraph[r[0].get_vertex()]);
			const LVertex& v1(m_lgraph[baseroute.empty() ? u : baseroute[0].get_vertex()]);
			const LSuperVertex& sv0(m_lsupervertices[v0.get_superindex()]);
			const LSuperVertex& sv1(m_lsupervertices[v1.get_superindex()]);
		std::ostringstream oss;
		oss << "lgraphdijkstra: final route invalid starting point " << sv0.get_ident() << " expected " << sv1.get_ident();
		m_signal_log(log_debug0, oss.str());
	}
	return r;
}

bool CFMUAutoroute::is_metric_invalid(double m)
{
	return std::isnan(m) || std::isinf(m) || m == std::numeric_limits<double>::max();
}

struct CFMUAutoroute::Combine {
        double operator()(double a, const CFMUAutoroute::LEdge& b) const
	{
                return a + b.m_metric;
        }
};

#if 0

CFMUAutoroute::LRoute CFMUAutoroute::lgraphnextsolution(void)
{
	LRoute r;
	for (;;) {
		if (m_solutionpool.empty()) {
			const LSuperVertex& sdep(m_lsupervertices[m_superdep]);
			const LSuperVertex& sdest(m_lsupervertices[m_superdest]);
			std::vector<LGraph::vertex_descriptor> predecessors(boost::num_vertices(m_lgraph), 0);
			dijkstra_shortest_paths(m_lgraph, sdep.get_v0(), boost::weight_map(boost::get(&LEdge::m_metric, m_lgraph)).
						predecessor_map(&predecessors[0]));
			r = lgraphdijkstraroute(sdep.get_v0(), sdest.get_v0(), predecessors);
			lgraphupdatemetric(r);
			if (is_metric_invalid(r.get_metric())) {
				r.clear();
				return r;
			}
			m_solutionpool.insert(r);
			// check if already in solution tree; return if not, otherwise use as solution to base other solutions off
			if (lgraphinsertsolutiontree(r))
				return r;
		}
		r = *m_solutionpool.begin();
		m_solutionpool.erase(m_solutionpool.begin());
		// verify if solution still exists in graph
		bool ok(r.is_existing(m_lgraph));
		if (true) {
			std::ostringstream oss;
			oss << "k shortest path: Route " << lgraphprint(r) << " metric " << r.get_metric();
			if (ok)
				oss << " still exists";
			else
				oss << " no longer exists";
			m_signal_log(log_debug0, oss.str());
		}
		if (ok)
			break;
	}
	lgraphinsertsolutiontree(r);
	// generate new solutions
	typedef std::map<LGraph::edge_descriptor,double> wmap_t;
	wmap_t wmap;	
	{
		LGraph::edge_iterator e0, e1;
		for (tie(e0, e1) = edges(m_lgraph); e0 != e1; ++e0) {
			const LEdge& ee(m_lgraph[*e0]);
			wmap.insert(wmap_t::value_type(*e0, ee.get_metric()));
		}
	}
	boost::associative_property_map<wmap_t> pmwmap(wmap); 
	LGraphSolutionTreeNode *tree(&m_solutiontree);
	const LSuperVertex& sdest(m_lsupervertices[m_superdest]);
	for (LRoute::size_type i(0); i + 1 < r.size(); ++i) {
		if (tree) {
			LGraphSolutionTreeNode::iterator ti(tree->find(LVertexEdge(r[i])));
			if (ti == tree->end())
				tree = 0;
			else
				tree = &ti->second;
		}
		if (tree) {
			for (LGraphSolutionTreeNode::iterator ti(tree->begin()), te(tree->end()); ti != te;) {
				LGraphSolutionTreeNode::iterator tix(ti);
				++ti;
				LGraph::edge_descriptor e;
                                bool ok;
				boost::tie(e, ok) = LRoute::get_edge(r[i], tix->first, m_lgraph);
				if (!ok) {
					if (true) {
						std::ostringstream oss;
						oss << "k shortest path: Leg " << lgraphvertexname(r[i].get_vertex())
						    << "--" << lgraphindex2airway(tix->first.get_edge(), true) << "->"
						    << lgraphvertexname(tix->first.get_vertex()) << " does not exist in graph";
						m_signal_log(log_debug0, oss.str());
					}
					tree->erase(tix);
					continue;
				}
				wmap_t::iterator wi(wmap.find(e));
				if (wi == wmap.end()) {
					if (true) {
						std::ostringstream oss;
						oss << "k shortest path: Leg " << lgraphvertexname(r[i].get_vertex())
						    << "--" << lgraphindex2airway(tix->first.get_edge(), true) << "->"
						    << lgraphvertexname(tix->first.get_vertex()) << " does not exist in weight map";
						m_signal_log(log_debug0, oss.str());
					}
					continue;
				}
				wi->second = std::numeric_limits<double>::max();
			}
		}
		// kill level changes immediately after SID endpoint
		if (i == 1) {
			LGraph::out_edge_iterator e0, e1;
                        for (boost::tie(e0, e1) = boost::out_edges(r[i].get_vertex(), m_lgraph); e0 != e1; ++e0) {
				const LEdge& ee(m_lgraph[*e0]);
				if (!ee.is_match(airwaymatchlevel))
					continue;
				wmap_t::iterator wi(wmap.find(*e0));
				if (wi == wmap.end())
					continue;
				wi->second = std::numeric_limits<double>::max();
			}
		}
		std::vector<LGraph::vertex_descriptor> predecessors(boost::num_vertices(m_lgraph), 0);
		dijkstra_shortest_paths(m_lgraph, r[i].get_vertex(), boost::weight_map(pmwmap).predecessor_map(&predecessors[0]));
		//std::vector<double> dists(boost::num_vertices(m_lgraph), 0);
		//dijkstra_shortest_paths(m_lgraph, r[i].get_vertex(), boost::weight_map(boost::get(boost::edge_bundle, m_lgraph)).predecessor_map(&predecessors[0]).distance_map(&dists[0]).distance_combine(Combine()));
		//dijkstra_shortest_paths(m_lgraph, r[i].get_vertex(), boost::weight_map(pmwmap).predecessor_map(&predecessors[0]).distance_map(&dists[0]));
		LRoute r1;
		{
			LGraph::vertex_descriptor u(r[i].get_vertex());
			LGraph::vertex_descriptor v(sdest.get_v0());
			while (v != u) {
				LGraph::vertex_descriptor uu(predecessors[v]);
				if (uu == v) {
					r1.clear();
					break;
				}
				std::pair<LGraph::edge_descriptor,bool> e(lgraphfindbestedge(uu, v));
				if (!e.second) {
					r1.clear();
					break;
				}
				r1.insert(r1.begin(), LVertexEdge(e.first, m_lgraph));
				// check that weight map exists and is not infinity
				wmap_t::iterator wi(wmap.find(e.first));
				if (wi == wmap.end()) {
					if (true) {
						std::ostringstream oss;
						oss << "k shortest path: Leg " << lgraphvertexname(boost::source(e.first, m_lgraph))
						    << "--" << lgraphindex2airway(m_lgraph[e.first].get_ident(), true) << "->"
						    << lgraphvertexname(boost::target(e.first, m_lgraph)) << " does not exist in weight map";
						m_signal_log(log_debug0, oss.str());
					}
					r1.clear();
					break;
				}
				if (is_metric_invalid(wi->second)) {
					if (true) {
						std::ostringstream oss;
						oss << "k shortest path: Leg " << lgraphvertexname(boost::source(e.first, m_lgraph))
						    << "--" << lgraphindex2airway(m_lgraph[e.first].get_ident(), true) << "->"
						    << lgraphvertexname(boost::target(e.first, m_lgraph)) << " has invalid metric";
						m_signal_log(log_debug0, oss.str());
					}
					r1.clear();
					break;
				}
				v = uu;
			}
		}
		if (false && tree) {
			LGraph::out_edge_iterator e0, e1;
			for (boost::tie(e0, e1) = boost::out_edges(r[i].get_vertex(), m_lgraph); e0 != e1; ++e0) {
				const LEdge& ee(m_lgraph[*e0]);
				if (boost::target(*e0, m_lgraph) == r[i+1].get_vertex() && 
				    ee.get_ident() == r[i+1].get_edge())
					continue;
				wmap_t::iterator wi(wmap.find(*e0));
				if (wi == wmap.end())
					continue;
				wi->second = ee.get_metric();
			}
		}
		// kill all out-edges to avoid revisiting a node already in the solution
		{
			const LVertex& vv(m_lgraph[r[i].get_vertex()]);
			const LSuperVertex& sv(m_lsupervertices[vv.get_superindex()]);
			for (LGraph::vertex_descriptor v0(sv.get_v0()); v0 <= sv.get_v1(); ++v0) {
				LGraph::out_edge_iterator e0, e1;
				for (boost::tie(e0, e1) = boost::out_edges(v0, m_lgraph); e0 != e1; ++e0) {
					wmap_t::iterator wi(wmap.find(*e0));
					if (wi == wmap.end())
						continue;
					wi->second = std::numeric_limits<double>::max();
					if (true) {
						std::ostringstream oss;
						oss << "k shortest path: Killing edge " << lgraphvertexname(boost::source(*e0, m_lgraph))
						    << "--" << lgraphindex2airway(m_lgraph[*e0].get_ident(), true) << "->"
						    << lgraphvertexname(boost::target(*e0, m_lgraph));
						m_signal_log(log_debug0, oss.str());
					}
				}
			}
		}
		if (r1.size() < 2)
			continue;
		r1.insert(r1.begin(), r.begin(), r.begin() + (i+1));
		lgraphupdatemetric(r1);
		if (is_metric_invalid(r1.get_metric())) {
			if (true) {
				std::ostringstream oss;
				oss << "k shortest path: Route";
		        	for (LRoute::const_iterator ri(r1.begin()), re(r1.end()); ri != re; ++ri) {
					if (ri->get_edge() != airwaymatchnone)
						oss << ' ' << lgraphindex2airway(ri->get_edge(), true);
					oss << ' ' << lgraphvertexname(ri->get_vertex());
				}
				oss << " has no metric";
				m_signal_log(log_debug0, oss.str());
			}
			continue;
		}
		if (true) {
			std::ostringstream oss;
			oss << "k shortest path: Route";
			for (LRoute::const_iterator ri(r1.begin()), re(r1.end()); ri != re; ++ri) {
				if (ri->get_edge() != airwaymatchnone)
					oss << ' ' << lgraphindex2airway(ri->get_edge(), true);
				oss << ' ' << lgraphvertexname(ri->get_vertex());
			}
			oss << " metric " << r1.get_metric() << " added to solution pool";
			m_signal_log(log_debug0, oss.str());
		}
		m_solutionpool.insert(r1);
	}
	// check for too many 
	while (!m_solutionpool.empty() && !m_solutionpool.begin()->is_existing(m_lgraph))
		m_solutionpool.erase(m_solutionpool.begin());
	{
		static const lgraphsolutionpool_t::size_type maxsz(16384);
		lgraphsolutionpool_t::size_type sz(m_solutionpool.size());
		if (sz > maxsz) {
			lgraphsolutionpool_t::size_type n(0);
			for (lgraphsolutionpool_t::iterator i(m_solutionpool.begin()), e(m_solutionpool.end()); i != e; ) {
				lgraphsolutionpool_t::iterator i2(i);
				++i;
				if (n >= maxsz || !i2->is_existing(m_lgraph)) {
					m_solutionpool.erase(i2);
					continue;
				}
				++n;
			}
		}
	}
	if (m_solutionpool.empty())
		return LRoute();
	return *m_solutionpool.begin();
}

#else

CFMUAutoroute::LRoute CFMUAutoroute::lgraphnextsolution(void)
{
	LRoute r;
	for (;;) {
		if (m_solutionpool.empty()) {
			const LSuperVertex& sdep(m_lsupervertices[m_superdep]);
			const LSuperVertex& sdest(m_lsupervertices[m_superdest]);
			if (false) {
				std::vector<LGraph::vertex_descriptor> predecessors(boost::num_vertices(m_lgraph), 0);
				dijkstra_shortest_paths(m_lgraph, sdep.get_v0(), boost::weight_map(boost::get(&LEdge::m_metric, m_lgraph)).
							predecessor_map(&predecessors[0]));
				r = lgraphdijkstraroute(sdep.get_v0(), sdest.get_v0(), predecessors);
			} else {
				r = lgraphdijkstra(sdep.get_v0(), sdest.get_v0());
			}
			lgraphupdatemetric(r);
			if (is_metric_invalid(r.get_metric())) {
				r.clear();
				return r;
			}
			m_solutionpool.insert(r);
			// check if already in solution tree; return if not, otherwise use as solution to base other solutions off
			if (lgraphinsertsolutiontree(r))
				return r;
		}
		r = *m_solutionpool.begin();
		m_solutionpool.erase(m_solutionpool.begin());
		// verify if solution still exists in graph
		bool ok(r.is_existing(m_lgraph));
		if (true) {
			std::ostringstream oss;
			oss << "k shortest path: Route " << lgraphprint(r) << " metric " << r.get_metric();
			if (ok)
				oss << " still exists";
			else
				oss << " no longer exists";
			m_signal_log(log_debug0, oss.str());
		}
		if (ok)
			break;
	}
	lgraphinsertsolutiontree(r);
	// generate new solutions
	// "abuse" solution flag to enable/disable edges
	{
		LGraph::edge_iterator e0, e1;
		for (tie(e0, e1) = edges(m_lgraph); e0 != e1; ++e0) {
			LEdge& ee(m_lgraph[*e0]);
			ee.set_solution(true);
		}
	}
        LGraphSolutionEdgeFilter filter(m_lgraph);
        typedef boost::filtered_graph<LGraph, LGraphSolutionEdgeFilter> fg_t;
        fg_t fg(m_lgraph, filter);
	LGraphSolutionTreeNode *tree(&m_solutiontree);
	const LSuperVertex& sdest(m_lsupervertices[m_superdest]);
	for (LRoute::size_type i(0); i + 1 < r.size(); ++i) {
		if (tree) {
			LGraphSolutionTreeNode::iterator ti(tree->find(LVertexEdge(r[i])));
			if (ti == tree->end())
				tree = 0;
			else
				tree = &ti->second;
		}
		if (tree) {
			for (LGraphSolutionTreeNode::iterator ti(tree->begin()), te(tree->end()); ti != te;) {
				LGraphSolutionTreeNode::iterator tix(ti);
				++ti;
				LGraph::edge_descriptor e;
                                bool ok;
				boost::tie(e, ok) = LRoute::get_edge(r[i], tix->first, m_lgraph);
				if (!ok) {
					if (true) {
						std::ostringstream oss;
						oss << "k shortest path: Leg " << lgraphvertexname(r[i].get_vertex())
						    << "--" << lgraphindex2airway(tix->first.get_edge(), true) << "->"
						    << lgraphvertexname(tix->first.get_vertex()) << " does not exist in graph";
						m_signal_log(log_debug0, oss.str());
					}
					tree->erase(tix);
					continue;
				}
				m_lgraph[e].set_solution(false);
			}
		}
		// kill level changes immediately after SID endpoint
		if (i == 1) {
			LGraph::out_edge_iterator e0, e1;
                        for (boost::tie(e0, e1) = boost::out_edges(r[i].get_vertex(), m_lgraph); e0 != e1; ++e0) {
				LEdge& ee(m_lgraph[*e0]);
				if (!ee.is_match(airwaymatchlevel))
					continue;
				ee.set_solution(false);
			}
		}
		// kill other vertices with the same ident
		{
			unsigned int superidx(m_lgraph[r[i].get_vertex()].get_superindex());
			const Glib::ustring& ident(m_lsupervertices[superidx].get_ident());
			for (unsigned int sidx(0); sidx < m_lsupervertices.size(); ++sidx) {
				if (sidx == superidx)
					continue;
				const LSuperVertex& sv(m_lsupervertices[sidx]);
				if (sv.get_ident() != ident)
					continue;
				for (LGraph::vertex_descriptor v0(sv.get_v0()); v0 <= sv.get_v1(); ++v0) {
					LGraph::out_edge_iterator e0, e1;
					for (boost::tie(e0, e1) = boost::out_edges(v0, m_lgraph); e0 != e1; ++e0) {
						LEdge& ee(m_lgraph[*e0]);
						ee.set_solution(false);
						if (false) {
							std::ostringstream oss;
							oss << "k shortest path: SV " << sidx
							    << " Killing edge " << lgraphvertexname(boost::source(*e0, m_lgraph))
							    << '(' << (unsigned int)boost::source(*e0, m_lgraph) << ')'
							    << "--" << lgraphindex2airway(ee.get_ident(), true)
							    << "->" << lgraphvertexname(boost::target(*e0, m_lgraph))
							    << '(' << (unsigned int)boost::target(*e0, m_lgraph) << ')';
							m_signal_log(log_debug0, oss.str());
						}
					}
				}
			}
		}
		LRoute r1;
		if (false) {
			std::vector<LGraph::vertex_descriptor> predecessors(boost::num_vertices(m_lgraph), 0);
			dijkstra_shortest_paths(fg, r[i].get_vertex(), boost::weight_map(boost::get(&LEdge::m_metric, m_lgraph)).predecessor_map(&predecessors[0]));
			{
				LGraph::vertex_descriptor u(r[i].get_vertex());
				LGraph::vertex_descriptor v(sdest.get_v0());
				while (v != u) {
					LGraph::vertex_descriptor uu(predecessors[v]);
					if (uu == v) {
						r1.clear();
						break;
					}
					std::pair<LGraph::edge_descriptor,bool> e(lgraphfindbestedge(uu, v, true));
					if (!e.second) {
						r1.clear();
						break;
					}
					r1.insert(r1.begin(), LVertexEdge(e.first, m_lgraph));
					// check that weight map exists and is not infinity
					LEdge& ee(m_lgraph[e.first]);
					if (!ee.is_solution()) {
						if (true) {
							std::ostringstream oss;
							oss << "k shortest path: Leg " << lgraphvertexname(boost::source(e.first, m_lgraph))
							    << "--" << lgraphindex2airway(ee.get_ident(), true) << "->"
							    << lgraphvertexname(boost::target(e.first, m_lgraph)) << " does not exist in filtered graph";
							m_signal_log(log_debug0, oss.str());
						}
						r1.clear();
						break;
					}
					v = uu;
				}
			}
			if (r1.size() >= 2)
				r1.insert(r1.begin(), r.begin(), r.begin() + (i+1));
		} else {
			r1 = lgraphdijkstra(r[i].get_vertex(), sdest.get_v0(), r, true);
		}
		if (false && tree) {
			LGraph::out_edge_iterator e0, e1;
			for (boost::tie(e0, e1) = boost::out_edges(r[i].get_vertex(), m_lgraph); e0 != e1; ++e0) {
				LEdge& ee(m_lgraph[*e0]);
				if (boost::target(*e0, m_lgraph) == r[i+1].get_vertex() &&
				    ee.get_ident() == r[i+1].get_edge())
					continue;
				ee.set_solution(false);
			}
		}
		// kill all out-edges to avoid revisiting a node already in the solution
		{
			unsigned int superidx(m_lgraph[r[i].get_vertex()].get_superindex());
			const LSuperVertex& sv(m_lsupervertices[superidx]);
			for (LGraph::vertex_descriptor v0(sv.get_v0()); v0 <= sv.get_v1(); ++v0) {
				LGraph::out_edge_iterator e0, e1;
				for (boost::tie(e0, e1) = boost::out_edges(v0, m_lgraph); e0 != e1; ++e0) {
					LEdge& ee(m_lgraph[*e0]);
					ee.set_solution(false);
					if (false) {
						std::ostringstream oss;
						oss << "k shortest path: SV " << superidx
						    << " Killing edge " << lgraphvertexname(boost::source(*e0, m_lgraph))
						    << '(' << (unsigned int)boost::source(*e0, m_lgraph) << ')'
						    << "--" << lgraphindex2airway(ee.get_ident(), true)
						    << "->" << lgraphvertexname(boost::target(*e0, m_lgraph))
						    << '(' << (unsigned int)boost::target(*e0, m_lgraph) << ')';
						m_signal_log(log_debug0, oss.str());
					}
				}
			}
		}
		if (r1.size() < 2)
			continue;
		lgraphupdatemetric(r1);
		if (is_metric_invalid(r1.get_metric())) {
			if (true) {
				std::ostringstream oss;
				oss << "k shortest path: Route";
		        	for (LRoute::const_iterator ri(r1.begin()), re(r1.end()); ri != re; ++ri) {
					if (ri->get_edge() != airwaymatchnone)
						oss << ' ' << lgraphindex2airway(ri->get_edge(), true);
					oss << ' ' << lgraphvertexname(ri->get_vertex());
				}
				oss << " has no metric";
				m_signal_log(log_debug0, oss.str());
			}
			continue;
		}
		if (true) {
			std::ostringstream oss;
			oss << "k shortest path: Route";
			for (LRoute::const_iterator ri(r1.begin()), re(r1.end()); ri != re; ++ri) {
				if (ri->get_edge() != airwaymatchnone)
					oss << ' ' << lgraphindex2airway(ri->get_edge(), true);
				oss << ' ' << lgraphvertexname(ri->get_vertex()) << '(' << (unsigned int)ri->get_vertex() << ')';
			}
			oss << " metric " << r1.get_metric() << " added to solution pool";
			m_signal_log(log_debug0, oss.str());
		}
		m_solutionpool.insert(r1);
	}
	// check for too many solutions in the pool
	while (!m_solutionpool.empty() && !m_solutionpool.begin()->is_existing(m_lgraph))
		m_solutionpool.erase(m_solutionpool.begin());
	{
		static const lgraphsolutionpool_t::size_type maxsz(16384);
		lgraphsolutionpool_t::size_type sz(m_solutionpool.size());
		if (sz > maxsz) {
			lgraphsolutionpool_t::size_type n(0);
			for (lgraphsolutionpool_t::iterator i(m_solutionpool.begin()), e(m_solutionpool.end()); i != e; ) {
				lgraphsolutionpool_t::iterator i2(i);
				++i;
				if (n >= maxsz || !i2->is_existing(m_lgraph)) {
					m_solutionpool.erase(i2);
					continue;
				}
				++n;
			}
		}
	}
	if (m_solutionpool.empty())
		return LRoute();
	return *m_solutionpool.begin();
}

#endif

void CFMUAutoroute::lgraphroute(void)
{
	for (;;) {
		if (m_superdep >= m_lsupervertices.size() || m_superdest >= m_lsupervertices.size() ||
		    m_superdep == m_superdest) {
			m_signal_log(log_debug0, "Invalid departure and/or destination, stopping...");
			stop();
			return;
		}
		// dispatch events, check for stop command
		{
			
			Glib::RefPtr<Glib::MainContext> ctx(Glib::MainContext::get_default());
			while (ctx->iteration(false));
			if (!m_running) {
				m_signal_log(log_debug0, "Interrupted by user, stopping...");
				return;
			}
		}
		lgraphclearcurrent();
		LRoute r;
		{
			Glib::TimeVal tv;
			tv.assign_current_time();
			r = lgraphnextsolution();
			{
				Glib::TimeVal tv1;
				tv1.assign_current_time();
				tv = tv1 - tv;
			}
			if (true) {
				std::ostringstream oss;
				oss << "Next route computation: " << std::fixed << std::setprecision(3) << tv.as_double()
				    << "s, solution pool size " << m_solutionpool.size();
				m_signal_log(log_debug0, oss.str());
			}
		}
		if (r.size() < 2) {
			if (true) {
				const LSuperVertex& sdep(m_lsupervertices[m_superdep]);
				const LSuperVertex& sdest(m_lsupervertices[m_superdest]);
				std::ostringstream oss;
				oss << "No route found; "
				    << boost::out_degree(sdep.get_v0(), m_lgraph) << " SID edges, "
				    << boost::in_degree(sdest.get_v0(), m_lgraph) << " STAR edges";
				m_signal_log(log_debug0, oss.str());
			}
			stop();
			return;
		}
		if (!lgraphsetcurrent(r)) {
			stop();
			return;
		}
		if (r.is_repeated_nodes(m_lgraph, m_lsupervertices)) {
			if (true) {
				std::ostringstream oss;
				oss << "Flight Plan has duplicate points " << lgraphprint(r);
				m_signal_log(log_debug0, oss.str());
			}
			continue;
		}
		if (r.size() >= 4 && r[2].is_match(airwaymatchlevel)) {
			if (true) {
				std::ostringstream oss;
				oss << "Flight Plan contains level change on SID endpoint " << lgraphprint(r);
		        	m_signal_log(log_debug0, oss.str());
			}
			continue;
		}
		m_signal_log(log_fplproposal, get_simplefpl());
		bool nochange(true);
		if (get_tfr_enabled()) {
			TrafficFlowRestrictions::Result res;
			bool work(lgraphlocaltfrcheck(res));
			nochange = res;
			if (!nochange) {
				LGMandatory m(lgraphmandatoryroute(res));
				if (!m.empty()) {
					if (true) {
						m_signal_log(log_debug0, "Trying Mandatory Routing");
						for (unsigned int i = 0; i < m.size(); ++i) {
							const LGMandatoryAlternative& alt(m[i]);
							{
								std::ostringstream oss;
								oss << "  Mandatory Rule " << i;
								m_signal_log(log_debug0, oss.str());
							}
							for (unsigned int j = 0; j < alt.size(); ++j) {
								const LGMandatorySequence& seq(alt[j]);
								std::ostringstream oss;
								oss << "    ";
								for (unsigned int k = 0; k < seq.size(); ++k) {
									const LGMandatoryPoint& pt(seq[k]);
									const LVertex& vv(m_lgraph[pt.get_v0()]);
									const LSuperVertex& sv(m_lsupervertices[vv.get_superindex()]);
									oss << sv.get_ident();
									unsigned int pi0(pt.get_v0() - sv.get_v0());
									unsigned int pi1(pt.get_v1() - sv.get_v0());
									if (pi0 < m_performance.size() && pi1 < m_performance.size())
										oss << " F" << std::setw(3) << std::setfill('0') << (m_performance[pi0].get_altitude() / 100)
										    << "..F" << std::setw(3) << std::setfill('0') << ((m_performance[pi1].get_altitude() + 99) / 100);
									if (k + 1 < seq.size())
										oss << ' ' << lgraphindex2airway(pt.get_airway(), true) << ' ';
								}
								m_signal_log(log_debug0, oss.str());
							}
						}
					}
					const LSuperVertex& sdep(m_lsupervertices[m_superdep]);
					const LSuperVertex& sdest(m_lsupervertices[m_superdest]);
					r = lgraphdijkstra(sdep.get_v0(), sdest.get_v0(), LRoute(), false, m);
					if (r.empty()) {
						m_signal_log(log_debug0, "No Mandatory Route found");
					} else {
						lgraphupdatemetric(r);
						lgraphinsertsolutiontree(r);
						if (true) {
							std::ostringstream oss;
							oss << "Mandatory Route " << lgraphprint(r) << " metric " << r.get_metric();
							m_signal_log(log_debug0, oss.str());
						}
						if (lgraphsetcurrent(r)) {
							m_signal_log(log_fplproposal, get_simplefpl());
							work = lgraphlocaltfrcheck(res) || work;
							nochange = res;
						}
					}
				}
			}
		}
		if (false) {
			Glib::RefPtr<Glib::MainContext> ctx(Glib::MainContext::get_default());
			ctx->iteration(false);
		}
		if (!m_running) {
			m_signal_log(log_debug0, "Interrupted by user, stopping...");
			return;
		}
		if (nochange)
			break;
		m_signal_log(log_fplvalidation, "");
		++m_iterationcount[0];
		if (true) {
			Glib::signal_idle().connect_once(sigc::mem_fun(*this, &CFMUAutoroute::lgraphroute), Glib::PRIORITY_HIGH_IDLE);
			return;
		}
	}
	lgraphsubmitroute();
}

bool CFMUAutoroute::lgraphlocaltfrcheck(TrafficFlowRestrictions::Result& res)
{
	if (!get_tfr_enabled()) {
		res = TrafficFlowRestrictions::Result();
		return false;
	}
	{
		Glib::TimeVal tv;
		tv.assign_current_time();
		res = m_tfr.check_fplan(m_route, 'G', m_aircraft.get_aircrafttype(), m_aircraft.get_equipment(),
					m_airportdb, m_navaiddb, m_waypointdb, m_airwaydb, m_airspacedb);
		{
			Glib::TimeVal tv1;
			tv1.assign_current_time();
			tv = tv1 - tv;
		}
		if (true) {
			std::ostringstream oss;
			oss << "Local TFR Check: " << std::fixed << std::setprecision(3) << tv.as_double()
			    << 's';
			m_signal_log(log_debug0, oss.str());
		}
	}
	// Log Messages
	bool nochange(res);
	for (unsigned int i = 0; i < res.messages_size(); ++i) {
		const TrafficFlowRestrictions::Message& msg(res.get_message(i));
		logmessage(msg);
		if (!nochange)
			lgraphtfrmessage(msg, res.get_fplan());
	}
	bool work(false);
	typedef TrafficFlowRestrictions::RuleResult::set_t set_t;
	for (unsigned int i = 0; i < res.size(); ++i) {
		const TrafficFlowRestrictions::RuleResult& rule(res[i]);
		{
			std::ostringstream oss;
			oss << rule.get_rule() << ' ';
			switch (rule.get_codetype()) {
			case TrafficFlowRestrictions::RuleResult::codetype_mandatory:
				oss << 'M';
				break;

			case TrafficFlowRestrictions::RuleResult::codetype_forbidden:
				oss << 'F';
				break;

			case TrafficFlowRestrictions::RuleResult::codetype_closed:
				oss << 'C';
				break;

			default:
				oss << '?';
				break;
			}
			if (rule.is_dct())
				oss << " (D)";
			if (!rule.get_vertexset().empty()) {
				oss << " V:";
				bool subseq(false);
				for (set_t::const_iterator i(rule.get_vertexset().begin()), e(rule.get_vertexset().end()); i != e; ++i) {
					if (subseq)
						oss << ',';
					subseq = true;
					oss << *i;
				}
			}
			if (!rule.get_edgeset().empty()) {
				oss << " E:";
				bool subseq(false);
				for (set_t::const_iterator i(rule.get_edgeset().begin()), e(rule.get_edgeset().end()); i != e; ++i) {
					if (subseq)
						oss << ',';
					subseq = true;
					oss << *i;
				}
			}
			if (!rule.get_reflocset().empty()) {
				oss << " L:";
				bool subseq(false);
				for (set_t::const_iterator i(rule.get_reflocset().begin()), e(rule.get_reflocset().end()); i != e; ++i) {
					if (subseq)
						oss << ',';
					subseq = true;
					oss << *i;
				}
			}
			for (unsigned int i = 0; i < rule.size(); ++i) {
				const TrafficFlowRestrictions::RuleResult::Alternative& alt(rule[i]);
				if (i)
					oss << " /";
				if (!alt.get_vertexset().empty()) {
					oss << " V:";
					bool subseq(false);
					for (set_t::const_iterator i(alt.get_vertexset().begin()), e(alt.get_vertexset().end()); i != e; ++i) {
						if (subseq)
							oss << ',';
						subseq = true;
						oss << *i;
					}
				}
				if (!alt.get_edgeset().empty()) {
					oss << " E:";
					bool subseq(false);
					for (set_t::const_iterator i(alt.get_edgeset().begin()), e(alt.get_edgeset().end()); i != e; ++i) {
						if (subseq)
							oss << ',';
						subseq = true;
						oss << *i;
					}
				}
				for (unsigned int j = 0; j < alt.size(); ++j) {
					const TrafficFlowRestrictions::RuleResult::Alternative::Sequence& el(alt[j]);
					oss << ' ' << el.get_type_string();
					switch (el.get_type()) {
					case TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_airway:
						oss << ' ' << el.get_ident();
						break;

					case TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_sid:
						oss << ' ' << el.get_ident() << ' ' << el.get_airport();
						break;

					case TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_star:
						oss << ' ' << el.get_ident() << ' ' << el.get_airport();
						break;

					case TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_airspace:
						oss << ' ' << el.get_ident() << '/'
						    << AirspacesDb::Airspace::get_class_string(el.get_bdryclass(), AirspacesDb::Airspace::typecode_ead);
						break;

					case TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_dct:
					case TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_point:
					default:
						break;
					}
					oss << " F" << std::setw(3) << std::setfill('0') << ((el.get_lower_alt() + 50) / 100) << "...F"
					    << std::setw(3) << std::setfill('0') << ((std::min((int32_t)el.get_upper_alt(), (int32_t)99900) + 50) / 100);
					if (el.get_type() == TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_airway ||
					    el.get_type() == TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_dct ||
					    el.get_type() == TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_point) {
						const TrafficFlowRestrictions::RuleResult::Alternative::Sequence::RulePoint& pt(el.get_point0());
						oss << ' ' << pt.get_ident();
					}
					if (el.get_type() == TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_airway ||
					    el.get_type() == TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_dct) {
						const TrafficFlowRestrictions::RuleResult::Alternative::Sequence::RulePoint& pt(el.get_point1());
						oss << ' ' << pt.get_ident();
					}
				}
			}
			m_signal_log(log_graphrule, oss.str());
			m_signal_log(log_graphruledesc, rule.get_desc());
			m_signal_log(log_graphruleoprgoal, rule.get_oprgoal());
			m_signal_log(log_fplvalidation, oss.str());
		}
		if (!nochange)
			work = lgraphtfrruleresult(rule, res.get_fplan()) || work;
	}
	return work;
}

CFMUAutoroute::LGMandatory CFMUAutoroute::lgraphmandatoryroute(const TrafficFlowRestrictions::Result& res)
{
	// check whether the remaining violated rules are all mandatory and can be handled
	if (!res.size())
		return LGMandatory();
	LGMandatory m;
	for (unsigned int i = 0; i < res.size(); ++i) {
		const TrafficFlowRestrictions::RuleResult& rule(res[i]);
		if (rule.get_codetype() != TrafficFlowRestrictions::RuleResult::codetype_mandatory)
			return LGMandatory();
		m.push_back(LGMandatoryAlternative());
		for (unsigned int i = 0; i < rule.size(); ++i) {
			const TrafficFlowRestrictions::RuleResult::Alternative& alt(rule[i]);
			if (!alt.size())
				continue;
			bool altok(true);
			LGMandatorySequence mseq;
			for (unsigned int j = 0; j < alt.size(); ++j) {
				const TrafficFlowRestrictions::RuleResult::Alternative::Sequence& seq(alt[j]);
				unsigned int superidx0(~0UL), superidx1(~0UL);
				unsigned int pimin(std::numeric_limits<unsigned int>::max()), pimax(std::numeric_limits<unsigned int>::min());
				lgraphairwayindex_t awy(seq.get_type() == TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_dct ? airwaydct : airwaymatchnone);
				switch (seq.get_type()) {
				default:
					altok = false;
					j = ~0U;
					break;

				case TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_airway:
					awy = lgraphairway2index(seq.get_ident(), false);
					if (awy >= airwaydct) {
						altok = false;
						j = ~0U;
						break;
					}
					// fall through

				case TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_dct:
					for (unsigned int i = 0, n = m_lsupervertices.size(); i < n; ++i) {
						if (m_lsupervertices[i].get_ident() != seq.get_point1().get_ident())
							continue;
						if (m_lsupervertices[i].get_coord().simple_distance_nmi(seq.get_point1().get_coord()) > 0.1)
							continue;
						superidx1 = i;
						break;
					}
					if (superidx1 >= m_lsupervertices.size()) {
						altok = false;
						j = ~0U;
						break;
					}
					// fall through

				case TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_point:
					for (unsigned int i = 0, n = m_lsupervertices.size(); i < n; ++i) {
						if (m_lsupervertices[i].get_ident() != seq.get_point0().get_ident())
							continue;
						if (m_lsupervertices[i].get_coord().simple_distance_nmi(seq.get_point0().get_coord()) > 0.1)
							continue;
						superidx0 = i;
						break;
					}
					if (superidx0 >= m_lsupervertices.size()) {
						altok = false;
						j = ~0U;
						break;
					}
					for (unsigned int pi = 0; pi < m_performance.size(); ++pi) {
						const Performance& perf(m_performance[pi]);
						if (perf.get_altitude() < seq.get_lower_alt() || perf.get_altitude() > seq.get_upper_alt())
							continue;
						pimin = std::min(pimin, pi);
 						pimax = std::max(pimax, pi);
					}
					if (pimin > pimax) {
						altok = false;
						j = ~0U;
						break;
					}
					const LSuperVertex& super0(m_lsupervertices[superidx0]);
					if (super0.get_v1() -  super0.get_v0() < pimax)
						mseq.push_back(LGMandatoryPoint(super0.get_v0(), super0.get_v0(), awy));
					else
						mseq.push_back(LGMandatoryPoint(super0.get_v0() + pimin, super0.get_v0() + pimax, awy));
					if (superidx1 < m_lsupervertices.size()) {
						const LSuperVertex& super1(m_lsupervertices[superidx1]);
						if (super1.get_v1() -  super1.get_v0() < pimax)
							mseq.push_back(LGMandatoryPoint(super1.get_v0(), super1.get_v0(), airwaymatchnone));
						else
							mseq.push_back(LGMandatoryPoint(super1.get_v0() + pimin, super1.get_v0() + pimax, airwaymatchnone));
					}
					break;
				}
			}
			if (!altok || mseq.empty())
				continue;
			for (unsigned int i = 1, n = mseq.size(); i < n;) {
				const LVertex& v0(m_lgraph[mseq[i-1].get_v0()]);
				const LVertex& v1(m_lgraph[mseq[i].get_v0()]);
				if (v0.get_superindex() != v1.get_superindex()) {
					++i;
					continue;
				}
				mseq.erase(mseq.begin() + i);
				n = mseq.size();
			}
			m.back().push_back(mseq);
		}
		if (m.back().empty())
			return LGMandatory();
	}
	return m;
}

void CFMUAutoroute::lgraphsubmitroute(void)
{
	++m_iterationcount[1];
	if (!m_childrun) {
		child_close();
		child_run();
	}
	if (!m_childrun) {
		m_signal_log(log_debug0, "Cannot restart child, stopping...");
		stop();
		return;
	}
	m_childchanstdin->write(get_simplefpl() + "\n");
	m_childchanstdin->flush();
	m_childtimeoutcount = 0;
	m_connchildtimeout.disconnect();
	m_connchildtimeout = Glib::signal_timeout().connect(sigc::mem_fun(*this, &CFMUAutoroute::child_timeout), 60000);
}

bool CFMUAutoroute::child_timeout(void)
{
	m_connchildtimeout.disconnect();
	if (!m_running) {
		m_signal_log(log_debug0, "Interrupted by user, stopping...");
		return false;
	}
	++m_childtimeoutcount;
	if (m_childtimeoutcount > 5) {
		m_signal_log(log_debug0, "Child timed out, stopping...");
		stop();
		child_close();
		return false;
	}
	child_close();
	child_run();
	if (!m_childrun) {
		m_signal_log(log_debug0, "Cannot restart child after timeout, stopping...");
		stop();
		return false;
	}		
	m_childchanstdin->write(get_simplefpl() + "\n");
	m_childchanstdin->flush();
	m_connchildtimeout = Glib::signal_timeout().connect(sigc::mem_fun(*this, &CFMUAutoroute::child_timeout), 60000);
	return true;
}

bool CFMUAutoroute::lgraphtfrruleresult(const TrafficFlowRestrictions::RuleResult& rule, const TrafficFlowRestrictions::Fpl& fpl)
{
	// actually modify the graph
	typedef TrafficFlowRestrictions::RuleResult::set_t set_t;
	switch (rule.get_codetype()) {
	case TrafficFlowRestrictions::RuleResult::codetype_forbidden:
	{
		bool work(false);
		for (unsigned int j = 0; j < rule.size(); ++j) {
			const TrafficFlowRestrictions::RuleResult::Alternative& alt(rule[j]);
			if (alt.get_edgeset().empty()) {
				unsigned int i;
				{
					set_t::const_iterator it(alt.get_vertexset().begin()), e(alt.get_vertexset().end());
					if (it == e)
						continue;
					i = *it;
					++it;
					if (it != e)
						continue;
				}
				if (i >= fpl.size())
					continue;
				if (!rule.get_edgeset().empty())
					continue;
				{
					set_t vset, vset1;
					vset1.insert(i);
					std::set_difference(alt.get_vertexset().begin(), alt.get_vertexset().end(), vset1.begin(), vset1.end(),
							    std::insert_iterator<set_t>(vset, vset.begin()));
					if (!vset.empty())
						continue;
					vset.clear();
					vset1.insert(0U);
					vset1.insert(fpl.size() - 1U);
					std::set_difference(rule.get_vertexset().begin(), rule.get_vertexset().end(), vset1.begin(), vset1.end(),
							    std::insert_iterator<set_t>(vset, vset.begin()));
					if (!vset.empty())
						continue;
				}
				const TrafficFlowRestrictions::FplWpt& wpt(fpl[i]);
				work = lgraphdisconnectsolutionvertex(wpt.get_ident()) || work;
				continue;
			}
			unsigned int i;
			{
				set_t::const_iterator it(alt.get_edgeset().begin()), e(alt.get_edgeset().end());
				if (it == e)
					continue;
				i = *it;
				++it;
				if (it != e)
					continue;
			}
			if (i + 1U >= fpl.size())
				continue;
			{
				set_t vset, vset1;
				vset1.insert(i);
				vset1.insert(i + 1U);
				std::set_difference(alt.get_vertexset().begin(), alt.get_vertexset().end(), vset1.begin(), vset1.end(),
						    std::insert_iterator<set_t>(vset, vset.begin()));
				if (!vset.empty())
					continue;
				vset.clear();
				vset1.insert(0U);
				vset1.insert(fpl.size() - 1U);
				std::set_difference(rule.get_vertexset().begin(), rule.get_vertexset().end(), vset1.begin(), vset1.end(),
						    std::insert_iterator<set_t>(vset, vset.begin()));
				if (!vset.empty())
					continue;
				vset.clear();
				std::set_difference(rule.get_edgeset().begin(), rule.get_edgeset().end(),
						    alt.get_edgeset().begin(), alt.get_edgeset().end(),
						    std::insert_iterator<set_t>(vset, vset.begin()));
				if (!vset.empty())
					continue;
			}
			const TrafficFlowRestrictions::FplWpt& wpt0(fpl[i]);
			const TrafficFlowRestrictions::FplWpt& wpt1(fpl[i + 1U]);
			work = lgraphdisconnectsolutionedge(wpt0.get_ident(), wpt1.get_ident()) || work;
		}
		if (work)
			return true;
		break;
	}

	case TrafficFlowRestrictions::RuleResult::codetype_closed:
	{
		if (rule.get_edgeset().empty()) {
			unsigned int i;
			{
				set_t vset, vset1;
				vset1.insert(0U);
				vset1.insert(fpl.size() - 1U);
				std::set_difference(rule.get_vertexset().begin(), rule.get_vertexset().end(), vset1.begin(), vset1.end(),
						    std::insert_iterator<set_t>(vset, vset.begin()));
				set_t::const_iterator it(vset.begin()), e(vset.end());
				if (it == e)
					break;
				i = *it;
				++it;
				if (it != e)
					break;
			}
			if (i >= fpl.size())
				break;
			const TrafficFlowRestrictions::FplWpt& wpt(fpl[i]);
			if (lgraphdisconnectsolutionvertex(wpt.get_ident()))
				return true;
			break;
		}
		unsigned int i;
		{
			set_t::const_iterator it(rule.get_edgeset().begin()), e(rule.get_edgeset().end());
			if (it == e)
				break;
			i = *it;
			++it;
			if (it != e)
				break;
		}
		if (i + 1U >= fpl.size())
			break;
		{
			set_t vset, vset1;
			vset1.insert(i);
			vset1.insert(i + 1U);
			vset1.insert(0U);
			vset1.insert(fpl.size() - 1U);
			std::set_difference(rule.get_vertexset().begin(), rule.get_vertexset().end(), vset1.begin(), vset1.end(),
					    std::insert_iterator<set_t>(vset, vset.begin()));
			if (!vset.empty())
				break;
		}
		const TrafficFlowRestrictions::FplWpt& wpt0(fpl[i]);
		const TrafficFlowRestrictions::FplWpt& wpt1(fpl[i + 1U]);
		if (lgraphdisconnectsolutionedge(wpt0.get_ident(), wpt1.get_ident()))
			return true;
		break;
	}

	default:
	case TrafficFlowRestrictions::RuleResult::codetype_mandatory:
		if (rule.is_dct()) {
			bool work(false);
			for (set_t::const_iterator it(rule.get_edgeset().begin()), e(rule.get_edgeset().end()); it != e; ++it) {
				unsigned int i(*it);
				if (i + 1U >= fpl.size())
					continue;
				const TrafficFlowRestrictions::FplWpt& wpt0(fpl[i]);
				const TrafficFlowRestrictions::FplWpt& wpt1(fpl[i + 1U]);
				work = lgraphdisconnectsolutionedge(wpt0.get_ident(), wpt1.get_ident()) || work;
			}
			if (work)
				return true;
		}
		break;
	}
	double scale(1.05);
	bool work(false);
	switch (rule.get_codetype()) {
	case TrafficFlowRestrictions::RuleResult::codetype_mandatory:
		scale = 0.95;
		// fall through
		
	case TrafficFlowRestrictions::RuleResult::codetype_forbidden:
		for (unsigned int j = 0; j < rule.size(); ++j) {
			const TrafficFlowRestrictions::RuleResult::Alternative& alt(rule[j]);
			if (alt.get_edgeset().empty()) {
				for (set_t::const_iterator i(alt.get_vertexset().begin()), e(alt.get_vertexset().end()); i != e; ++i) {
					if (*i >= fpl.size())
						continue;
					const TrafficFlowRestrictions::FplWpt& wpt(fpl[*i]);
					work = lgraphscalesolutionvertex(wpt.get_ident(), scale) || work;
				}
				continue;
			}
			for (set_t::const_iterator i(alt.get_edgeset().begin()), e(alt.get_edgeset().end()); i != e; ++i) {
				if (*i >= fpl.size() + 1U)
					continue;
				const TrafficFlowRestrictions::FplWpt& wpt0(fpl[*i]);
				const TrafficFlowRestrictions::FplWpt& wpt1(fpl[*i + 1U]);
				work = lgraphscalesolutionedge(wpt0.get_ident(), wpt1.get_ident(), scale) || work;
			}
		}
		break;

	case TrafficFlowRestrictions::RuleResult::codetype_closed:
		if (rule.get_edgeset().empty()) {
			for (set_t::const_iterator i(rule.get_vertexset().begin()), e(rule.get_vertexset().end()); i != e; ++i) {
				if (*i >= fpl.size())
					continue;
				const TrafficFlowRestrictions::FplWpt& wpt(fpl[*i]);
				work = lgraphscalesolutionvertex(wpt.get_ident(), scale) || work;
			}
			break;
		}
		for (set_t::const_iterator i(rule.get_edgeset().begin()), e(rule.get_edgeset().end()); i != e; ++i) {
			if (*i >= fpl.size() + 1U)
				continue;
			const TrafficFlowRestrictions::FplWpt& wpt0(fpl[*i]);
			const TrafficFlowRestrictions::FplWpt& wpt1(fpl[*i + 1U]);
			work = lgraphscalesolutionedge(wpt0.get_ident(), wpt1.get_ident(), scale) || work;
		}
		break;

	default:
		break;
	}
	if (work)
		lgraphmodified();
	return work;
}

void CFMUAutoroute::lgraphtfrmessage(const TrafficFlowRestrictions::Message& msg, const TrafficFlowRestrictions::Fpl& fpl)
{
	if (msg.get_type() != TrafficFlowRestrictions::Message::type_error)
		return;
	// actually modify the graph
	typedef TrafficFlowRestrictions::RuleResult::set_t set_t;
	if (msg.get_edgeset().empty()) {
		unsigned int i;
		{
			set_t vset, vset1;
			vset1.insert(0U);
			vset1.insert(fpl.size() - 1U);
			std::set_difference(msg.get_vertexset().begin(), msg.get_vertexset().end(), vset1.begin(), vset1.end(),
					    std::insert_iterator<set_t>(vset, vset.begin()));
			set_t::const_iterator it(vset.begin()), e(vset.end());
			if (it == e)
				return;
			i = *it;
			++it;
			if (it != e)
				return;
		}
		if (i >= fpl.size())
			return;
		const TrafficFlowRestrictions::FplWpt& wpt(fpl[i]);
		if (lgraphdisconnectsolutionvertex(wpt.get_ident()))
			return;
		return;
	}
	unsigned int i;
	{
		set_t::const_iterator it(msg.get_edgeset().begin()), e(msg.get_edgeset().end());
		if (it == e)
			return;
		i = *it;
		++it;
		if (it != e)
			return;
	}
	if (i + 1U >= fpl.size())
		return;
	{
		set_t vset, vset1;
		vset1.insert(i);
		vset1.insert(i + 1U);
		vset1.insert(0U);
		vset1.insert(fpl.size() - 1U);
		std::set_difference(msg.get_vertexset().begin(), msg.get_vertexset().end(), vset1.begin(), vset1.end(),
				    std::insert_iterator<set_t>(vset, vset.begin()));
		if (!vset.empty())
			return;
	}
	const TrafficFlowRestrictions::FplWpt& wpt0(fpl[i]);
	const TrafficFlowRestrictions::FplWpt& wpt1(fpl[i + 1U]);
	if (lgraphdisconnectsolutionedge(wpt0.get_ident(), wpt1.get_ident()))
		return;
}

bool CFMUAutoroute::lgraphkilllevelchg(unsigned int superindex)
{
	if (superindex >= m_lsupervertices.size())
		return false;
	return lgraphkilllevelchg(m_lsupervertices[superindex]);
}

bool CFMUAutoroute::lgraphkilllevelchg(const LSuperVertex& super)
{
	if (super.get_v0() == super.get_v1())
		return false;
	bool work(false);
	for (LGraph::vertex_descriptor v0(super.get_v0()); v0 < super.get_v1(); ++v0) {
		LGraph::vertex_descriptor v1(v0);
		for (;;) {
			bool done(true);
			LGraph::out_edge_iterator e0, e1;
			for (tie(e0, e1) = out_edges(v0, m_lgraph); e0 != e1; ) {
				LGraph::out_edge_iterator ex(e0);
				++e0;
				if (target(*ex, m_lgraph) != v1)
					continue;
				remove_edge(ex, m_lgraph);
				work = true;
				if (lgraphrestartedgescan) {
					done = false;
					break;
				}
			}
			if (done)
				break;
		}
		for (;;) {
			bool done(true);
			LGraph::out_edge_iterator e0, e1;
			for (tie(e0, e1) = out_edges(v1, m_lgraph); e0 != e1; ) {
				LGraph::out_edge_iterator ex(e0);
				++e0;
				if (target(*ex, m_lgraph) != v0)
					continue;
				remove_edge(ex, m_lgraph);
				work = true;
				if (lgraphrestartedgescan) {
					done = false;
					break;
				}
			}
			if (done)
				break;
		}
	}
	return work;
}

Glib::ustring CFMUAutoroute::lgraphvertexname(LGraph::vertex_descriptor v)
{
	const LVertex& vv(m_lgraph[v]);
	const LSuperVertex& sv(m_lsupervertices[vv.get_superindex()]);
	std::ostringstream oss;
	oss << sv.get_ident();
	if (sv.get_v0() == sv.get_v1())
		return oss.str();
	const Performance& perf(m_performance[v - sv.get_v0()]);
	oss << ":F" << std::setw(3) << std::setfill('0') << ((perf.get_altitude() + 50) / 100);
	return oss.str();
}

void CFMUAutoroute::lgraphlogkilledge(LGraph::edge_descriptor e)
{
	const LEdge& ee(m_lgraph[e]);
	std::ostringstream oss;
	oss << "- " << lgraphvertexname(source(e, m_lgraph))
	    << ' ' << lgraphindex2airway(ee.get_ident(), true) << ' '
	    << lgraphvertexname(target(e, m_lgraph));
	m_signal_log(log_graphchange, oss.str());
}

void CFMUAutoroute::lgraphlogaddedge(LGraph::edge_descriptor e)
{
	const LEdge& ee(m_lgraph[e]);
	std::ostringstream oss;
	oss << "+ " << lgraphvertexname(source(e, m_lgraph))
	    << ' ' << lgraphindex2airway(ee.get_ident(), true) << ' '
	    << lgraphvertexname(target(e, m_lgraph));
	m_signal_log(log_graphchange, oss.str());
}

void CFMUAutoroute::lgraphlogrenameedge(LGraph::edge_descriptor e, lgraphairwayindex_t awyname)
{
	const LEdge& ee(m_lgraph[e]);
	std::ostringstream oss;
	oss << "R " << lgraphvertexname(source(e, m_lgraph))
	    << ' ' << lgraphindex2airway(ee.get_ident(), true) << ' '
	    << lgraphvertexname(target(e, m_lgraph)) << " -> " << lgraphindex2airway(awyname, true);
	m_signal_log(log_graphchange, oss.str());
}

void CFMUAutoroute::lgraphlogscaleedge(LGraph::edge_descriptor e, double scale)
{
	const LEdge& ee(m_lgraph[e]);
	std::ostringstream oss;
	oss << "S " << lgraphvertexname(source(e, m_lgraph))
	    << ' ' << lgraphindex2airway(ee.get_ident(), true) << ' '
	    << lgraphvertexname(target(e, m_lgraph)) << " : "
	    << ee.get_metric() << " * " << scale << " -> " << (ee.get_metric() * scale);
	m_signal_log(log_graphchange, oss.str());
}

bool CFMUAutoroute::lgraphhasdctedge(LGraph::edge_descriptor e)
{
	return lgraphhasdctedge(source(e, m_lgraph), target(e, m_lgraph));
}

bool CFMUAutoroute::lgraphhasdctedge(LGraph::vertex_descriptor u, LGraph::vertex_descriptor v)
{
	LGraph::out_edge_iterator e0, e1;
	for (tie(e0, e1) = out_edges(u, m_lgraph); e0 != e1; ++e0) {
		if (target(*e0, m_lgraph) != v)
			continue;
		if (m_lgraph[*e0].is_dct())
			return true;
	}
	return false;
}

bool CFMUAutoroute::lgraphcanadddctedge(LGraph::vertex_descriptor u, LGraph::vertex_descriptor v)
{
	// basic sanity checks
	const LVertex& uu(m_lgraph[u]);
	const LSuperVertex& su(m_lsupervertices[uu.get_superindex()]);
	if (su.get_v0() == su.get_v1())
		return false;
	const LVertex& vv(m_lgraph[v]);
	const LSuperVertex& sv(m_lsupervertices[vv.get_superindex()]);
	if (sv.get_v0() == sv.get_v1())
		return false;;
	unsigned int pi(u - su.get_v0());
	if (pi != v - sv.get_v0() || pi >= m_performance.size())
		return false;
	// check distance
	{
		double dist(su.get_coord().spheric_distance_nmi_dbl(sv.get_coord()));
		if (dist > get_dctlimit())
			return false;
	}
	// check whether there already is a DCT
	if (lgraphhasdctedge(u, v))
		return false;
	// check TFR rules
	if (get_tfr_enabled()) {
		TrafficFlowRestrictions::DctParameters dctrule(su.get_ident(), su.get_coord(),
							       sv.get_ident(), sv.get_coord(),
							       m_performance[pi].get_altitude(), m_performance[pi].get_altitude()+1);
		std::vector<TrafficFlowRestrictions::Message> msg;
		m_tfr.check_dct(msg, dctrule);
		for (unsigned int i = 0; i < msg.size(); ++i)
			logmessage(msg[i]);
		if (dctrule.get_alt(0).is_empty())
			return false;
	}
	return true;
}

bool CFMUAutoroute::lgraphdisconnectvertex(const Glib::ustring& ident, int lvlfrom, int lvlto)
{
	bool work(false);
	Glib::ustring identu(ident.uppercase());
	lvlfrom *= 100;
	lvlto *= 100;
	for (solutionvertices_t::const_iterator svi(m_solutionvertices.begin()), sve(m_solutionvertices.end()); svi != sve; ++svi) {
		const LVertex& vv(m_lgraph[*svi]);
		const LSuperVertex& sv(m_lsupervertices[vv.get_superindex()]);
		if (sv.get_ident() != identu)
			continue;
		// prevent departure or destination from being disconnected if an intersection has the same ident
		if ((lvlfrom == 0 && lvlto == 60000) && (vv.get_superindex() == m_superdep || vv.get_superindex() == m_superdest))
			continue;
		const Performance& perf(m_performance[*svi - sv.get_v0()]);
		if (perf.get_altitude() < lvlfrom || perf.get_altitude() > lvlto)
			continue;
		for (LGraph::vertex_descriptor v0(sv.get_v0()), v1(sv.get_v1()); v0 <= v1; ++v0) {
			for (;;) {
				bool done(true);
				LGraph::out_edge_iterator e0, e1;
				for (tie(e0, e1) = out_edges(v0, m_lgraph); e0 != e1; ) {
					LGraph::out_edge_iterator ex(e0);
					++e0;
					LGraph::vertex_descriptor u(target(*ex, m_lgraph));
					const LVertex& uu(m_lgraph[u]);
					if (vv.get_superindex() == uu.get_superindex())
						continue;
					lgraphlogkilledge(*ex);
					remove_edge(ex, m_lgraph);
					work = true;
					if (lgraphrestartedgescan) {
						done = false;
						break;
					}
				}
				if (done)
					break;
			}
			for (;;) {
				bool done(true);
				LGraph::in_edge_iterator e0, e1;
				for (tie(e0, e1) = in_edges(v0, m_lgraph); e0 != e1; ) {
					LGraph::in_edge_iterator ex(e0);
					++e0;
					LGraph::vertex_descriptor u(source(*ex, m_lgraph));
					const LVertex& uu(m_lgraph[u]);
					if (vv.get_superindex() == uu.get_superindex())
						continue;
					lgraphlogkilledge(*ex);
					remove_edge(*ex, m_lgraph);
					work = true;
					if (lgraphrestartedgescan) {
						done = false;
						break;
					}
				}
				if (done)
					break;
			}
		}
	}
	return work;
}

bool CFMUAutoroute::lgraphdisconnectsolutionvertex(const Glib::ustring& ident)
{
	bool work(false);
	Glib::ustring identu(ident.uppercase());
	for (;;) {
		bool done(true);
		LGraph::edge_iterator e0, e1;
		for (tie(e0, e1) = edges(m_lgraph); e0 != e1;) {
			LGraph::edge_iterator ex(e0);
			++e0;
			LEdge& ee(m_lgraph[*ex]);
			if (!ee.is_solution())
				continue;
			const LVertex& uu(m_lgraph[source(*ex, m_lgraph)]);
			const LSuperVertex& su(m_lsupervertices[uu.get_superindex()]);
			if (su.get_ident() != identu) {
				const LVertex& vv(m_lgraph[target(*ex, m_lgraph)]);
				const LSuperVertex& sv(m_lsupervertices[vv.get_superindex()]);
				if (sv.get_ident() != identu)
					continue;
			}
			lgraphlogkilledge(*ex);
			remove_edge(*ex, m_lgraph);
			work = true;
			if (lgraphrestartedgescan) {
				done = false;
				break;
			}
		}
		if (done)
			break;
	}
	return work;
}

bool CFMUAutoroute::lgraphdisconnectsolutionedge(const Glib::ustring& id0, const Glib::ustring& id1)
{
	bool work(false);
	Glib::ustring id0u(id0.uppercase());
	Glib::ustring id1u(id1.uppercase());
	for (;;) {
		bool done(true);
		LGraph::edge_iterator e0, e1;
		for (tie(e0, e1) = edges(m_lgraph); e0 != e1;) {
			LGraph::edge_iterator ex(e0);
			++e0;
			LEdge& ee(m_lgraph[*ex]);
			if (!ee.is_solution())
				continue;
			LGraph::vertex_descriptor u(source(*ex, m_lgraph));
			const LVertex& uu(m_lgraph[u]);
			const LSuperVertex& su(m_lsupervertices[uu.get_superindex()]);
			if (su.get_ident() != id0u)
				continue;
			LGraph::vertex_descriptor v(target(*ex, m_lgraph));
			const LVertex& vv(m_lgraph[v]);
			const LSuperVertex& sv(m_lsupervertices[vv.get_superindex()]);
			if (sv.get_ident() != id1u)
				continue;
			if (ee.is_match(airwaymatchawy) && lgraphcanadddctedge(u, v)) {
				lgraphlogrenameedge(*ex, airwaydct);
				ee.set_ident(airwaydct);
				ee.set_metric(ee.get_metric() * get_dctpenalty());
				work = true;
				continue;
        		}
			lgraphlogkilledge(*ex);
			remove_edge(*ex, m_lgraph);
			work = true;
			if (lgraphrestartedgescan) {
				done = false;
				break;
			}
		}
		if (done)
			break;
	}
	return work;
}

bool CFMUAutoroute::lgraphscalesolutionvertex(const Glib::ustring& ident, double scale)
{
	bool work(false);
	Glib::ustring identu(ident.uppercase());
	LGraph::edge_iterator e0, e1;
	for (tie(e0, e1) = edges(m_lgraph); e0 != e1;) {
		LGraph::edge_iterator ex(e0);
		++e0;
		LEdge& ee(m_lgraph[*ex]);
		if (!ee.is_solution())
			continue;
		const LVertex& uu(m_lgraph[source(*ex, m_lgraph)]);
		const LSuperVertex& su(m_lsupervertices[uu.get_superindex()]);
		if (su.get_ident() != identu) {
			const LVertex& vv(m_lgraph[target(*ex, m_lgraph)]);
			const LSuperVertex& sv(m_lsupervertices[vv.get_superindex()]);
			if (sv.get_ident() != identu)
				continue;
		}
		double newmetric(ee.get_metric() * scale);
		if (newmetric != ee.get_metric()) {
			lgraphlogscaleedge(*ex, scale);
			work = true;
		}
		ee.set_metric(newmetric);
	}
	return work;
}

bool CFMUAutoroute::lgraphscalesolutionedge(const Glib::ustring& id0, const Glib::ustring& id1, double scale)
{
	bool work(false);
	Glib::ustring id0u(id0.uppercase());
	Glib::ustring id1u(id1.uppercase());
	LGraph::edge_iterator e0, e1;
	for (tie(e0, e1) = edges(m_lgraph); e0 != e1;) {
		LGraph::edge_iterator ex(e0);
		++e0;
		LEdge& ee(m_lgraph[*ex]);
		if (!ee.is_solution())
			continue;
		const LVertex& uu(m_lgraph[source(*ex, m_lgraph)]);
		const LSuperVertex& su(m_lsupervertices[uu.get_superindex()]);
		if (su.get_ident() != id0u)
			continue;
		const LVertex& vv(m_lgraph[target(*ex, m_lgraph)]);
		const LSuperVertex& sv(m_lsupervertices[vv.get_superindex()]);
		if (sv.get_ident() != id1u)
			continue;
		double newmetric(ee.get_metric() * scale);
		if (newmetric != ee.get_metric()) {
			lgraphlogscaleedge(*ex, scale);
			work = true;
		}
		ee.set_metric(newmetric);
	}
	return work;
}

bool CFMUAutoroute::lgraphmodifyedge(const Glib::ustring& id0, const Glib::ustring& id1, lgraphairwayindex_t awyremove,
				     lgraphairwayindex_t awyadd, int lvlfrom, int lvlto, bool bidir)
{
	bool work(false);
	Glib::ustring id0u(id0.uppercase());
	Glib::ustring id1u(id1.uppercase());
	lvlfrom *= 100;
	lvlto *= 100;
	for (solutionvertices_t::const_iterator sui(m_solutionvertices.begin()), sue(m_solutionvertices.end()); sui != sue; ++sui) {
		const LVertex& uu(m_lgraph[*sui]);
		const LSuperVertex& su(m_lsupervertices[uu.get_superindex()]);
		if (su.get_ident() != id0u)
			continue;
		for (solutionvertices_t::const_iterator svi(m_solutionvertices.begin()), sve(m_solutionvertices.end()); svi != sve; ++svi) {
			if (sui == svi)
				continue;
			const LVertex& vv(m_lgraph[*svi]);
			const LSuperVertex& sv(m_lsupervertices[vv.get_superindex()]);
			if (sv.get_ident() != id1u)
				continue;
			double dist(su.get_coord().spheric_distance_nmi_dbl(sv.get_coord()));
			for (unsigned int pi = 0; pi < m_performance.size(); ++pi) {
				const Performance& perf(m_performance[pi]);
				int alt(perf.get_altitude());
				if (alt < lvlfrom || alt > lvlto)
					continue;
				LGraph::vertex_descriptor u(su.get_v0() + pi), v(sv.get_v0() + pi);
				if (u > su.get_v1() || v > sv.get_v1())
					continue;
				bool doadd(awyadd != airwaymatchnone);
				for (;;) {
					bool done(true);
					LGraph::out_edge_iterator e0, e1;
					for (tie(e0, e1) = out_edges(u, m_lgraph); e0 != e1; ) {
						LGraph::out_edge_iterator ex(e0);
						++e0;
						if (target(*ex, m_lgraph) != v)
							continue;
						const LEdge& ee(m_lgraph[*ex]);
						if (ee.is_match(awyadd)) {
							doadd = false;
							continue;
						}
						if (!ee.is_match(awyremove))
							continue;
						lgraphlogkilledge(*ex);
						remove_edge(ex, m_lgraph);
						work = true;
						if (lgraphrestartedgescan) {
							done = false;
							break;
						}
					}
					if (done)
						break;
				}
				if (doadd && awyadd == airwaydct && !lgraphcanadddctedge(u, v))
					doadd = false;
				if (doadd) {
					LGraph::edge_descriptor e;
					bool added;
					double x(dist * perf.get_secpernmi());
					if (is_optfuel())
						x *= perf.get_fuelpersec();
					if (awyadd == airwaydct)
						x *= get_dctpenalty();
					tie(e, added) = boost::add_edge(u, v, LEdge(awyadd, x), m_lgraph);
					if (added)
						lgraphlogaddedge(e);
					work = work || added;
				}
				if (!bidir)
					continue;
				doadd = awyadd != airwaymatchnone;
				for (;;) {
					bool done(true);
					LGraph::out_edge_iterator e0, e1;
					for (tie(e0, e1) = out_edges(v, m_lgraph); e0 != e1; ) {
						LGraph::out_edge_iterator ex(e0);
						++e0;
						if (target(*ex, m_lgraph) != u)
							continue;
						const LEdge& ee(m_lgraph[*ex]);
						if (ee.is_match(awyadd)) {
							doadd = false;
							continue;
						}
						if (!ee.is_match(awyremove))
							continue;
						lgraphlogkilledge(*ex);
						remove_edge(ex, m_lgraph);
						work = true;
						if (lgraphrestartedgescan) {
							done = false;
							break;
						}
					}
					if (done)
						break;
				}
				if (doadd && awyadd == airwaydct && !lgraphcanadddctedge(v, u))
					doadd = false;
				if (doadd) {
					LGraph::edge_descriptor e;
					bool added;
					double x(dist * perf.get_secpernmi());
					if (is_optfuel())
						x *= perf.get_fuelpersec();
					if (awyadd == airwaydct)
						x *= get_dctpenalty();
					tie(e, added) = boost::add_edge(v, u, LEdge(awyadd, x), m_lgraph);
					if (added)
						lgraphlogaddedge(e);
					work = work || added;
				}
			}
		}
	}
	return work;
}

bool CFMUAutoroute::lgraphmodifyawysegment(const Glib::ustring& id0, const Glib::ustring& id1, lgraphairwayindex_t awymatch,
					   lgraphairwayindex_t awyrename, int lvlfrom, int lvlto, bool bidir, bool solutiononly)
{
	if (awymatch == awyrename)
		return false;
	bool work(false);
	Glib::ustring id0u(id0.uppercase());
	Glib::ustring id1u(id1.uppercase());
	lvlfrom *= 100;
	lvlto *= 100;
#if 0
	// construct weight map
	typedef std::map<LGraph::edge_descriptor, double> weights_t;
	weights_t weights;
	weights.clear();
	{
		LGraph::edge_iterator e0, e1;
		for (tie(e0, e1) = edges(m_lgraph); e0 != e1; ++e0) {
			LEdge& ee(m_lgraph[*e0]);
			int w(1);
			if (!ee.is_match(awymatch))
				//w = std::numeric_limits<double>::max();
				w = 1e200;
			weights[*e0] = w;
		}
	}
	boost::associative_property_map<weights_t> weightsmap(weights);
#else
	LGraphAwyEdgeFilter filter(m_lgraph, awymatch);
        typedef boost::filtered_graph<LGraph, LGraphAwyEdgeFilter> fg_t;
        fg_t fg(m_lgraph, filter);
#endif
	for (solutionvertices_t::const_iterator sui(m_solutionvertices.begin()), sue(m_solutionvertices.end()); sui != sue; ++sui) {
		const LVertex& uu(m_lgraph[*sui]);
		const LSuperVertex& su(m_lsupervertices[uu.get_superindex()]);
		if (su.get_ident() != id0u)
			continue;
		for (solutionvertices_t::const_iterator svi(m_solutionvertices.begin()), sve(m_solutionvertices.end()); svi != sve; ++svi) {
			if (sui == svi)
				continue;
			const LVertex& vv(m_lgraph[*svi]);
			const LSuperVertex& sv(m_lsupervertices[vv.get_superindex()]);
			if (sv.get_ident() != id1u)
				continue;
			for (unsigned int pi = 0; pi < m_performance.size(); ++pi) {
				const Performance& perf(m_performance[pi]);
				int alt(perf.get_altitude());
				if (alt < lvlfrom || alt > lvlto)
					continue;
				LGraph::vertex_descriptor u(su.get_v0() + pi), v(sv.get_v0() + pi);
				if (u > su.get_v1() || v > sv.get_v1())
					continue;
				std::vector<LGraph::vertex_descriptor> predecessors(boost::num_vertices(m_lgraph), 0);
#if 0
				dijkstra_shortest_paths(m_lgraph, u, boost::weight_map(weightsmap).
							predecessor_map(&predecessors[0]));
#else
				dijkstra_shortest_paths(fg, u, boost::weight_map(boost::get(&LEdge::m_metric, m_lgraph)).
							predecessor_map(&predecessors[0]));
#endif
				LGraph::vertex_descriptor ve(v);
				while (ve != u) {
					LGraph::vertex_descriptor ue(predecessors[ve]);
					for (;;) {
						bool done(true);
						unsigned int renedges(0);
						LGraph::out_edge_iterator e0, e1;
						for (tie(e0, e1) = out_edges(ue, m_lgraph); e0 != e1; ) {
							LGraph::out_edge_iterator ex(e0);
							++e0;
							if (target(*ex, m_lgraph) != ve)
								continue;
							LEdge& ee(m_lgraph[*ex]);
							if (ee.is_match(awyrename)) {
								++renedges;
								if (renedges >= 2) {
									lgraphlogkilledge(*ex);
									remove_edge(ex, m_lgraph);
									work = true;
									if (lgraphrestartedgescan) {
										done = false;
										break;
									}
								}
							}
							if (!ee.is_match(awymatch))
								continue;
							if (solutiononly && !ee.is_solution())
								continue;
							if (renedges < 1 && awyrename != airwaymatchnone &&
							    (awyrename != airwaydct || lgraphcanadddctedge(ue, ve))) {
								ee.set_ident(awyrename);
								lgraphlogrenameedge(*ex, awyrename);
								++renedges;
								continue;
							}
							lgraphlogkilledge(*ex);
							remove_edge(ex, m_lgraph);
							work = true;
							if (lgraphrestartedgescan) {
								done = false;
								break;
							}
						}
						if (done)
							break;
					}
					if (!bidir) {
						if (ue == ve)
							break;
						ve = ue;
						continue;
					}
					for (;;) {
						bool done(true);
						unsigned int renedges(0);
						LGraph::out_edge_iterator e0, e1;
						for (tie(e0, e1) = out_edges(ve, m_lgraph); e0 != e1; ) {
							LGraph::out_edge_iterator ex(e0);
							++e0;
							if (target(*ex, m_lgraph) != ue)
								continue;
							LEdge& ee(m_lgraph[*ex]);
							if (ee.is_match(awyrename)) {
								++renedges;
								if (renedges >= 2) {
									lgraphlogkilledge(*ex);
									remove_edge(ex, m_lgraph);
									work = true;
									if (lgraphrestartedgescan) {
										done = false;
										break;
									}
									continue;
								}
							}
							if (!ee.is_match(awymatch))
								continue;
							if (solutiononly && !ee.is_solution())
								continue;
							if (renedges < 1 && awyrename != airwaymatchnone &&
							    (awyrename != airwaydct || lgraphcanadddctedge(ve, ue))) {
								ee.set_ident(awyrename);
								lgraphlogrenameedge(*ex, awyrename);
								++renedges;
								continue;
							}
							lgraphlogkilledge(*ex);
							remove_edge(ex, m_lgraph);
							work = true;
							if (lgraphrestartedgescan) {
								done = false;
								break;
							}
						}
						if (done)
							break;
					}
					if (ue == ve)
						break;
					ve = ue;
				}
			}
		}
	}
	return work;
}

bool CFMUAutoroute::lgraphchangeoutgoing(const Glib::ustring& id, lgraphairwayindex_t awyname)
{
	bool work(false);
	Glib::ustring identu(id.uppercase());
	for (solutionvertices_t::const_iterator svi(m_solutionvertices.begin()), sve(m_solutionvertices.end()); svi != sve; ++svi) {
		const LVertex& vv(m_lgraph[*svi]);
		const LSuperVertex& sv(m_lsupervertices[vv.get_superindex()]);
		if (sv.get_ident() != identu)
			continue;
		for (LGraph::vertex_descriptor v0(sv.get_v0()), v1(sv.get_v1()); v0 <= v1; ++v0) {
			for (;;) {
				bool done(true);
				LGraph::out_edge_iterator e0, e1;
				for (tie(e0, e1) = out_edges(v0, m_lgraph); e0 != e1; ) {
					LGraph::out_edge_iterator ex(e0);
					++e0;
					LEdge& ee(m_lgraph[*ex]);
					if (!ee.is_match(awyname))
						continue;
					work = true;
					if (awyname != airwaydct && lgraphcanadddctedge(v0, target(*ex, m_lgraph))) {
						lgraphlogrenameedge(*ex, airwaydct);
						ee.set_ident(airwaydct);
						ee.set_metric(ee.get_metric() * get_dctpenalty());
						continue;
					}
					lgraphlogkilledge(*ex);
					remove_edge(ex, m_lgraph);
					if (lgraphrestartedgescan) {
						done = false;
						break;
					}
				}
				if (done)
					break;
			}
		}
	}
	return work;
}

bool CFMUAutoroute::lgraphchangeincoming(const Glib::ustring& id, lgraphairwayindex_t awyname)
{
	bool work(false);
	Glib::ustring identu(id.uppercase());
	for (solutionvertices_t::const_iterator svi(m_solutionvertices.begin()), sve(m_solutionvertices.end()); svi != sve; ++svi) {
		const LVertex& vv(m_lgraph[*svi]);
		const LSuperVertex& sv(m_lsupervertices[vv.get_superindex()]);
		if (sv.get_ident() != identu)
			continue;
		for (LGraph::vertex_descriptor v0(sv.get_v0()), v1(sv.get_v1()); v0 <= v1; ++v0) {
			for (;;) {
				bool done(true);
				LGraph::in_edge_iterator e0, e1;
				for (tie(e0, e1) = in_edges(v0, m_lgraph); e0 != e1; ) {
					LGraph::in_edge_iterator ex(e0);
					++e0;
					LEdge& ee(m_lgraph[*ex]);
					if (!ee.is_match(awyname))
						continue;
					work = true;
					if (awyname != airwaydct && lgraphcanadddctedge(source(*ex, m_lgraph), v0)) {
						lgraphlogrenameedge(*ex, airwaydct);
						ee.set_ident(airwaydct);
						ee.set_metric(ee.get_metric() * get_dctpenalty());
						continue;
					}
					lgraphlogkilledge(*ex);
					remove_edge(*ex, m_lgraph);
					if (lgraphrestartedgescan) {
						done = false;
						break;
					}
				}
				if (done)
					break;
			}
		}
	}
	return work;
}

bool CFMUAutoroute::lgraphedgetodct(lgraphairwayindex_t awy)
{
	if (awy == airwaydct)
		return false;
	bool work(false);
	for (;;) {
		bool done(true);
		LGraph::edge_iterator e0, e1;
		for (tie(e0, e1) = edges(m_lgraph); e0 != e1; ) {
			LGraph::edge_iterator ex(e0);
			++e0;
			LEdge& ee(m_lgraph[*ex]);
			if (!ee.is_solution())
				continue;
			if (!ee.is_match(awy))
				continue;
			if (lgraphcanadddctedge(source(*ex, m_lgraph), target(*ex, m_lgraph))) {
				lgraphlogrenameedge(*ex, airwaydct);
				ee.set_ident(airwaydct);
				ee.set_metric(ee.get_metric() * get_dctpenalty());
				work = true;
				continue;
			}
			lgraphlogkilledge(*ex);
			remove_edge(*ex, m_lgraph);
			if (lgraphrestartedgescan) {
				done = false;
				break;
			}
		}
		if (done)
			break;
	}
	return work;
}

bool CFMUAutoroute::lgraphkilloutgoing(const Glib::ustring& id)
{
	Glib::ustring identu(id.uppercase());
	bool work(false);
	for (;;) {
		bool done(true);
		LGraph::edge_iterator e0, e1;
		for (tie(e0, e1) = edges(m_lgraph); e0 != e1; ) {
			LGraph::edge_iterator ex(e0);
			++e0;
			LEdge& ee(m_lgraph[*ex]);
			if (!ee.is_solution())
				continue;
			LGraph::vertex_descriptor v(source(*ex, m_lgraph));
			const LVertex& vv(m_lgraph[v]);
			const LSuperVertex& sv(m_lsupervertices[vv.get_superindex()]);
			if (sv.get_ident() != identu)
				continue;
			lgraphlogkilledge(*ex);
			remove_edge(*ex, m_lgraph);
			work = true;
			if (lgraphrestartedgescan) {
				done = false;
				break;
			}
		}
		if (done)
			break;
	}
	return work;
}

bool CFMUAutoroute::lgraphkillincoming(const Glib::ustring& id)
{
	Glib::ustring identu(id.uppercase());
	bool work(false);
	for (;;) {
		bool done(true);
		LGraph::edge_iterator e0, e1;
		for (tie(e0, e1) = edges(m_lgraph); e0 != e1; ) {
			LGraph::edge_iterator ex(e0);
			++e0;
			LEdge& ee(m_lgraph[*ex]);
			if (!ee.is_solution())
				continue;
			LGraph::vertex_descriptor v(target(*ex, m_lgraph));
			const LVertex& vv(m_lgraph[v]);
			const LSuperVertex& sv(m_lsupervertices[vv.get_superindex()]);
			if (sv.get_ident() != identu)
				continue;
			lgraphlogkilledge(*ex);
			remove_edge(*ex, m_lgraph);
			work = true;
			if (lgraphrestartedgescan) {
				done = false;
				break;
			}
		}
		if (done)
			break;
	}
	return work;
}

bool CFMUAutoroute::lgraphsolutiongroundclearance(void)
{
	m_signal_log(log_graphrule, "Ground Clearance");
	bool work(false);
	for (;;) {
		bool done(true);
		LGraph::edge_iterator e0, e1;
		for (tie(e0, e1) = edges(m_lgraph); e0 != e1;) {
			LGraph::edge_iterator ex(e0);
			++e0;
			LEdge& ee(m_lgraph[*ex]);
			if (!ee.is_solution())
				continue;
			LGraph::vertex_descriptor u(source(*ex, m_lgraph));
			const LVertex& uu(m_lgraph[u]);
			const LSuperVertex& su(m_lsupervertices[uu.get_superindex()]);
			if (su.get_v0() == su.get_v1())
				continue;
			LGraph::vertex_descriptor v(target(*ex, m_lgraph));
			const LVertex& vv(m_lgraph[v]);
			const LSuperVertex& sv(m_lsupervertices[vv.get_superindex()]);
			if (sv.get_v0() == sv.get_v1())
				continue;
			if (uu.get_superindex() == vv.get_superindex())
				continue;
			int32_t minalt(0);
			{
				TopoDb30::Profile prof(m_topodb.get_profile(su.get_coord(), sv.get_coord(), 5));
				if (prof.empty()) {
					std::ostringstream oss;
					oss << "Error computing ground clearance: " << lgraphvertexname(source(*ex, m_lgraph))
					    << ' ' << lgraphindex2airway(ee.get_ident(), true) << ' '
					    << lgraphvertexname(target(*ex, m_lgraph));
					m_signal_log(log_normal, oss.str());
					continue;
				}
				int32_t alt(prof.get_maxelev() * Point::m_to_ft);
				minalt = alt + 1000;
				if (alt >= 5000)
					minalt += 1000;
				{
					std::ostringstream oss;
					oss << "Leg: " << lgraphvertexname(source(*ex, m_lgraph))
					    << ' ' << lgraphindex2airway(ee.get_ident(), true) << ' '
					    << lgraphvertexname(target(*ex, m_lgraph))
					    << " ground elevation " << alt << "ft minimum level FL" << ((minalt + 99) / 100);
					m_signal_log(log_normal, oss.str());
				}
			}
			for (unsigned int i = 0; i < m_performance.size(); ++i) {
				const Performance& perf(m_performance[i]);
				if (perf.get_altitude() >= minalt)
					break;
				if (su.get_v0() + i > su.get_v1())
					break;
				if (sv.get_v0() + i > sv.get_v1())
					break;
				LGraph::edge_descriptor e;
                                bool ok;
				tie(e, ok) = boost::edge(su.get_v0() + i, sv.get_v0() + i, m_lgraph);
				if (ok) {
					LEdge& ee(m_lgraph[e]);
					if (ee.is_solution())
						work = true;
					done = false;
					lgraphlogkilledge(e);
					remove_edge(e, m_lgraph);
				}
				tie(e, ok) = boost::edge(sv.get_v0() + i, su.get_v0() + i, m_lgraph);
				if (ok) {
					LEdge& ee(m_lgraph[e]);
					if (ee.is_solution())
						work = true;
					done = false;
					lgraphlogkilledge(e);
					remove_edge(e, m_lgraph);
				}
			}
			if (!done)
				break;
		}
		if (done)
			break;
	}
	return work;
}

void CFMUAutoroute::parse_response(void)
{
	m_connchildtimeout.disconnect();
	if (!m_running) {
		m_signal_log(log_debug0, "Interrupted by user, stopping...");
		return;
	}
	m_signal_statuschange(statusmask_newvalidateresponse);
	bool work(m_validationresponse.size() == 1 && m_validationresponse[0] == "NO ERRORS");
	if (!work) {
		work = true;
		for (validationresponse_t::const_iterator i(m_validationresponse.begin()), e(m_validationresponse.end()); i != e; ++i) {
			if (i->empty())
				continue;
			const char * const *ignrx(ignoreregex);
			work = false;
			while (*ignrx) {
				Glib::RefPtr<Glib::Regex> rx(Glib::Regex::create(*ignrx++));
				Glib::MatchInfo minfo;
				if (rx->match(*i, minfo)) {
					work = true;
					break;
				}
			}
			if (!work)
				break;
		}
	}
	if (work) {
		if (m_pathprobe < 0) {
			if (lgraphsolutiongroundclearance()) {
				lgraphroute();
				return;
			}
			m_done = true;
			m_signal_log(log_fplproposal, get_simplefpl());
			stop();
			return;
		}
		for (;;) {
			++m_pathprobe;
			if (m_pathprobe + 1 >= m_route.get_nrwpt())
				break;
			const FPlanWaypoint& wpt(m_route[m_pathprobe]);
			if (wpt.get_flags() & FPlanWaypoint::altflag_ifr &&
			    wpt.get_pathcode() != FPlanWaypoint::pathcode_sid &&
			    wpt.get_pathcode() != FPlanWaypoint::pathcode_star)
				break;
		}
		if (m_pathprobe + 1 >= m_route.get_nrwpt()) {
			m_pathprobe = -1;
			m_signal_log(log_fplproposal, get_simplefpl());
			m_signal_log(log_normal, "No progress made; stopping...");
			stop();
			return;
		}
		lgraphroute();
		return;
	}
	work = false;
	for (validationresponse_t::const_iterator i(m_validationresponse.begin()), e(m_validationresponse.end()); i != e; ++i) {
		if (i->empty())
			continue;
		m_signal_log(log_graphrule, *i);
		{
			Glib::ustring::size_type n(i->rfind('['));
			if (n != Glib::ustring::npos) {
				Glib::ustring::size_type ne(i->find(']', n + 1));
				if (ne != Glib::ustring::npos && ne - n >= 2 && ne - n <= 11) {
					std::vector<TrafficFlowRestrictions::RuleResult> rr(m_tfr.find_rules(i->substr(n + 1, ne - n - 1), true));
					if (!rr.empty()) {
						m_signal_log(log_graphruledesc, rr[0].get_desc());
						m_signal_log(log_graphruleoprgoal, rr[0].get_oprgoal());
					}
				}
			}
		}
		if (false)
			std::cerr << "Message: " << *i << std::endl;
		const struct parsefunctions *pf(parsefunctions);
		for (; pf->regex; ++pf) {
			Glib::RefPtr<Glib::Regex> rx(Glib::Regex::create(pf->regex));
			Glib::MatchInfo minfo;
			if (!rx->match(*i, minfo))
				continue;
			if (pf->func)
				work = (this->*(pf->func))(minfo) || work;
			break;
		}
		if (!pf->regex) {
			const char * const *ignrx(ignoreregex);
			while (*ignrx) {
				Glib::RefPtr<Glib::Regex> rx(Glib::Regex::create(*ignrx++));
				Glib::MatchInfo minfo;
				if (!rx->match(*i, minfo))
					continue;
				break;
			}
			if (!*ignrx)
				m_signal_log(log_normal, "Error Message not understood: \"" + *i + "\"");
		}
	}
	if (!work && m_pathprobe >= 0 && m_pathprobe + 1 < m_route.get_nrwpt()) {
		Glib::ustring id0(m_route[m_pathprobe].get_icao());
		if (id0.empty())
			id0 = m_route[m_pathprobe].get_name();
		Glib::ustring id1(m_route[m_pathprobe + 1].get_icao());
		if (id1.empty())
			id1 = m_route[m_pathprobe + 1].get_name();
		work = lgraphmodifyawysegment(id0, id1, airwaymatchall, airwaymatchnone, 0, 600, false, true);
	}
	// disable path probing
	if (work || true) {
		if (m_pathprobe >= 0)
			m_signal_log(log_normal, "Resuming normal operation...");
		m_pathprobe = -1;
		lgraphroute();
		return;
	}
	for (;;) {
		++m_pathprobe;
		if (m_pathprobe + 1 >= m_route.get_nrwpt())
			break;
		const FPlanWaypoint& wpt(m_route[m_pathprobe]);
		if (wpt.get_flags() & FPlanWaypoint::altflag_ifr &&
		    wpt.get_pathcode() != FPlanWaypoint::pathcode_sid &&
		    wpt.get_pathcode() != FPlanWaypoint::pathcode_star)
			break;
	}
	if (m_pathprobe + 1 >= m_route.get_nrwpt()) {
		m_pathprobe = -1;
		m_signal_log(log_fplproposal, get_simplefpl());
		m_signal_log(log_normal, "No progress made; stopping...");
		stop();
		return;
	}
	if (m_pathprobe == 0)
		m_signal_log(log_normal, "No progress made; starting path probing...");
	lgraphroute();
}

const struct CFMUAutoroute::parsefunctions CFMUAutoroute::parsefunctions[] = {
	{ "^ROUTE49: THE POINT ([[:alnum:]]+) IS UNKNOWN IN THE CONTEXT OF THE ROUTE", &CFMUAutoroute::parse_response_route49 },
	{ "^ROUTE52: THE DCT SEGMENT ([[:alnum:]]+)..([[:alnum:]]+) IS FORBIDDEN", &CFMUAutoroute::parse_response_route52 },
	{ "^ROUTE130: UNKNOWN DESIGNATOR ([[:alnum:]]+)", &CFMUAutoroute::parse_response_route130 },
	{ "^ROUTE134: THE STAR LIMIT IS EXCEEDED FOR AERODROME .*? CONNECTING TO ([[:alnum:]]+).", &CFMUAutoroute::parse_response_route134 },
	{ "^ROUTE135: THE SID LIMIT IS EXCEEDED FOR AERODROME .*? CONNECTING TO ([[:alnum:]]+).", &CFMUAutoroute::parse_response_route135 },
	{ "^ROUTE139: ([[:alnum:]]+) IS PRECEDED BY ([[:alnum:]]+) WHICH IS NOT ONE OF ITS POINTS", &CFMUAutoroute::parse_response_route139 },
	{ "^ROUTE140: ([[:alnum:]]+) IS FOLLOWED BY ([[:alnum:]]+) WHICH IS NOT ONE OF ITS POINTS", &CFMUAutoroute::parse_response_route140 },
	{ "^ROUTE165: THE DCT SEGMENT ([[:alnum:]]+)..([[:alnum:]]+)", &CFMUAutoroute::parse_response_route165 },
	{ "^ROUTE168: INVALID DCT ([[:alnum:]]+)..([[:alnum:]]+).", &CFMUAutoroute::parse_response_route168 },
	{ "^ROUTE172: MULTIPLE ROUTES BETWEEN ([[:alnum:]]+) AND ([[:alnum:]]+). ([[:alnum:]]+) IS SUGGESTED.", &CFMUAutoroute::parse_response_route172 },
	{ "^PROF50: CLIMBING/DESCENDING OUTSIDE THE VERTICAL LIMITS OF SEGMENT ([[:alnum:]]+) ([[:alnum:]]+) ([[:alnum:]]+)", &CFMUAutoroute::parse_response_prof50 },
	{ "^PROF194: ([[:alnum:]]+) ([[:alnum:]]+) ([[:alnum:]]+) IS NOT AVAILABLE IN FL RANGE F([[:digit:]]+)..F([[:digit:]]+)", &CFMUAutoroute::parse_response_prof194 },
	{ "^PROF195: ([[:alnum:]]+) ([[:alnum:]]+) ([[:alnum:]]+) DOES NOT EXIST IN FL RANGE F([[:digit:]]+)..F([[:digit:]]+)", &CFMUAutoroute::parse_response_prof195 },
	{ "^PROF197: RS: ([[:alnum:]]+) ([[:alnum:]]+) ([[:alnum:]]+):F([[:digit:]]+)..F([[:digit:]]+) IS CLOSED FOR CRUISING", &CFMUAutoroute::parse_response_prof197 },
	{ "^PROF198: ([[:alnum:]]+) ([[:alnum:]]+) ([[:alnum:]]+) IS A CDR 3 IN FL RANGE F([[:digit:]]+)..F([[:digit:]]+)", &CFMUAutoroute::parse_response_prof198 },
	{ "^PROF199: ([[:alnum:]]+) ([[:alnum:]]+) ([[:alnum:]]+) IS A CLOSED CDR 2 IN FL RANGE F([[:digit:]]+)..F([[:digit:]]+)", &CFMUAutoroute::parse_response_prof199 },
	{ "^PROF201: CANNOT CLIMB OR DESCEND ON ([[:alnum:]]+) ([[:alnum:]]+) ([[:alnum:]]+) IN FL RANGE CLOSED", &CFMUAutoroute::parse_response_prof201 },
	{ "^PROF201: CANNOT CLIMB OR DESCEND ON ([[:alnum:]]+) ([[:alnum:]]+) ([[:alnum:]]+) IN FL RANGE F([[:digit:]]+)..F([[:digit:]]+)", &CFMUAutoroute::parse_response_prof201b },
	{ "^PROF204: RS: TRAFFIC VIA ((?:[[:alnum:]]+)(?: [[:alnum:]]+)*?) IS ON FORBIDDEN ROUTE", &CFMUAutoroute::parse_response_prof204 },
	{ "^PROF204: RS: TRAFFIC VIA ((?:[[:alnum:]]+)(?: [[:alnum:]]+)*?):F([[:digit:]]+)..F([[:digit:]]+) IS ON FORBIDDEN ROUTE", &CFMUAutoroute::parse_response_prof204b },
	{ "^PROF204: RS: TRAFFIC VIA ([[:alnum:]]+) ([[:alnum:]]+):F([[:digit:]]+)..F([[:digit:]]+) IS ON FORBIDDEN ROUTE", &CFMUAutoroute::parse_response_prof204c },
	{ "^PROF204: RS: TRAFFIC VIA ([[:alnum:]]+) ([[:alnum:]]+) ([[:alnum:]]+):F([[:digit:]]+)..F([[:digit:]]+) IS ON FORBIDDEN ROUTE", &CFMUAutoroute::parse_response_prof204d },
	{ "^PROF205: RS: TRAFFIC VIA ([[:alnum:]]+) IS OFF MANDATORY ROUTE", &CFMUAutoroute::parse_response_prof205 },
	{ "^PROF205: RS: TRAFFIC VIA ([[:alnum:]]+) ([[:alnum:]]+) ([[:alnum:]]+) IS OFF MANDATORY ROUTE", &CFMUAutoroute::parse_response_prof205b },
	{ "^PROF205: RS: TRAFFIC VIA ([[:alnum:]]+) ([[:alnum:]]+) ([[:alnum:]]+):F([[:digit:]]+)..F([[:digit:]]+) IS OFF MANDATORY ROUTE", &CFMUAutoroute::parse_response_prof205c },
	{ "^PROF206: THE DCT SEGMENT ([[:alnum:]]+) .. ([[:alnum:]]+) IS NOT AVAILABLE IN FL RANGE F([[:digit:]]+)..F([[:digit:]]+)", &CFMUAutoroute::parse_response_prof206 },
	{ 0, 0 }
};

const char *CFMUAutoroute::ignoreregex[] = {
	// ignore 8.33 carriage errors
	"^PROF188:",
	"^PROF189:",
	"^PROF190:",
	0
};

bool CFMUAutoroute::parse_response_route49(Glib::MatchInfo& minfo)
{
	return lgraphdisconnectvertex(minfo.fetch(1).uppercase());
}

bool CFMUAutoroute::parse_response_route52(Glib::MatchInfo& minfo)
{
	return lgraphmodifyedge(minfo.fetch(1), minfo.fetch(2), airwaydct, airwaymatchnone, 0, 600, true);
}

bool CFMUAutoroute::parse_response_route130(Glib::MatchInfo& minfo)
{
	Glib::ustring idu(minfo.fetch(1).uppercase());
	bool work(lgraphdisconnectvertex(idu));
	lgraphairwayindex_t awy(lgraphairway2index(idu));
	if (awy != airwaydct && awy != airwaymatchnone)
		work = lgraphedgetodct(awy) || work;
	return work;
}

bool CFMUAutoroute::parse_response_route134(Glib::MatchInfo& minfo)
{
	bool work(false);
	for (;;) {
		bool done(true);
		LGraph::edge_iterator e0, e1;
		for (tie(e0, e1) = edges(m_lgraph); e0 != e1; ) {
			LGraph::edge_iterator ex(e0);
			++e0;
			LEdge& ee(m_lgraph[*ex]);
			if (!ee.is_solution())
				continue;
			if (!ee.is_match(airwaystar))
				continue;
			lgraphlogkilledge(*ex);
			remove_edge(*ex, m_lgraph);
			work = true;
			if (lgraphrestartedgescan) {
				done = false;
				break;
			}
		}
		if (done)
			break;
	}
	return work;
}

bool CFMUAutoroute::parse_response_route135(Glib::MatchInfo& minfo)
{
	bool work(false);
	for (;;) {
		bool done(true);
		LGraph::edge_iterator e0, e1;
		for (tie(e0, e1) = edges(m_lgraph); e0 != e1; ) {
			LGraph::edge_iterator ex(e0);
			++e0;
			LEdge& ee(m_lgraph[*ex]);
			if (!ee.is_solution())
				continue;
			if (!ee.is_match(airwaysid))
				continue;
			lgraphlogkilledge(*ex);
			remove_edge(*ex, m_lgraph);
			work = true;
			if (lgraphrestartedgescan) {
				done = false;
				break;
			}
		}
		if (done)
			break;
	}
	return work;
}

bool CFMUAutoroute::parse_response_route139(Glib::MatchInfo& minfo)
{
	lgraphairwayindex_t awy(lgraphairway2index(minfo.fetch(1)));
	if (awy == airwaymatchnone || awy == airwaydct)
		return false;
	return lgraphchangeoutgoing(minfo.fetch(2), awy);
}

bool CFMUAutoroute::parse_response_route140(Glib::MatchInfo& minfo)
{
	lgraphairwayindex_t awy(lgraphairway2index(minfo.fetch(1)));
	if (awy == airwaymatchnone || awy == airwaydct)
		return false;
	return lgraphchangeincoming(minfo.fetch(2), awy);
}

bool CFMUAutoroute::parse_response_route165(Glib::MatchInfo& minfo)
{
	return lgraphmodifyedge(minfo.fetch(1), minfo.fetch(2), airwaydct, airwaymatchnone, 0, 600, true);
}

bool CFMUAutoroute::parse_response_route168(Glib::MatchInfo& minfo)
{
	return lgraphmodifyedge(minfo.fetch(1), minfo.fetch(2), airwaydct, airwaymatchnone, 0, 600, true);
}

bool CFMUAutoroute::parse_response_route172(Glib::MatchInfo& minfo)
{
	lgraphairwayindex_t awy(lgraphairway2index(minfo.fetch(3), true));
	if (awy == airwaydct)
		return false;
	return lgraphmodifyedge(minfo.fetch(1), minfo.fetch(2), airwaydct, awy, 0, 600, false);
}

bool CFMUAutoroute::parse_response_prof50(Glib::MatchInfo& minfo)
{
	lgraphairwayindex_t awy(lgraphairway2index(minfo.fetch(2)));
	if (awy != airwaymatchnone && lgraphmodifyawysegment(minfo.fetch(1), minfo.fetch(3), awy, airwaymatchnone, 0, 600, true, true))
		return true;
	return lgraphmodifyawysegment(minfo.fetch(1), minfo.fetch(3), airwaydct, airwaymatchnone, 0, 600, true, true);
}

bool CFMUAutoroute::parse_response_prof194(Glib::MatchInfo& minfo)
{
	return parse_response_prof195(minfo);
}

bool CFMUAutoroute::parse_response_prof195(Glib::MatchInfo& minfo)
{
	int altfrom(strtol(minfo.fetch(4).c_str(), 0, 10));
	int altto(strtol(minfo.fetch(5).c_str(), 0, 10));
	if (altfrom > altto)
		return false;
	lgraphairwayindex_t awy(lgraphairway2index(minfo.fetch(2)));
	if (awy != airwaymatchnone && lgraphmodifyawysegment(minfo.fetch(1), minfo.fetch(3), awy, airwaymatchnone, altfrom, altto, true, false))
		return true;
	return lgraphmodifyawysegment(minfo.fetch(1), minfo.fetch(3), airwaydct, airwaymatchnone, altfrom, altto, true, false);
}

bool CFMUAutoroute::parse_response_prof197(Glib::MatchInfo& minfo)
{
	int altfrom(strtol(minfo.fetch(4).c_str(), 0, 10));
	int altto(strtol(minfo.fetch(5).c_str(), 0, 10));
	if (altfrom > altto)
		return false;
	lgraphairwayindex_t awy(lgraphairway2index(minfo.fetch(2)));
	if (awy != airwaymatchnone && lgraphmodifyawysegment(minfo.fetch(1), minfo.fetch(3), awy, airwaymatchnone, altfrom, altto, true, true))
		return true;
	return lgraphmodifyawysegment(minfo.fetch(1), minfo.fetch(3), airwaydct, airwaymatchnone, altfrom, altto, true, true);
}

bool CFMUAutoroute::parse_response_prof198(Glib::MatchInfo& minfo)
{
	return parse_response_prof195(minfo);
}

bool CFMUAutoroute::parse_response_prof199(Glib::MatchInfo& minfo)
{
	return parse_response_prof195(minfo);
}

bool CFMUAutoroute::parse_response_prof201(Glib::MatchInfo& minfo)
{
	return parse_response_prof50(minfo);
}

bool CFMUAutoroute::parse_response_prof201b(Glib::MatchInfo& minfo)
{
	int altfrom(strtol(minfo.fetch(4).c_str(), 0, 10));
	int altto(strtol(minfo.fetch(5).c_str(), 0, 10));
	if (altfrom > altto)
		return false;
	lgraphairwayindex_t awy(lgraphairway2index(minfo.fetch(2)));
	if (awy != airwaymatchnone && lgraphmodifyawysegment(minfo.fetch(1), minfo.fetch(3), awy, airwaymatchnone, altfrom, altto, true, true))
		return true;
	return lgraphmodifyawysegment(minfo.fetch(1), minfo.fetch(3), airwaydct, airwaymatchnone, altfrom, altto, true, true);
}

bool CFMUAutoroute::parse_response_prof204(Glib::MatchInfo& minfo)
{
	bool work(false);
	if (false)
		std::cerr << "PROF204: \"" << minfo.fetch(1) << "\" \"" << minfo.fetch(0) << "\"" << std::endl;
	typedef std::vector<Glib::ustring> token_t;
	token_t token(Glib::Regex::split_simple(" +", minfo.fetch(1)));
 	for (int i = 0; i < token.size(); ++i) {
 	 	if (i + 2 < token.size() &&
		    !Engine::AirwayGraphResult::Vertex::is_ident_numeric(token[i]) &&
		    Engine::AirwayGraphResult::Vertex::is_ident_numeric(token[i + 1]) &&
		    !Engine::AirwayGraphResult::Vertex::is_ident_numeric(token[i + 2])) {
			if (true)
				std::cerr << "PROF204: kill " << token[i] << "--" << token[i + 1] << "->" << token[i + 2] << std::endl;
			lgraphairwayindex_t awy(lgraphairway2index(token[i + 1]));
			if (awy != airwaymatchnone && awy != airwaydct) {
				bool work2(lgraphmodifyawysegment(token[i], token[i + 2], awy, airwaymatchnone, 0, 600, true, true));
				if (!work2) {
					if (true)
						std::cerr << "PROF204: " << token[i] << "--" << lgraphindex2airway(awy) << "->" << token[i + 2]
							  << " did not match, trying DCT" << std::endl;
					work2 = lgraphmodifyawysegment(token[i], token[i + 2], airwaydct, airwaymatchnone, 0, 600, true, true);
					if (!work2) {
						if (true)
							std::cerr << "PROF204: " << token[i] << "--->" << token[i + 2] << " did not match, trying airway convert to DCT" << std::endl;
						work2 = lgraphedgetodct(awy);
						if (!work2) {
							std::cerr << "PROF204: could not convert " << lgraphindex2airway(awy)
								  << " to DCT, removing incoming/outgoing edges" << std::endl;
							work2 = lgraphkilloutgoing(token[i]);
							work2 = lgraphkillincoming(token[i + 2]) || work2;
						}
					}
				}
				work = work2 || work;
			}
			i += 2;
			continue;
		}
		if (true)
			std::cerr << "PROF204: kill " << token[i] << std::endl;
		work = lgraphdisconnectsolutionvertex(token[i]) || work;
		if (!work && token[i].size() >= 4) {
			opendb();
			AirspacesDb::elementvector_t ev(m_airspacedb.find_by_icao(token[i], 0, /* AirspacesDb::comp_startswith */ AirspacesDb::comp_exact,
										  0, AirspacesDb::Airspace::subtables_all));
			for (AirspacesDb::elementvector_t::const_iterator ei(ev.begin()), ee(ev.end()); ei != ee; ++ei) {
				const AirspacesDb::Airspace& aspc(*ei);
				if (!aspc.is_valid() || aspc.get_typecode() != AirspacesDb::Airspace::typecode_ead)
					continue;
				{
					std::ostringstream oss;
					oss << "A " << aspc.get_icao_name();
					m_signal_log(log_graphchange, oss.str());
				}
				const MultiPolygonHole& poly(aspc.get_polygon());
				Rect bbox(poly.get_bbox());
				for (;;) {
					bool done(true);
					LGraph::edge_iterator e0, e1;
					for (tie(e0, e1) = edges(m_lgraph); e0 != e1; ) {
						LGraph::edge_iterator ex(e0);
						++e0;
						LEdge& ee(m_lgraph[*ex]);
						if (!ee.is_solution())
							continue;
						LGraph::vertex_descriptor u(source(*ex, m_lgraph));
						const LVertex& uu(m_lgraph[u]);
						const LSuperVertex& su(m_lsupervertices[uu.get_superindex()]);
						LGraph::vertex_descriptor v(target(*ex, m_lgraph));
						const LVertex& vv(m_lgraph[v]);
						const LSuperVertex& sv(m_lsupervertices[vv.get_superindex()]);
						bool insideu(bbox.is_inside(su.get_coord()) && poly.windingnumber(su.get_coord()));
						bool insidev(bbox.is_inside(sv.get_coord()) && poly.windingnumber(sv.get_coord()));
						bool deledge(insideu || insidev || poly.is_intersection(su.get_coord(), sv.get_coord()));
						if (!deledge)
							continue;
						lgraphlogkilledge(*ex);
						remove_edge(*ex, m_lgraph);
						work = true;
						if (lgraphrestartedgescan) {
							done = false;
							break;
						}
					}
					if (done)
						break;
				}
			}
		}
	}
	return work;
}

bool CFMUAutoroute::parse_response_prof204b(Glib::MatchInfo& minfo)
{
	int altfrom(strtol(minfo.fetch(2).c_str(), 0, 10));
	int altto(strtol(minfo.fetch(3).c_str(), 0, 10));
	if (altfrom > altto)
		return false;
	bool work(false);
	if (false)
		std::cerr << "PROF204: \"" << minfo.fetch(1) << "\" \"" << minfo.fetch(0) << "\"" << std::endl;
	typedef std::vector<Glib::ustring> token_t;
	token_t token(Glib::Regex::split_simple(" +", minfo.fetch(1)));
 	for (int i = 0; i < token.size(); ++i) {
 	 	if (i + 2 < token.size() &&
		    !Engine::AirwayGraphResult::Vertex::is_ident_numeric(token[i]) &&
		    Engine::AirwayGraphResult::Vertex::is_ident_numeric(token[i + 1]) &&
		    !Engine::AirwayGraphResult::Vertex::is_ident_numeric(token[i + 2])) {
			if (true)
				std::cerr << "PROF204: kill " << token[i] << "--" << token[i + 1] << "->" << token[i + 2] << std::endl;
			lgraphairwayindex_t awy(lgraphairway2index(token[i + 1]));
			if (awy != airwaymatchnone && awy != airwaydct) {
				bool work2(lgraphmodifyawysegment(token[i], token[i + 2], awy, airwaymatchnone, altfrom, altto, true, true));
				if (!work2) {
					if (true)
						std::cerr << "PROF204: " << token[i] << "--" << lgraphindex2airway(awy) << "->" << token[i + 2]
							  << " did not match, trying DCT" << std::endl;
					work2 = lgraphmodifyawysegment(token[i], token[i + 2], airwaydct, airwaymatchnone, altfrom, altto, true, true);
					if (!work2) {
						if (true)
							std::cerr << "PROF204: " << token[i] << "--->" << token[i + 2] << " did not match, trying airway convert to DCT" << std::endl;
						work2 = lgraphedgetodct(awy);
						if (!work2) {
							std::cerr << "PROF204: could not convert " << lgraphindex2airway(awy)
								  << " to DCT, removing incoming/outgoing edges" << std::endl;
							work2 = lgraphkilloutgoing(token[i]);
							work2 = lgraphkillincoming(token[i + 2]) || work2;
						}
					}
				}
				work = work2 || work;
			}
			i += 2;
			continue;
		}
		if (true)
			std::cerr << "PROF204: kill " << token[i] << std::endl;
		work = lgraphdisconnectvertex(token[i], altfrom, altto) || work;
	}
	return work;
}

bool CFMUAutoroute::parse_response_prof204c(Glib::MatchInfo& minfo)
{
	int altfrom(strtol(minfo.fetch(3).c_str(), 0, 10));
	int altto(strtol(minfo.fetch(4).c_str(), 0, 10));
	if (altfrom > altto)
		return false;
	return lgraphmodifyedge(minfo.fetch(1), minfo.fetch(2), airwaymatchdctawy, airwaymatchnone, altfrom, altto, true);
}

bool CFMUAutoroute::parse_response_prof204d(Glib::MatchInfo& minfo)
{
	int altfrom(strtol(minfo.fetch(4).c_str(), 0, 10));
	int altto(strtol(minfo.fetch(5).c_str(), 0, 10));
	if (altfrom > altto)
		return false;
	lgraphairwayindex_t awy(lgraphairway2index(minfo.fetch(2)));
	if (awy != airwaymatchnone && lgraphmodifyawysegment(minfo.fetch(1), minfo.fetch(3), awy, airwaymatchnone, altfrom, altto, true, true))
		return true;
	return lgraphmodifyawysegment(minfo.fetch(1), minfo.fetch(3), airwaydct, airwaymatchnone, altfrom, altto, true, true);
}

bool CFMUAutoroute::parse_response_prof205(Glib::MatchInfo& minfo)
{
	return lgraphdisconnectsolutionvertex(minfo.fetch(1));
}

bool CFMUAutoroute::parse_response_prof205b(Glib::MatchInfo& minfo)
{
	return parse_response_prof50(minfo);
}

bool CFMUAutoroute::parse_response_prof205c(Glib::MatchInfo& minfo)
{
	return parse_response_prof204d(minfo);
}

bool CFMUAutoroute::parse_response_prof206(Glib::MatchInfo& minfo)
{
	int altfrom(strtol(minfo.fetch(3).c_str(), 0, 10));
	int altto(strtol(minfo.fetch(4).c_str(), 0, 10));
	if (altfrom > altto)
		return false;
	return lgraphmodifyedge(minfo.fetch(1), minfo.fetch(2), airwaydct, airwaymatchnone, altfrom, altto, true);
}
//PROF201: CANNOT CLIMB OR DESCEND ON SOPER N851 PIXOS IN FL RANGE CLOSED, BECAUSE OF UNAVAILABLE LEVELS ON N851
//PROF50: CLIMBING/DESCENDING OUTSIDE THE VERTICAL LIMITS OF SEGMENT NUNRI T103 KUDIS
//PROF205: RS: TRAFFIC VIA EGHL+4 IS OFF MANDATORY ROUTE REF:[EG2380D] GWC COMP FOR TRAFFIC 1.DEP FARNBOROUGH G
//PROF205: RS: TRAFFIC VIA EGHH EGHI EGHO EGHL+4 IS OFF MANDATORY ROUTE REF:[EGOG5112A] GWC DCT SFD APP.4
//PROF204: RS: TRAFFIC VIA EDGGADF IS ON FORBIDDEN ROUTE REF:[ED3115A] EDGGADF

// Airspace BG has 49 navaids, 108 intersections
// Airspace BI has 82 navaids, 77 intersections
// Airspace CFMU1 CFMU1 has 3893 navaids, 14369 intersections
// Airspace CY-AERODB has 1294 navaids, 14666 intersections
// Airspace CY has 987 navaids, 10145 intersections
// Airspace DA has 89 navaids, 128 intersections
// Airspace DT has 25 navaids, 154 intersections
// Airspace EB MIL EB MIL DOWNLOAD has 114 navaids, 322 intersections
// Airspace EB-EAD-1 has 85 navaids, 194 intersections
// Airspace EB-EAD-2 has 0 navaids, 3 intersections
// Airspace EB-EAD-3 has 1 navaids, 3 intersections
// Airspace EB-EAD-5 has 3 navaids, 18 intersections
// Airspace EBDWNLOAD1 EBDOWNLOAD1 has 681 navaids, 3320 intersections
// Airspace EBDWNLOAD2 EBDOWNLOAD2 has 231 navaids, 1035 intersections
// Airspace ED has 355 navaids, 3156 intersections
// Airspace EE has 20 navaids, 94 intersections
// Airspace EF has 104 navaids, 385 intersections
// Airspace EG2 has 56 navaids, 143 intersections
// Airspace EG has 362 navaids, 1126 intersections
// Airspace EH-EAD-1 EH-EAD-1 has 101 navaids, 605 intersections
// Airspace EH-EAD-2 EH-EAD-2 has 0 navaids, 33 intersections
// Airspace EI-EAD-1 EI-EAD-1 has 60 navaids, 217 intersections
// Airspace EI-EAD-2 EI-EAD-2 has 0 navaids, 6 intersections
// Airspace EI-EAD-3 EI-EAD-3 has 0 navaids, 1 intersections
// Airspace EI-EAD-4 EI-EAD-4 has 0 navaids, 22 intersections
// Airspace EI-EAD-5 EI-EAD-5 has 0 navaids, 19 intersections
// Airspace EK-EAD-1 EK-EAD-1 has 74 navaids, 403 intersections
// Airspace EK-EAD-2 EK-EAD-2 has 0 navaids, 51 intersections
// Airspace EK-EAD-3 EK-EAD-3 has 5 navaids, 26 intersections
// Airspace EK-EAD-4 EK-EAD-4 has 6 navaids, 20 intersections
// Airspace EN-EAD-1 EN EAD FIR PLUS DELEGATED AREAS has 270 navaids, 1185 intersections
// Airspace EN-EAD-2 NO_FIR AREA NEAR NORWAY has 0 navaids, 10 intersections
// Airspace EN-EAD-3 JAN MAYEN ISLAND (JAN NDB LOCATION) has 1 navaids, 0 intersections
// Airspace EP has 146 navaids, 584 intersections
// Airspace ES has 241 navaids, 1028 intersections
// Airspace EV-EAD-1 EV-EAD-1 has 14 navaids, 137 intersections
// Airspace EV-EAD-3 EV-EAD-3 has 3 navaids, 17 intersections
// Airspace EV-EAD-4 EV-EAD-4 has 2 navaids, 9 intersections
// Airspace EV-EAD-5 EV-EAD-5 has 0 navaids, 1 intersections
// Airspace EY has 17 navaids, 159 intersections
// Airspace FA has 187 navaids, 256 intersections
// Airspace FABEC1N FABEC PROJECT PHASE 1 NORTH has 57 navaids, 531 intersections
// Airspace FABEC1W FABEC PROJECT PHASE 1 WEST has 102 navaids, 305 intersections
// Airspace FABEC2N FABEC PROJECT PHASE 2 NORTH has 159 navaids, 814 intersections
// Airspace FABECSOUTH FABEC PROJECT PHASE SOUTH has 184 navaids, 852 intersections
// Airspace FD_EAD has 4 navaids, 6 intersections
// Airspace FX_EAD has 4 navaids, 3 intersections
// Airspace FY_EAD has 4 navaids, 3 intersections
// Airspace G1AA AREA OF RESPONSIBILITY 1 has 5384 navaids, 18698 intersections
// Airspace G1BB AREA OF RESPONSIBILITY 2 has 453 navaids, 6887 intersections
// Airspace G1CC AREA OF RESPONSIBILITY 3 has 2333 navaids, 9117 intersections
// Airspace GM has 59 navaids, 158 intersections
// Airspace GO has 97 navaids, 334 intersections
// Airspace GV has 7 navaids, 39 intersections
// Airspace HE has 56 navaids, 93 intersections
// Airspace HL has 61 navaids, 123 intersections
// Airspace KCHART1 COVERAGE OF CHARTS U.S.A. has 969 navaids, 13909 intersections
// Airspace LA has 5 navaids, 30 intersections
// Airspace LB has 36 navaids, 118 intersections
// Airspace LC has 10 navaids, 67 intersections
// Airspace LD-EAD-1 EAD AIRSPACE CROATIA has 43 navaids, 131 intersections
// Airspace LD-EAD-2 EAD AIRSPACE DELEGATED TO CROATIA has 1 navaids, 4 intersections
// Airspace LD-EAD-3 AIRSPACE DELEGATION TO CROATIA has 0 navaids, 10 intersections
// Airspace LE_1 has 265 navaids, 826 intersections
// Airspace LE_2 has 50 navaids, 115 intersections
// Airspace LF has 482 navaids, 1549 intersections
// Airspace LG has 119 navaids, 225 intersections
// Airspace LH has 79 navaids, 252 intersections
// Airspace LI has 269 navaids, 766 intersections
// Airspace LICD LAMPEDUSA has 4 navaids, 5 intersections
// Airspace LJ has 17 navaids, 79 intersections
// Airspace LK has 97 navaids, 443 intersections
// Airspace LL has 28 navaids, 51 intersections
// Airspace LM has 14 navaids, 93 intersections
// Airspace LO has 50 navaids, 214 intersections
// Airspace LP has 65 navaids, 351 intersections
// Airspace LQ has 29 navaids, 65 intersections
// Airspace LR has 72 navaids, 360 intersections
// Airspace LS has 42 navaids, 204 intersections
// Airspace LT has 207 navaids, 403 intersections
// Airspace LU has 11 navaids, 49 intersections
// Airspace LW has 12 navaids, 29 intersections
// Airspace LY has 61 navaids, 172 intersections
// Airspace LZ-EAD-1 LZ-EAD-1 has 32 navaids, 138 intersections
// Airspace LZ-EAD-2 LZ-EAD-2 has 0 navaids, 0 intersections
// Airspace NZ-LNIA LOWER NORTH ISLAND ARC has 40 navaids, 434 intersections
// Airspace NZ-NOEN NORTH ENRC has 64 navaids, 885 intersections
// Airspace NZ-SOEN SOUTH ENRC has 74 navaids, 939 intersections
// Airspace NZ-UNIA UPPER NORTH ISLAND ARC has 33 navaids, 500 intersections
// Airspace NZ-USIA UPPER SOUTH ISLAND ARC has 36 navaids, 366 intersections
// Airspace NZ has 557 navaids, 4093 intersections
// Airspace OI has 180 navaids, 305 intersections
// Airspace OJ has 22 navaids, 59 intersections
// Airspace OL has 7 navaids, 30 intersections
// Airspace OS has 24 navaids, 40 intersections
// Airspace RC has 71 navaids, 97 intersections
// Airspace RP has 92 navaids, 295 intersections
// Airspace UB-EAD-1 AZERBAIJAN EAD AREA EAST has 30 navaids, 80 intersections
// Airspace UB-EAD-2 AZERBAIJAN EAD AREA WEST has 5 navaids, 3 intersections
// Airspace UD has 13 navaids, 36 intersections
// Airspace UG has 20 navaids, 35 intersections
// Airspace UK has 149 navaids, 401 intersections
// Airspace UM BELARUS has 32 navaids, 153 intersections
// Airspace UT-2 TAJIKISTAN has 9 navaids, 26 intersections
// Airspace U_ADR-1 has 187 navaids, 594 intersections
// Airspace U_ADR-2 has 2 navaids, 24 intersections
// Airspace WS SINGAPORE has 26 navaids, 164 intersections
// Airspace XX has 8 navaids, 3 intersections
